"""
Generates a 3-D colon mesh along the central line, with variable
numbers of elements around, along and through wall, with
variable radius and thickness along.
"""

import copy

from cmlibs.zinc.element import Element
from cmlibs.zinc.field import Field
from cmlibs.zinc.node import Node
from scaffoldmaker.annotation.annotationgroup import AnnotationGroup, findOrCreateAnnotationGroupForTerm, \
    getAnnotationGroupForTerm
from scaffoldmaker.annotation.colon_terms import get_colon_term
from scaffoldmaker.meshtypes.meshtype_1d_path1 import MeshType_1d_path1
from scaffoldmaker.meshtypes.meshtype_3d_colonsegment1 import MeshType_3d_colonsegment1, \
    ColonSegmentTubeMeshInnerPoints, getTeniaColi, createFlatCoordinatesTeniaColi, createColonCoordinatesTeniaColi, \
    createNodesAndElementsTeniaColi
from scaffoldmaker.meshtypes.scaffold_base import Scaffold_base
from scaffoldmaker.scaffoldpackage import ScaffoldPackage
from scaffoldmaker.utils import interpolation as interp
from scaffoldmaker.utils import tubemesh
from scaffoldmaker.utils import vector
from scaffoldmaker.utils.zinc_utils import exnode_string_from_nodeset_field_parameters, \
    get_nodeset_path_field_parameters


class MeshType_3d_colon1(Scaffold_base):
    '''
    Generates a 3-D colon mesh with variable numbers
    of elements around, along the central line, and through wall.
    The colon is created by a function that generates a colon
    segment and uses tubemesh to map the segment along a central
    line profile.
    '''

    centralPathDefaultScaffoldPackages = {
        'Cattle 1': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 52
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [ Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3 ], [
                (1,  [ [ -245.30, 444.60, -49.10 ], [ -267.70,  -53.10, -20.20 ], [   0.00,   0.00,  35.00 ], [   0.00,  0.00,  -4.41 ], [  -6.81,  34.33,  0.00 ], [   0.00,  0.00,  -4.41] ] ),
                (2,  [ [ -380.30, 484.80, -45.00 ], [   24.50,  102.70,  15.70 ], [   0.00,   0.00,  31.47 ], [   0.00,  0.00,  -2.65 ], [  30.61,  -7.30,  0.00 ], [   0.00,  0.00,  -2.65] ] ),
                (3,  [ [ -298.10, 510.40, -36.80 ], [   73.60,    9.90, -16.40 ], [   0.00,   0.00,  29.34 ], [ -12.52, -3.10, -14.34 ], [   3.91, -29.08,  0.00 ], [ -12.52, -3.10, -14.34] ] ),
                (4,  [ [ -213.10, 527.90, -22.50 ], [   -1.00,  -10.80, 125.60 ], [ -26.23,  -6.50,   1.62 ], [  -4.10, -6.11, -25.37 ], [   6.36, -26.22, -2.20 ], [  -4.10, -6.11, -25.37] ] ),
                (5,  [ [ -315.50, 570.20,  18.90 ], [ -107.90,    9.30,  21.90 ], [  -2.62, -12.12, -20.82 ], [  14.35,  4.82, -10.40 ], [   0.65, -20.97, 12.13 ], [  14.35,  4.82, -10.40] ] ),
                (6,  [ [ -417.40, 555.00,  14.60 ], [  -83.00,  -41.30,  -0.80 ], [   4.22,   1.17, -21.45 ], [   0.54,  4.88,   0.73 ], [   9.74, -19.59,  0.85 ], [   0.54,  4.88,   0.73] ] ),
                (7,  [ [ -497.30, 488.90,  13.60 ], [  -44.60,  -81.60,  10.00 ], [  -1.56,  -2.39, -19.34 ], [  -2.12, -0.60,   2.13 ], [  17.14,  -9.40, -0.22 ], [  -2.12, -0.60,   2.13] ] ),
                (8,  [ [ -527.00, 392.50,   2.70 ], [   47.40,  -82.00,  -7.90 ], [   0.00,   0.00, -17.20 ], [   0.68,  1.04,   1.96 ], [  14.89,   8.61,  0.00 ], [   0.68,  1.04,   1.96] ] ),
                (9,  [ [ -461.20, 345.90,  -0.80 ], [   56.90,  -44.50,   2.40 ], [   0.00,   0.00, -15.38 ], [   0.00,  0.00,   1.09 ], [   9.47,  12.11,  0.00 ], [   0.00,  0.00,   1.09] ] ),
                (10, [ [ -415.60, 293.80,   3.90 ], [   93.20,  -62.60,   3.10 ], [   0.00,   0.00, -14.91 ], [   0.00,  0.00,   0.43 ], [   8.31,  12.38,  0.00 ], [   0.00,  0.00,   0.43] ] ),
                (11, [ [ -232.20, 264.90,   0.20 ], [  140.10,   58.20,  -1.00 ], [   0.00,   0.00, -14.58 ], [   0.00,  0.00,   0.25 ], [  -5.59,  13.46,  0.00 ], [   0.00,  0.00,   0.25] ] ),
                (12, [ [ -168.40, 357.20,   1.30 ], [   10.10,   78.60,  -3.20 ], [   0.00,   0.00, -14.38 ], [   0.00,  0.00,   0.15 ], [ -14.26,   1.83,  0.00 ], [   0.00,  0.00,   0.15] ] ),
                (13, [ [ -185.30, 419.10,  -0.70 ], [  -45.10,   57.10,  -0.90 ], [   0.00,   0.00, -14.27 ], [   0.00,  0.00,   0.13 ], [ -11.20,  -8.84,  0.00 ], [   0.00,  0.00,   0.13] ] ),
                (14, [ [ -253.20, 466.70,  -0.30 ], [  -63.40,   24.70,   0.20 ], [   0.00,   0.00, -14.13 ], [   0.00,  0.00,   0.13 ], [  -5.13, -13.17,  0.00 ], [   0.00,  0.00,   0.13] ] ),
                (15, [ [ -323.80, 482.50,   0.10 ], [  -68.20,    2.90,  -1.20 ], [   0.00,   0.00, -14.00 ], [   0.00,  0.00,   0.12 ], [  -0.59, -13.99,  0.00 ], [   0.00,  0.00,   0.12] ] ),
                (16, [ [ -387.50, 485.40,  -0.20 ], [  -44.20,  -17.10,  -1.00 ], [   0.00,   0.00, -13.89 ], [   0.00,  0.00,   0.12 ], [   5.01, -12.95,  0.00 ], [   0.00,  0.00,   0.12] ] ),
                (17, [ [ -435.60, 433.50,   3.30 ], [    3.40, -109.50,   1.40 ], [   0.00,   0.00, -13.76 ], [   0.00,  0.00,   0.14 ], [  13.75,   0.43,  0.00 ], [   0.00,  0.00,   0.14] ] ),
                (18, [ [ -370.60, 376.30,  -1.10 ], [   66.90,  -29.20,  -0.90 ], [   0.00,   0.00, -13.60 ], [   0.00,  0.00,   0.12 ], [   5.44,  12.46,  0.00 ], [   0.00,  0.00,   0.12] ] ),
                (19, [ [ -313.00, 357.90,  -0.10 ], [   40.00,  -33.50,   9.60 ], [   0.00,   0.00, -13.50 ], [   0.00,  0.00,   0.10 ], [   8.67,  10.35,  0.00 ], [   0.00,  0.00,   0.10] ] ),
                (20, [ [ -259.20, 340.70,   2.10 ], [   48.90,    6.40,   1.40 ], [   0.00,   0.00, -13.40 ], [   0.00,  0.00,   0.09 ], [  -1.74,  13.29,  0.00 ], [   0.00,  0.00,   0.09] ] ),
                (21, [ [ -246.50, 380.30,  -0.80 ], [  -29.70,   33.60,  -0.70 ], [   0.00,   0.00, -13.31 ], [   0.00,  0.00,   0.09 ], [  -9.97,  -8.82,  0.00 ], [   0.00,  0.00,   0.09] ] ),
                (22, [ [ -297.30, 387.10,   0.60 ], [  -59.70,   12.60,  -0.00 ], [   0.00,   0.00, -13.22 ], [   0.00,  0.00,   0.09 ], [  -2.73, -12.94,  0.00 ], [   0.00,  0.00,   0.09] ] ),
                (23, [ [ -340.20, 415.60,  -1.00 ], [  -86.20,   28.90,  -2.90 ], [   0.00,   0.00, -13.13 ], [   0.00,  0.00,   0.10 ], [  -4.17, -12.45,  0.00 ], [   0.00,  0.00,   0.10] ] ),
                (24, [ [ -398.30, 443.10,  -0.10 ], [   10.60,   82.10,  -2.60 ], [   0.00,   0.00, -13.01 ], [   0.00,  0.00,   0.12 ], [ -12.90,   1.67,  0.00 ], [   0.00,  0.00,   0.12] ] ),
                (25, [ [ -329.80, 449.10,  -2.10 ], [   53.20,   14.00,  -0.50 ], [   0.00,   0.00, -12.88 ], [   0.00,  0.00,   0.14 ], [  -3.28,  12.46,  0.00 ], [   0.00,  0.00,   0.14] ] ),
                (26, [ [ -251.30, 425.90,  -0.30 ], [   43.90,  -19.30,   0.00 ], [   0.00,   0.00, -12.74 ], [   0.00,  0.00,   0.11 ], [   5.13,  11.66,  0.00 ], [   0.00,  0.00,   0.11] ] ),
                (27, [ [ -209.10, 390.60,   0.00 ], [   26.00,  -38.80,   0.90 ], [   0.00,   0.00, -12.65 ], [   0.00,  0.00,   0.08 ], [  10.51,   7.04,  0.00 ], [   0.00,  0.00,   0.08] ] ),
                (28, [ [ -207.80, 350.80,   1.40 ], [   -9.40,  -43.60,   1.80 ], [   0.00,   0.00, -12.57 ], [   0.00,  0.00,   0.09 ], [  12.29,  -2.65,  0.00 ], [   0.00,  0.00,   0.09] ] ),
                (29, [ [ -245.80, 299.40,   7.60 ], [  -70.30,  -36.00,   1.40 ], [   0.00,   0.00, -12.46 ], [   0.00,  0.00,   0.14 ], [   5.68, -11.09,  0.00 ], [   0.00,  0.00,   0.14] ] ),
                (30, [ [ -345.30, 304.10,   3.10 ], [ -100.20,   27.90,  -1.90 ], [   0.00,   0.00, -12.29 ], [   0.00,  0.00,   0.17 ], [  -3.30, -11.84,  0.00 ], [   0.00,  0.00,   0.17] ] ),
                (31, [ [ -418.40, 361.10,  -0.20 ], [  -57.80,   55.80,  -1.70 ], [   0.00,   0.00, -12.13 ], [   0.00,  0.00,   0.15 ], [  -8.42,  -8.73,  0.00 ], [   0.00,  0.00,   0.15] ] ),
                (32, [ [ -479.20, 415.60,   2.20 ], [   -8.80,   73.10,  -1.60 ], [   0.00,   0.00, -11.98 ], [   0.00,  0.00,   0.15 ], [ -11.89,  -1.43,  0.00 ], [   0.00,  0.00,   0.15] ] ),
                (33, [ [ -439.60, 495.70,  -2.10 ], [   61.10,   57.10,  -1.30 ], [   0.00,   0.00, -11.82 ], [   0.00,  0.00,   0.15 ], [  -8.07,   8.64,  0.00 ], [   0.00,  0.00,   0.15] ] ),
                (34, [ [ -361.60, 522.60,  -3.00 ], [   78.60,    9.90,   0.20 ], [   0.00,   0.00, -11.68 ], [   0.00,  0.00,   0.15 ], [  -1.46,  11.59,  0.00 ], [   0.00,  0.00,   0.15] ] ),
                (35, [ [ -270.10, 506.50,  -3.80 ], [  103.60,  -33.30,   1.00 ], [   0.00,   0.00, -11.52 ], [   0.00,  0.00,   0.19 ], [   3.53,  10.97,  0.00 ], [   0.00,  0.00,   0.19] ] ),
                (36, [ [ -148.90, 441.40,  -2.10 ], [   79.70,  -91.50,   2.80 ], [   0.00,   0.00, -11.28 ], [   0.00,  0.00,   0.23 ], [   8.51,   7.41,  0.00 ], [   0.00,  0.00,   0.23] ] ),
                (37, [ [ -130.90, 313.30,   4.00 ], [   -4.00, -107.20,   3.10 ], [   0.00,   0.00, -11.05 ], [   0.00,  0.00,   0.18 ], [  11.04,  -0.41,  0.00 ], [   0.00,  0.00,   0.18] ] ),
                (38, [ [ -183.90, 251.00,   3.80 ], [  -65.50,  -60.20,   3.60 ], [   0.00,   0.00, -10.90 ], [   0.00,  0.00,   0.16 ], [   7.38,  -8.03,  0.00 ], [   0.00,  0.00,   0.16] ] ),
                (39, [ [ -280.30, 213.00,   3.40 ], [ -165.10,  -18.60,   0.10 ], [   0.00,   0.00, -10.72 ], [   0.00,  0.00,   0.20 ], [   1.20, -10.65,  0.00 ], [   0.00,  0.00,   0.20] ] ),
                (40, [ [ -400.80, 247.50,   6.80 ], [ -127.10,   36.80,   1.30 ], [   0.00,   0.00, -10.51 ], [   0.00,  0.00,   0.23 ], [  -2.92, -10.10,  0.00 ], [   0.00,  0.00,   0.23] ] ),
                (41, [ [ -530.50, 290.70,   5.20 ], [  -89.00,   86.50,   0.30 ], [   0.00,   0.00, -10.27 ], [   0.00,  0.00,   0.21 ], [  -7.16,  -7.36,  0.00 ], [   0.00,  0.00,   0.21] ] ),
                (42, [ [ -568.80, 392.30,   6.90 ], [  -77.40,   67.70,  -5.50 ], [   0.00,   0.00, -10.08 ], [   0.00,  0.00,   0.23 ], [  -6.64,  -7.59,  0.00 ], [   0.00,  0.00,   0.23] ] ),
                (43, [ [ -511.20, 535.10,   2.50 ], [   86.20,  111.40,  -1.00 ], [   0.00,   0.00,  -9.80 ], [   0.00,  0.00,   0.25 ], [  -7.75,   6.00,  0.00 ], [   0.00,  0.00,   0.25] ] ),
                (44, [ [ -405.00, 601.70,   6.40 ], [  143.60,   52.20,   2.60 ], [   0.00,   0.00,  -9.58 ], [   1.49, -1.44,   0.82 ], [  -3.27,   9.00,  0.00 ], [   1.49, -1.44,   0.82] ] ),
                (45, [ [ -238.80, 615.90,  16.60 ], [   63.30,   -9.10,  19.10 ], [   3.45,  -3.33,  -7.95 ], [  -5.13,  0.21,   1.79 ], [   2.06,   8.63, -2.72 ], [  -5.13,  0.21,   1.79] ] ),
                (46, [ [ -146.20, 605.90,  36.50 ], [   49.30,   -9.90, -50.60 ], [  -6.72,  -1.05,  -6.07 ], [  -3.12,  3.94,   7.67 ], [   0.10,   8.96, -1.66 ], [  -3.12,  3.94,   7.67] ] ),
                (47, [ [ -218.40, 585.30,  -2.00 ], [ -124.00,    0.40, -37.50 ], [  -3.04,   4.50,   7.19 ], [   2.76,  2.30,   8.98 ], [   1.33,   7.80, -4.32 ], [   2.76,  2.30,   8.98] ] ),
                (48, [ [ -376.30, 579.60, -40.80 ], [ -189.20,  -50.70,  -8.80 ], [  -1.85,   1.22,   8.78 ], [   4.16,  0.49,  -2.60 ], [  -2.23,   8.62, -1.67 ], [   4.16,  0.49,  -2.60] ] ),
                (49, [ [ -557.90, 493.90, -24.90 ], [  -30.30,   24.10, 152.80 ], [   6.19,   6.65,   0.69 ], [   1.49, -2.48,  -8.73 ], [  -6.35,   6.14, -2.23 ], [   1.49, -2.48,  -8.73] ] ),
                (50, [ [ -484.80, 594.40,   0.70 ], [  132.70,   97.00,   3.50 ], [   3.25,  -1.18,  -8.47 ], [  -2.78, -5.06,  -5.12 ], [  -5.07,   7.03, -2.92 ], [  -2.78, -5.06,  -5.12] ] ),
                (51, [ [ -318.10, 641.90,  -8.50 ], [  166.70,   17.60,   5.50 ], [   0.67,  -2.88,  -8.72 ], [  -2.15, -0.82,  -0.11 ], [  -0.82,   8.69, -2.93 ], [  -2.15, -0.82,  -0.11] ] ),
                (52, [ [ -158.30, 634.70,  -1.90 ], [  176.50,  -14.00,  10.80 ], [  -1.08,  -2.89,  -8.71 ], [  -0.81,  0.27,  -0.12 ], [   0.87,   8.70, -2.99 ], [  -0.81,  0.27,  -0.12] ] ),
                (53, [ [   32.70, 611.70,  13.60 ], [  205.50,  -32.20,  20.00 ], [  -0.76,  -2.28,  -8.98 ], [   1.46,  0.96,  -0.42 ], [   1.62,   8.84, -2.38 ], [   1.46,  0.96,  -0.42] ] ) ] ),

            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-8',
                    'name': get_colon_term('right colon')[0],
                    'ontId': get_colon_term('right colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '9-46',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '47-52',
                    'name': get_colon_term('left colon')[0],
                    'ontId': get_colon_term('left colon')[1]
                }]
        }),
        'Human 1': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 8
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [ Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1, [ [   0.00,   0.00, 0.00 ], [ -50.70,  178.20, 0.00 ], [ -37.97,  -9.49, -18.98 ], [ -6.86, -11.39,  -2.36 ], [ -18.61,  -3.98, 39.12 ], [ -14.00,  -1.00, -12.00] ] ),
                (2, [ [ -47.40, 188.60, 0.00 ], [ -19.30,  177.10, 0.00 ], [ -35.79,  -6.51, -13.01 ], [ 11.23,  17.36,  14.31 ], [ -12.66,  -3.99, 36.28 ], [  -4.00,  19.00,  22.00] ] ),
                (3, [ [  -4.40, 396.50, 0.00 ], [ 206.00,   40.10, 0.00 ], [ -13.89,  27.78,  11.11 ], [ 13.54,  -1.87,  21.51 ], [  -6.05, -12.50, 29.93 ], [  -6.00,   0.00,  51.00] ] ),
                (4, [ [ 130.00, 384.10, 0.00 ], [ 130.80,  -40.50, 0.00 ], [  -5.35,   4.28,  31.06 ], [  5.83,  -8.41,   8.86 ], [ -15.28, -27.78,  2.51 ], [   0.00,   1.00,  24.00] ] ),
                (5, [ [ 279.40, 383.00, 0.00 ], [ 118.00,   48.70, 0.00 ], [  -2.51,  12.57,  27.65 ], [  9.30,   9.73, -10.83 ], [  12.83, -24.62, 12.58 ], [   5.00,  25.00, -20.00] ] ),
                (6, [ [ 443.90, 390.80, 0.00 ], [ 111.30,  -97.00, 0.00 ], [  14.07,  23.92,   8.44 ], [ 12.36,  -3.74, -23.50 ], [  -9.40,  -3.01, 27.28 ], [   1.00,  -6.00, -35.00] ] ),
                (7, [ [ 475.20, 168.00, 0.00 ], [  -0.80, -112.40, 0.00 ], [  20.78,   0.00, -20.78 ], [ -2.41, -19.32, -15.36 ], [  20.78,  -0.00, 20.78 ], [  15.00,  -1.00, -10.00] ] ),
                (8, [ [ 432.60, -32.30, 0.00 ], [ -90.50,  -59.00, 0.00 ], [  10.09, -15.13, -23.54 ], [ -9.58,  -7.07,  -2.37 ], [  13.01, -19.62, 18.19 ], [   8.00, -11.00, -13.00] ] ),
                (9, [ [ 272.40,   7.50, 0.00 ], [ -79.00,   47.40, 0.00 ], [   1.42, -15.65, -25.60 ], [ -7.76,   6.05,  -1.75 ], [  -5.22, -26.72, 12.68 ], [   4.00, -12.00, -12.00] ] ) ] ),

            'userAnnotationGroups': [
            {
                '_AnnotationGroup': True,
                'dimension': 1,
                'identifierRanges': '1-2',
                'name': get_colon_term('ascending colon')[0],
                'ontId': get_colon_term('ascending colon')[1]
            },
            {
                '_AnnotationGroup': True,
                'dimension': 1,
                'identifierRanges': '3-5',
                'name': get_colon_term('transverse colon')[0],
                'ontId': get_colon_term('transverse colon')[1]
            },
            {
                '_AnnotationGroup': True,
                'dimension': 1,
                'identifierRanges': '6-8',
                'name': get_colon_term('descending colon')[0],
                'ontId': get_colon_term('descending colon')[1]
            }]
        }),
        'Human 2': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 8
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [ Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1, [ [   0.00,    0.00,   0.00 ], [  -56.81,  105.14,  -38.05 ], [ -37.97,  -9.49, -18.98 ], [ -3.67, -11.77,  -1.56 ], [ -21.20,   5.92, 37.52 ], [ -14.00,  -1.00, -12.00 ] ] ),
                (2, [ [ -34.50,  114.00, -18.10 ], [   -9.51,  117.91,    3.65 ], [ -33.74,  -6.13, -12.27 ], [ 12.13,  18.49,  14.99 ], [ -12.58,  -4.62, 33.86 ], [  -4.00,  19.00,  22.00 ] ] ),
                (3, [ [ -19.10,  218.50,   5.50 ], [   79.23,   66.40,   77.49 ], [ -13.72,  27.44,  10.98 ], [ 14.61,   7.18,  21.72 ], [ -22.92, -17.43, 15.26 ], [  -6.00,   0.00,  51.00 ] ] ),
                (4, [ [  82.50,  189.10,  94.20 ], [  140.70,   -1.06,   48.14 ], [  -5.34,   4.27,  30.95 ], [  5.75,  -8.29,   9.07 ], [  11.54, -25.97, 14.03 ], [   0.00,   1.00,  24.00 ] ] ),
                (5, [ [ 226.60,  218.70,  85.70 ], [  164.08,  101.90,  -75.52 ], [  -2.53,  12.64,  27.81 ], [  8.00,   9.47,  -9.25 ], [  19.49, -20.43, 11.94 ], [   5.00,  25.00, -20.00 ] ] ),
                (6, [ [ 325.50,  381.70, -57.90 ], [  187.36, -116.61, -173.53 ], [  14.07,  23.92,   8.44 ], [ 12.05,  -5.07, -23.99 ], [   5.51, -10.97, 26.28 ], [   1.00,  -6.00, -35.00 ] ] ),
                (7, [ [ 354.00,  105.30, -24.40 ], [  -20.59, -269.54,   30.48 ], [  20.87,   0.00, -20.87 ], [ -2.92, -19.10, -14.61 ], [  20.81,   5.79, 20.11 ], [  15.00,  -1.00, -10.00 ] ] ),
                (8, [ [ 296.50, -121.20,  -0.60 ], [ -170.98, -102.19,  -18.39 ], [  10.15, -15.22, -23.67 ], [ -9.48,  -6.06,  -2.33 ], [  13.09, -19.73, 18.29 ], [   8.00, -11.00, -13.00 ] ] ),
                (9, [ [ 169.80,  -73.40, -33.50 ], [  -42.47,  101.91,  -24.43 ], [   1.43, -15.71, -25.71 ], [ -7.96,   5.07,  -1.75 ], [ -16.72, -21.84, 12.39 ], [   4.00, -12.00, -12.00 ] ] ) ] ),

            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-2',
                    'name': get_colon_term('ascending colon')[0],
                    'ontId': get_colon_term('ascending colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '3-5',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '6-8',
                    'name': get_colon_term('descending colon')[0],
                    'ontId': get_colon_term('descending colon')[1]
                }]
        }),
        'Human 3': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 48
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1,  [ [ -87.21, -111.06,  890.54 ], [  -4.75,   0.41,  12.39 ], [  21.05,  -2.83,  7.13 ], [  2.46, -0.39, -2.95 ], [  2.79, 22.12,  0.41 ], [  1.83,  0.46, -4.31 ] ] ),
                (2,  [ [ -89.99, -110.57,  902.65 ], [  -2.05,   0.53,  12.72 ], [  22.40,  -2.97,  3.74 ], [  0.07,  0.14, -2.81 ], [  3.10, 22.66, -0.58 ], [  0.43,  0.10, -3.97 ] ] ),
                (3,  [ [ -91.25, -110.01,  915.92 ], [  -0.55,   0.53,  13.39 ], [  21.27,  -2.54,  0.97 ], [ -1.46,  0.55, -1.92 ], [  2.57, 21.27, -0.82 ], [ -1.42,  0.54, -2.13 ] ] ),
                (4,  [ [ -91.08, -109.52,  929.39 ], [   0.13,   0.65,  13.47 ], [  19.48,  -1.87, -0.09 ], [ -0.53,  0.22, -1.06 ], [  1.86, 19.46, -1.05 ], [ -0.53,  0.21, -1.17 ] ] ),
                (5,  [ [ -91.00, -108.72,  942.85 ], [   0.83,   0.67,  13.37 ], [  20.22,  -2.11, -1.15 ], [ -0.88,  0.27, -1.85 ], [  2.05, 20.22, -1.27 ], [ -0.92,  0.27, -2.03 ] ] ),
                (6,  [ [ -89.43, -108.18,  956.08 ], [   2.81,   0.32,  13.11 ], [  17.75,  -1.34, -3.78 ], [ -1.96,  0.62, -1.35 ], [  1.22, 18.09, -0.76 ], [ -2.00,  0.62, -1.49 ] ] ),
                (7,  [ [ -85.39, -108.09,  968.92 ], [   3.10,   0.44,  13.08 ], [  16.31,  -0.87, -3.84 ], [ -0.76,  0.32,  0.66 ], [  0.72, 16.68, -0.80 ], [ -0.73,  0.31,  0.71 ] ] ),
                (8,  [ [ -83.24, -107.30,  982.14 ], [   2.05,   0.85,  13.23 ], [  16.22,  -0.70, -2.47 ], [ -0.24,  0.09,  0.57 ], [  0.54, 16.36, -1.26 ], [ -0.21,  0.08,  0.63 ] ] ),
                (9,  [ [ -81.29, -106.40,  995.37 ], [   2.32,   1.34,  13.29 ], [  15.83,  -0.69, -2.70 ], [  0.14,  0.24, -0.11 ], [  0.41, 15.95, -1.86 ], [  0.05,  0.31, -0.57 ] ] ),
                (10, [ [ -78.60, -104.62, 1008.65 ], [   2.14,  -1.50,  13.27 ], [  16.51,  -0.22, -2.69 ], [ -0.91,  1.85, -0.22 ], [  0.74, 16.43,  3.01 ], [ -0.95,  1.98, -0.15 ] ] ),
                (11, [ [ -77.17, -109.18, 1020.82 ], [   4.07,  -9.07,   9.68 ], [  14.05,   2.96, -3.13 ], [ -2.15,  4.06,  0.80 ], [ -0.01, 10.40, 10.89 ], [ -1.97,  3.89,  1.43 ] ] ),
                (12, [ [ -71.19, -121.03, 1025.58 ], [   7.48, -11.42,   0.49 ], [  12.23,   7.97, -0.98 ], [ -1.27,  2.69,  3.17 ], [  0.00,  1.72, 15.29 ], [ -1.27,  2.72,  3.23 ] ] ),
                (13, [ [ -63.35, -130.56, 1022.18 ], [   7.99,  -9.44,  -3.40 ], [  11.46,   8.59,  3.10 ], [ -0.81,  0.70,  2.42 ], [  0.00, -4.74, 14.55 ], [ -0.91,  0.85,  2.43 ] ] ),
                (14, [ [ -55.27, -139.84, 1018.80 ], [   8.88,  -8.59,  -3.55 ], [  10.61,   9.36,  3.87 ], [ -1.07,  0.87,  0.81 ], [  0.00, -5.35, 14.33 ], [ -1.09,  0.87,  0.81 ] ] ),
                (15, [ [ -45.66, -147.66, 1015.10 ], [   9.90,  -7.39,  -3.36 ], [   9.31,  10.33,  4.71 ], [ -1.63,  1.38,  0.03 ], [ -0.01, -5.79, 14.11 ], [ -1.62,  1.36,  0.02 ] ] ),
                (16, [ [ -35.52, -154.60, 1012.08 ], [  11.08,  -6.08,  -1.98 ], [   7.36,  12.12,  3.95 ], [ -1.99,  1.15,  0.25 ], [ -0.01, -4.33, 14.74 ], [ -1.97,  1.13,  0.24 ] ] ),
                (17, [ [ -23.71, -159.68, 1011.20 ], [  11.97,  -4.32,  -1.79 ], [   5.34,  12.62,  5.23 ], [ -1.74, -0.44,  2.51 ], [  0.01, -5.34, 14.28 ], [ -1.79, -0.44,  2.52 ] ] ),
                (18, [ [ -11.72, -163.20, 1008.53 ], [  12.26,  -2.58,  -2.06 ], [   3.87,  11.25,  8.97 ], [ -2.16, -0.11,  1.44 ], [ -0.01, -8.84, 12.28 ], [ -2.14, -0.12,  1.43 ] ] ),
                (19, [ [   0.67, -164.82, 1007.10 ], [  12.71,  -0.73,  -0.48 ], [   1.02,  12.38,  8.14 ], [ -2.67,  0.76, -0.81 ], [ -0.01, -7.75, 13.06 ], [ -2.58,  0.75, -0.81 ] ] ),
                (20, [ [  13.53, -164.63, 1007.59 ], [  12.75,   0.98,   0.85 ], [  -1.47,  12.77,  7.36 ], [ -1.75,  0.42, -1.04 ], [ -0.28, -7.05, 13.49 ], [ -1.71,  0.43, -1.04 ] ] ),
                (21, [ [  26.09, -162.88, 1008.79 ], [  12.63,   1.81,   1.23 ], [  -2.49,  13.21,  6.05 ], [ -2.26, -0.16, -1.14 ], [ -0.39, -5.90, 14.09 ], [ -2.23, -0.14, -1.14 ] ] ),
                (22, [ [  38.73, -161.01, 1010.04 ], [  11.76,   4.53,   2.77 ], [  -6.00,  12.45,  5.07 ], [ -3.08, -0.42, -3.34 ], [ -0.85, -5.62, 14.19 ], [ -3.12, -0.44, -3.35 ] ] ),
                (23, [ [  48.79, -154.12, 1014.15 ], [   9.86,   7.10,   4.16 ], [  -8.65,  12.37, -0.62 ], [ -1.93, -0.52, -2.09 ], [ -4.15, -2.21, 15.11 ], [ -1.97, -0.53, -2.09 ] ] ),
                (24, [ [  58.41, -146.84, 1018.34 ], [   9.34,   7.75,   4.35 ], [  -9.86,  11.41,  0.85 ], [ -1.82, -1.10,  1.29 ], [ -3.19, -3.77, 15.02 ], [ -1.80, -1.06,  1.30 ] ] ),
                (25, [ [  67.43, -138.64, 1022.84 ], [   7.99,   8.77,   4.67 ], [ -12.30,  10.17,  1.95 ], [ -2.50, -1.97, -0.34 ], [ -2.29, -5.48, 15.73 ], [ -2.59, -2.16, -0.41 ] ] ),
                (26, [ [  74.34, -129.42, 1027.61 ], [   5.08,   9.97,   3.83 ], [ -14.87,   7.50,  0.22 ], [  0.03, -2.06,  0.33 ], [ -2.33, -5.10, 16.50 ], [ -1.53, -1.17,  0.49 ] ] ),
                (27, [ [  77.56, -119.25, 1030.43 ], [   5.04,  10.78,  -0.51 ], [ -12.53,   5.98,  2.38 ], [  0.10,  1.23, -1.77 ], [  3.79,  1.01, 17.93 ], [  0.40,  1.97, -2.97 ] ] ),
                (28, [ [  83.86, -109.53, 1026.21 ], [   6.93,   6.51,  -9.45 ], [ -14.97,  10.33, -3.86 ], [ -2.81,  1.30, -3.77 ], [  6.42, 13.14, 12.02 ], [ -1.24,  0.30, -4.60 ] ] ),
                (29, [ [  89.55, -108.42, 1013.56 ], [   3.91,   0.86, -13.16 ], [ -18.20,   8.23, -4.87 ], [ -1.85, -1.68,  0.87 ], [  7.52, 18.71,  3.84 ], [ -2.14, -1.95,  1.17 ] ] ),
                (30, [ [  91.64, -107.83, 1000.37 ], [   1.78,   0.55, -13.27 ], [ -18.73,   6.95, -2.22 ], [ -0.80, -1.12,  1.76 ], [  6.79, 18.82,  1.85 ], [ -0.86, -1.12,  1.92 ] ] ),
                (31, [ [  93.10, -107.33,  987.04 ], [   1.01,   0.30, -13.42 ], [ -19.80,   6.00, -1.36 ], [ -0.56, -0.65,  1.35 ], [  5.95, 19.84,  0.98 ], [ -0.58, -0.66,  1.49 ] ] ),
                (32, [ [  93.66, -107.24,  973.55 ], [  -0.02,   1.05, -13.41 ], [ -19.85,   5.64,  0.47 ], [  0.93, -2.25,  1.33 ], [  5.66, 19.79,  1.71 ], [  0.92, -2.27,  1.47 ] ] ),
                (33, [ [  93.06, -105.25,  960.32 ], [  -0.77,   2.25, -13.20 ], [ -17.96,   1.50,  1.30 ], [  0.92, -2.34,  0.15 ], [  1.70, 17.72,  3.23 ], [  0.93, -2.36,  0.17 ] ] ),
                (34, [ [  92.12, -102.74,  947.16 ], [  -0.42,   2.82, -13.16 ], [ -18.00,   0.97,  0.78 ], [  0.74, -0.11, -0.85 ], [  1.11, 17.57,  4.15 ], [  0.73, -0.09, -0.94 ] ] ),
                (35, [ [  92.23,  -99.62,  934.04 ], [   0.47,   1.78, -13.24 ], [ -16.47,   1.28, -0.41 ], [  1.01,  0.01, -1.14 ], [  1.22, 16.31,  2.49 ], [  1.03,  0.02, -1.27 ] ] ),
                (36, [ [  93.04,  -99.17,  920.86 ], [   1.21,  -0.74, -13.33 ], [ -15.97,   1.00, -1.51 ], [  0.28, -0.61,  0.54 ], [  1.07, 16.00, -0.87 ], [  0.28, -0.69, -0.02 ] ] ),
                (37, [ [  94.65, -101.12,  907.56 ], [  -0.61,  -1.62, -13.47 ], [ -15.91,   0.05,  0.71 ], [ -0.21, -1.55,  3.28 ], [ -0.04, 15.82, -2.00 ], [ -0.49, -1.71,  2.07 ] ] ),
                (38, [ [  91.84, -102.35,  894.35 ], [  -3.47,  -1.64, -11.95 ], [ -16.38,  -2.10,  5.05 ], [  0.91, -2.20,  3.65 ], [ -2.64, 16.98, -1.59 ], [ -0.73, -2.58,  3.87 ] ] ),
                (39, [ [  87.88, -104.31,  883.75 ], [  -5.21,  -2.13, -10.32 ], [ -14.30,  -4.34,  8.12 ], [  2.73, -2.87,  2.50 ], [ -5.97, 18.48, -0.99 ], [  0.46, -3.19,  4.53 ] ] ),
                (40, [ [  81.45, -106.59,  873.90 ], [  -6.99,  -2.14,  -9.27 ], [ -10.90,  -7.86, 10.04 ], [  0.33, -3.15,  2.69 ], [ -9.26, 19.09,  1.96 ], [  2.23, -2.67,  3.37 ] ] ),
                (41, [ [  73.98, -108.57,  865.27 ], [  -8.31,  -0.22,  -8.52 ], [ -13.53, -10.65, 13.47 ], [ -1.28, -1.76,  1.51 ], [ -8.01, 18.53,  7.34 ], [  3.18, -1.46,  2.05 ] ] ),
                (42, [ [  65.06, -106.92,  857.15 ], [  -8.97,   1.81,  -7.69 ], [ -13.39, -11.32, 12.96 ], [  1.28,  0.43, -0.09 ], [ -4.01, 17.55, 11.89 ], [  2.48,  0.16,  0.52 ] ] ),
                (43, [ [  56.07, -104.97,  849.91 ], [ -10.00,   1.92,  -6.88 ], [ -11.02,  -9.85, 13.27 ], [  2.49,  5.00,  1.79 ], [ -2.03, 16.12, 10.97 ], [  2.28,  4.69,  1.10 ] ] ),
                (44, [ [  45.11, -103.12,  843.53 ], [ -11.00,   4.81,  -5.26 ], [  -8.40,  -0.98, 16.68 ], [  2.56,  4.66,  2.88 ], [  6.12, 16.27,  4.28 ], [  2.42,  4.35,  0.87 ] ] ),
                (45, [ [  34.86,  -95.59,  839.81 ], [  -9.59,   8.68,  -2.67 ], [  -5.91,  -0.68, 19.02 ], [  0.94, -0.41,  0.56 ], [ 11.03, 11.58,  3.43 ], [  0.00,  0.16, -0.54 ] ] ),
                (46, [ [  26.14,  -85.99,  838.23 ], [  -8.26,  10.16,  -2.00 ], [  -6.49,  -1.78, 17.83 ], [ -0.32, -0.12, -0.85 ], [ 12.91,  9.40,  5.55 ], [ -1.05, -0.47,  0.41 ] ] ),
                (47, [ [  18.39,  -75.31,  835.82 ], [  -8.42,  10.08,  -2.66 ], [  -6.54,  -0.88, 17.32 ], [ -0.56,  0.61,  0.80 ], [ 14.15,  8.98,  5.99 ], [ -0.36, -0.28,  1.59 ] ] ),
                (48, [ [   9.35,  -65.88,  832.92 ], [ -11.90,  13.24,  -4.28 ], [  -7.61,  -0.56, 19.43 ], [ -0.85,  0.48,  1.28 ], [ 15.89, 10.29,  6.56 ], [ -1.39, -0.78,  1.04 ] ] ),
                (49, [ [  -5.12,  -48.68,  827.11 ], [ -17.03,  21.16,  -7.34 ], [  -8.07,   0.18, 19.25 ], [ -0.08,  1.01, -1.64 ], [ 15.89, 10.29,  6.56 ], [ -1.39, -0.78,  1.04 ] ] ) ] ),

            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-48',
                    'name': get_colon_term('colon')[0],
                    'ontId': get_colon_term('colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-10',
                    'name': get_colon_term('ascending colon')[0],
                    'ontId': get_colon_term('ascending colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '11-26',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '27-48',
                    'name': get_colon_term('descending colon')[0],
                    'ontId': get_colon_term('descending colon')[1]
                }]
        }),
        'Mouse 1': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 7
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2,  Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1, [ [   0.00,   0.00,  0.00 ], [  6.00, 12.00,  -2.00 ], [ 0.67,  0.33,  0.67 ], [ -0.03,  0.95, -0.68 ], [  0.86,  0.04, -0.51 ], [ 6.00, 0.00, 3.00 ] ] ),
                (2, [ [  -2.00,  11.00, -3.00 ], [ -8.00,  4.00,   9.00 ], [ 0.64,  0.64,  0.32 ], [ -0.03, -0.34, -0.02 ], [ -0.37,  0.66, -0.58 ], [ 0.00, 1.00, 2.00 ] ] ),
                (3, [ [  -3.00,   2.00,  3.00 ], [ -4.00, -8.00,   0.00 ], [ 0.61, -0.30,  0.61 ], [ -0.15, -0.65,  0.01 ], [ -0.55,  0.27,  0.68 ], [ 1.00, 0.00, 2.00 ] ] ),
                (4, [ [ -11.00,  -3.00, -4.00 ], [ -8.00, -3.00,  -7.00 ], [ 0.33, -0.67,  0.33 ], [ -0.18, -0.18, -0.31 ], [ -0.32,  0.10,  0.75 ], [ 0.00, 0.00, 0.50 ] ] ),
                (5, [ [ -16.00,  -4.00,  0.00 ], [  4.00, -3.00,  14.00 ], [ 0.23, -0.70,  0.00 ], [ -0.16, -0.02, -0.19 ], [  0.71,  0.18,  0.05 ], [ 0.00, 0.00, 0.50 ] ] ),
                (6, [ [  -7.00,  -8.00,  0.00 ], [  5.00, -1.00, -14.00 ], [ 0.00, -0.70,  0.00 ], [  0.03,  0.04, -0.12 ], [ -0.64, -0.00, -0.28 ], [ 0.00, 0.00, 0.50 ] ] ),
                (7, [ [  -1.00,  -6.00, -1.00 ], [  2.00, -2.00,   9.00 ], [ 0.21, -0.63, -0.21 ], [  0.12,  0.11, -0.23 ], [  0.64,  0.25, -0.11 ], [ 0.00, 0.00, 0.50 ] ] ),
                (8, [ [  -2.00, -14.00,  5.00 ], [ -2.00, -4.00,   2.00 ], [ 0.23, -0.47, -0.47 ], [ -0.08,  0.22, -0.28 ], [  0.53, -0.17,  0.42 ], [ 0.00, 0.00, 0.50 ] ] ) ] ),

            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-2',
                    'name': get_colon_term('right colon')[0],
                    'ontId': get_colon_term('right colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '3-4',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '5-7',
                    'name': get_colon_term('left colon')[0],
                    'ontId': get_colon_term('left colon')[1]
                }]
        }),
        'Mouse 2': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 4
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2,  Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1, [ [   0.00,  0.00,   0.00 ], [ 0.00,  0.00,  13.00 ], [ 0.00, -1.00, 0.00 ], [ 0.00,  0.03, 0.00 ], [  1.00, -0.00,  0.00 ], [ 0.00, 0.00, 0.50 ] ] ),
                (2, [ [   0.00,  0.00,  13.00 ], [ 0.00,  2.00,  28.00 ], [ 0.00, -0.96, 0.00 ], [ 0.00,  0.05, 0.00 ], [  0.95, -0.00, -0.07 ], [ 0.00, 0.00, 0.50 ] ] ),
                (3, [ [ -14.00, -2.00,  13.00 ], [ 0.00, -3.00, -19.00 ], [ 0.00, -0.88, 0.00 ], [ 0.00,  0.13, 0.00 ], [ -0.87, -0.02, -0.14 ], [ 0.00, 0.00, 0.50 ] ] ),
                (4, [ [ -14.00, -1.00, -10.00 ], [ 1.00,  1.00, -17.00 ], [ 0.00, -0.70, 0.00 ], [ 0.00,  0.08, 0.00 ], [ -0.70, -0.00, -0.00 ], [ 0.00, 0.00, 0.50 ] ] ),
                (5, [ [ -14.00,  0.00, -28.00 ], [ 0.00,  0.00, -11.00 ], [ 0.00, -0.70, 0.00 ], [ 0.00, -0.08, 0.00 ], [ -0.70, -0.00,  0.00 ], [ 0.00, 0.00, 0.50 ] ] ) ] ),
                
            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-2',
                    'name': get_colon_term('right colon')[0],
                    'ontId': get_colon_term('right colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '3',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '4',
                    'name': get_colon_term('left colon')[0],
                    'ontId': get_colon_term('left colon')[1]
                }]
        }),
        'Pig 1': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'Coordinate dimensions': 3,
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Length': 1.0,
                'Number of elements': 39
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2,  Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1,  [ [   -7.20,   83.30,  -20.70 ], [  -65.20,   -8.10,   7.60 ], [  0.00,  0.00, 36.00 ], [  0.00,  0.00, -0.50 ], [  -8.49,  34.98,  0.48 ], [  0.00,  0.00, -0.50 ] ] ),
                (2,  [ [  -68.50,   52.80,   -9.60 ], [  -40.10,  -36.10,  10.70 ], [  0.00,  0.00, 35.44 ], [  0.00,  0.00, -0.63 ], [ -28.28,  21.31,  1.34 ], [  0.00,  0.00, -0.63 ] ] ),
                (3,  [ [  -97.40,  -26.30,    5.70 ], [   18.00,  -93.20,  13.70 ], [  0.00,  0.00, 34.73 ], [  0.00,  0.00, -0.67 ], [ -32.82, -11.34,  0.71 ], [  0.00,  0.00, -0.67 ] ] ),
                (4,  [ [  -56.80,  -90.50,   14.10 ], [   65.50,  -41.40,   7.30 ], [  0.00,  0.00, 34.10 ], [  0.00,  0.00, -0.74 ], [ -15.44, -30.39,  0.30 ], [  0.00,  0.00, -0.74 ] ] ),
                (5,  [ [   48.90, -100.80,   24.00 ], [  112.20,   40.10,  19.00 ], [  0.00,  0.00, 33.22 ], [  0.00,  0.00, -0.91 ], [  15.90, -29.15,  0.82 ], [  0.00,  0.00, -0.91 ] ] ),
                (6,  [ [  114.80,  -12.60,   38.70 ], [    8.20,   96.10,  14.20 ], [  0.00,  0.00, 32.28 ], [  0.00,  0.00, -0.93 ], [  32.22,   1.92,  0.68 ], [  0.00,  0.00, -0.93 ] ] ),
                (7,  [ [   60.30,   83.50,   43.70 ], [ -108.70,   54.10,  22.40 ], [  0.00,  0.00, 31.35 ], [  0.00,  0.00, -0.89 ], [   8.73,  30.09,  1.03 ], [  0.00,  0.00, -0.89 ] ] ),
                (8,  [ [  -41.20,   90.70,   56.30 ], [  -89.00,  -32.40,  14.40 ], [  0.00,  0.00, 30.50 ], [  0.00,  0.00, -0.93 ], [ -14.57,  26.79,  0.69 ], [  0.00,  0.00, -0.93 ] ] ),
                (9,  [ [ -107.90,   -9.70,   76.60 ], [   11.10,  -94.40,  11.30 ], [  0.00,  0.00, 29.47 ], [  0.00,  0.00, -0.91 ], [ -28.66,  -6.85,  0.41 ], [  0.00,  0.00, -0.91 ] ] ),
                (10, [ [  -57.30,  -91.90,   81.30 ], [   71.20,  -31.20,   5.70 ], [  0.00,  0.00, 28.65 ], [  0.00,  0.00, -0.86 ], [  -9.56, -27.01,  0.15 ], [  0.00,  0.00, -0.86 ] ] ),
                (11, [ [   51.20,  -89.40,   97.20 ], [   99.10,   55.40,  12.90 ], [  0.00,  0.00, 27.75 ], [  0.00,  0.00, -0.90 ], [  16.17, -22.55,  0.35 ], [  0.00,  0.00, -0.90 ] ] ),
                (12, [ [   91.60,    9.30,  103.60 ], [    4.70,   51.20,   3.40 ], [  0.00,  0.00, 26.86 ], [  0.00,  0.00, -0.90 ], [  26.85,  -0.69,  0.12 ], [  0.00,  0.00, -0.90 ] ] ),
                (13, [ [   61.60,  111.80,  109.60 ], [  -85.20,   46.10,   2.60 ], [  0.00,  0.00, 25.95 ], [  0.00,  0.00, -0.95 ], [  11.73,  23.14,  0.02 ], [  0.00,  0.00, -0.95 ] ] ),
                (14, [ [  -54.60,   91.90,  129.40 ], [  -92.70,  -55.00,  14.50 ], [  0.00,  0.00, 24.95 ], [  0.00,  0.00, -0.94 ], [ -15.45,  19.58,  0.44 ], [  0.00,  0.00, -0.94 ] ] ),
                (15, [ [ -109.00,    5.60,  156.90 ], [   23.60, -108.20,  27.70 ], [  0.00,  0.00, 24.05 ], [  0.00,  0.00, -0.80 ], [ -21.59, -10.51,  1.42 ], [  0.00,  0.00, -0.80 ] ] ),
                (16, [ [  -59.10,  -62.50,  170.80 ], [   74.00,  -20.10,  14.40 ], [  0.00,  0.00, 23.34 ], [  0.00,  0.00, -0.70 ], [  -1.93, -23.24,  0.79 ], [  0.00,  0.00, -0.70 ] ] ),
                (17, [ [   23.50,  -53.20,  179.70 ], [   84.60,   47.00,   6.90 ], [  0.00,  0.00, 22.65 ], [  0.00,  0.00, -0.73 ], [  12.38, -18.97,  0.11 ], [  0.00,  0.00, -0.73 ] ] ),
                (18, [ [   62.30,   30.10,  187.50 ], [  -12.80,   58.00,   0.80 ], [  0.00,  0.00, 21.88 ], [  0.00,  0.00, -0.55 ], [  21.30,   5.00,  0.00 ], [  0.00,  0.00, -0.55 ] ] ),
                (19, [ [   22.40,   45.20,  181.10 ], [  -23.60,  -34.50,  -7.40 ], [  0.00,  0.00, 21.44 ], [  0.00,  0.00, -0.41 ], [ -15.35,  14.96,  0.65 ], [  0.00,  0.00, -0.41 ] ] ),
                (20, [ [   -1.90,    4.90,  180.50 ], [  -41.30,  -30.90,   7.50 ], [  0.00,  0.00, 21.06 ], [  0.00,  0.00, -0.39 ], [ -14.89,  14.89,  0.44 ], [  0.00,  0.00, -0.39 ] ] ),
                (21, [ [  -45.10,  -12.60,  194.40 ], [  -40.50,   -4.60,   6.90 ], [  0.00,  0.00, 20.67 ], [  0.00,  0.00, -0.30 ], [  -5.68,  19.87,  0.58 ], [  0.00,  0.00, -0.30 ] ] ),
                (22, [ [  -71.70,   -2.20,  197.20 ], [  -25.20,   35.80,  -6.80 ], [  0.00,  0.00, 20.42 ], [  0.00,  0.00, -0.30 ], [  18.28,   9.08,  0.48 ], [  0.00,  0.00, -0.30 ] ] ),
                (23, [ [  -65.80,   42.10,  182.30 ], [   26.60,   37.60, -15.60 ], [  0.00,  0.00, 20.03 ], [  0.00,  0.00, -0.46 ], [  11.97, -15.92,  2.06 ], [  0.00,  0.00, -0.46 ] ] ),
                (24, [ [  -14.10,   81.20,  163.50 ], [   41.00,   10.30,  -9.50 ], [  0.00,  0.00, 19.48 ], [  0.00,  0.00, -0.59 ], [   0.59, -19.44,  0.94 ], [  0.00,  0.00, -0.59 ] ] ),
                (25, [ [   61.70,   86.10,  156.40 ], [   77.90,  -40.70,   8.90 ], [  0.00,  0.00, 18.85 ], [  0.00,  0.00, -0.62 ], [  -7.01, -17.50,  0.19 ], [  0.00,  0.00, -0.62 ] ] ),
                (26, [ [   92.90,   20.50,  150.30 ], [    0.00,  -73.30,  -5.20 ], [  0.00,  0.00, 18.23 ], [  0.00,  0.00, -0.69 ], [ -18.19,   1.29,  0.09 ], [  0.00,  0.00, -0.69 ] ] ),
                (27, [ [   48.90,  -65.00,  142.80 ], [  -82.80,  -80.00,  -1.90 ], [  0.00,  0.00, 17.44 ], [  0.00,  0.00, -0.84 ], [ -11.91,  12.74,  0.00 ], [  0.00,  0.00, -0.84 ] ] ),
                (28, [ [  -54.30,  -90.80,  134.00 ], [  -60.10,   26.40,  -8.20 ], [  0.00,  0.00, 16.54 ], [  0.00,  0.00, -0.92 ], [   8.46,  14.21,  0.25 ], [  0.00,  0.00, -0.92 ] ] ),
                (29, [ [  -89.90,   11.20,  115.00 ], [   34.90,  125.10, -27.90 ], [  0.00,  0.00, 15.59 ], [  0.00,  0.00, -0.88 ], [  13.83,  -7.18,  0.69 ], [  0.00,  0.00, -0.88 ] ] ),
                (30, [ [  -17.40,   74.20,   91.10 ], [   78.80,   19.10, -15.40 ], [  0.00,  0.00, 14.77 ], [  0.00,  0.00, -0.67 ], [   0.79, -14.74,  0.51 ], [  0.00,  0.00, -0.67 ] ] ),
                (31, [ [   43.40,   50.20,   73.30 ], [   30.20,  -36.00,  -9.90 ], [  0.00,  0.00, 14.20 ], [  0.00,  0.00, -0.52 ], [ -12.48,  -6.73,  0.60 ], [  0.00,  0.00, -0.52 ] ] ),
                (32, [ [   62.40,   -5.10,   63.50 ], [   10.90,  -54.20,  -2.70 ], [  0.00,  0.00, 13.72 ], [  0.00,  0.00, -0.48 ], [ -13.56,  -2.05,  0.03 ], [  0.00,  0.00, -0.48 ] ] ),
                (33, [ [   32.70,  -51.70,   56.10 ], [  -38.60,  -29.80,  -8.10 ], [  0.00,  0.00, 13.24 ], [  1.37,  2.33, -1.30 ], [  -6.29,  11.65,  0.36 ], [  1.59,  2.70, -1.41 ] ] ),
                (34, [ [  -38.10,  -28.60,   46.80 ], [  -62.50,   82.60, -19.20 ], [  3.28,  5.57, 10.81 ], [  1.59, -2.89, -2.75 ], [   9.22,   6.17, -5.96 ], [  1.85, -3.40, -3.08 ] ] ),
                (35, [ [    5.70,   40.40,   22.40 ], [  144.80,   18.60, -20.50 ], [  2.68, -8.27,  7.65 ], [ -4.75, -6.37, -4.66 ], [   0.33,  -7.82, -8.54 ], [ -5.62, -7.55, -5.37 ] ] ),
                (36, [ [   53.00,  -14.70,   -4.10 ], [   -6.00,  -25.70, -46.70 ], [ -5.67, -8.15,  1.68 ], [ -3.78,  1.55, -1.86 ], [  -8.24,   3.65, -4.49 ], [ -4.60,  1.69, -2.12 ] ] ),
                (37, [ [   24.80,   -0.40,  -48.80 ], [  -13.40,   23.90, -30.60 ], [ -6.57, -5.70,  2.41 ], [  4.46,  2.06,  2.31 ], [  -1.87,   3.95,  7.90 ], [  5.41,  2.39,  2.88 ] ] ),
                (38, [ [  -20.90,   15.30,  -77.90 ], [  -51.20,  -30.60,  21.10 ], [  3.74, -4.06,  6.43 ], [ -1.12,  2.71, -1.85 ], [  -2.63,   5.93,  5.46 ], [ -1.44,  3.31, -2.27 ] ] ),
                (39, [ [  -47.60,   33.90, -112.20 ], [   32.60,   30.70, -27.80 ], [ -8.36, -0.32, -1.04 ], [ -7.25,  2.76, -6.03 ], [  -3.75,   2.25,  7.20 ], [ -8.97,  3.42, -7.45 ] ] ),
                (40, [ [   19.60,   96.00, -167.50 ], [   19.90,   19.10, -18.40 ], [ -6.99,  0.72, -4.50 ], [  9.97, -0.68, -0.89 ], [  -2.97,   5.84,  5.17 ], [ 12.30, -0.82, -1.11 ] ] ) ] ),
                
            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-34',
                    'name': get_colon_term('spiral colon')[0],
                    'ontId': get_colon_term('spiral colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '35-36',
                    'name': get_colon_term('transverse colon')[0],
                    'ontId': get_colon_term('transverse colon')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '37-39',
                    'name': get_colon_term('descending colon')[0],
                    'ontId': get_colon_term('descending colon')[1]
                }]
        }),
    }

    @staticmethod
    def getName():
        return '3D Colon 1'

    @staticmethod
    def getParameterSetNames():
        return [
            'Default',
            'Cattle 1',
            'Human 1',
            'Human 2',
            'Human 3',
            'Mouse 1',
            'Mouse 2',
            'Pig 1']

    @classmethod
    def getDefaultOptions(cls, parameterSetName='Default'):
        if 'Human 2' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Human 2']
        elif 'Human 3' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Human 3']
        elif 'Cattle 1' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Cattle 1']
        elif 'Mouse 1' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Mouse 1']
        elif 'Mouse 2' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Mouse 2']
        elif 'Pig 1' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Pig 1']
        else:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Human 1']

        if 'Human 3' in parameterSetName:
            segmentProfileOption = ScaffoldPackage(MeshType_3d_colonsegment1, defaultParameterSetName='Human 2')
        elif 'Cattle' in parameterSetName:
            segmentProfileOption = ScaffoldPackage(MeshType_3d_colonsegment1, defaultParameterSetName='Cattle 1')
        elif 'Mouse' in parameterSetName:
            segmentProfileOption = ScaffoldPackage(MeshType_3d_colonsegment1, defaultParameterSetName='Mouse 1')
        elif 'Pig' in parameterSetName:
            segmentProfileOption = ScaffoldPackage(MeshType_3d_colonsegment1, defaultParameterSetName='Pig 1')
        else:
            segmentProfileOption = ScaffoldPackage(MeshType_3d_colonsegment1, defaultParameterSetName='Human 1')
        options = {
            'Central path': copy.deepcopy(centralPathOption),
            'Segment profile': segmentProfileOption,
            'Number of segments': 30,
            'Start phase': 0.0,
            'Proximal tenia coli width': 10.0,
            'Proximal-transverse tenia coli width': 10.0,
            'Transverse-distal tenia coli width': 10.0,
            'Distal tenia coli width': 10.0,
            'Refine': False,
            'Refine number of elements around': 1,
            'Refine number of elements along': 1,
            'Refine number of elements through wall': 1
        }
        if 'Cattle 1' in parameterSetName:
            options['Number of segments'] = 40
            options['Proximal tenia coli width'] = 12.0
            options['Proximal-transverse tenia coli width'] = 6.0
            options['Transverse-distal tenia coli width'] = 3.0
            options['Distal tenia coli width'] = 3.0
        elif 'Human 3' in parameterSetName:
            options['Number of segments'] = 33
        elif 'Mouse' in parameterSetName:
            options['Number of segments'] = 10
            options['Proximal tenia coli width'] = 0.5
            options['Proximal-transverse tenia coli width'] = 0.5
            options['Transverse-distal tenia coli width'] = 0.5
            options['Distal tenia coli width'] = 0.5
        elif 'Pig 1' in parameterSetName:
            options['Number of segments'] = 120
            options['Proximal tenia coli width'] = 5.0
            options['Proximal-transverse tenia coli width'] = 4.0
            options['Transverse-distal tenia coli width'] = 3.0
            options['Distal tenia coli width'] = 1.5
        return options

    @staticmethod
    def getOrderedOptionNames():
        return [
            'Central path',
            'Segment profile',
            'Number of segments',
            'Start phase',
            'Proximal tenia coli width',
            'Proximal-transverse tenia coli width',
            'Transverse-distal tenia coli width',
            'Distal tenia coli width',
            'Refine',
            'Refine number of elements around',
            'Refine number of elements along',
            'Refine number of elements through wall']

    @classmethod
    def getOptionValidScaffoldTypes(cls, optionName):
        if optionName == 'Central path':
            return [MeshType_1d_path1]
        if optionName == 'Segment profile':
            return [MeshType_3d_colonsegment1]
        return []

    @classmethod
    def getOptionScaffoldTypeParameterSetNames(cls, optionName, scaffoldType):
        if optionName == 'Central path':
            return list(cls.centralPathDefaultScaffoldPackages.keys())
        assert scaffoldType in cls.getOptionValidScaffoldTypes(optionName), \
            cls.__name__ + '.getOptionScaffoldTypeParameterSetNames.  ' + \
            'Invalid option \'' + optionName + '\' scaffold type ' + scaffoldType.getName()
        return scaffoldType.getParameterSetNames()

    @classmethod
    def getOptionScaffoldPackage(cls, optionName, scaffoldType, parameterSetName=None):
        '''
        :param parameterSetName:  Name of valid parameter set for option Scaffold, or None for default.
        :return: ScaffoldPackage.
        '''
        if parameterSetName:
            assert parameterSetName in cls.getOptionScaffoldTypeParameterSetNames(optionName, scaffoldType), \
                'Invalid parameter set ' + str(parameterSetName) + ' for scaffold ' + str(scaffoldType.getName()) + \
                ' in option ' + str(optionName) + ' of scaffold ' + cls.getName()
        if optionName == 'Central path':
            if not parameterSetName:
                parameterSetName = list(cls.centralPathDefaultScaffoldPackages.keys())[0]
            return copy.deepcopy(cls.centralPathDefaultScaffoldPackages[parameterSetName])
        if optionName == 'Segment profile':
            if not parameterSetName:
                parameterSetName = scaffoldType.getParameterSetNames()[0]
            return ScaffoldPackage(scaffoldType, defaultParameterSetName=parameterSetName)
        assert False, cls.__name__ + '.getOptionScaffoldPackage:  Option ' + optionName + ' is not a scaffold'

    @classmethod
    def checkOptions(cls, options):
        if not options['Central path'].getScaffoldType() in cls.getOptionValidScaffoldTypes('Central path'):
            options['Central path'] = cls.getOptionScaffoldPackage('Central path', MeshType_1d_path1)
        if not options['Segment profile'].getScaffoldType() in cls.getOptionValidScaffoldTypes('Segment profile'):
            options['Segment profile'] = cls.getOptionScaffoldPackage('Segment profile', MeshType_3d_colonsegment1)
        for key in [
            'Number of segments',
            'Refine number of elements around',
            'Refine number of elements along',
            'Refine number of elements through wall']:
            if options[key] < 1:
                options[key] = 1
        for key in [
            'Proximal tenia coli width',
            'Proximal-transverse tenia coli width',
            'Transverse-distal tenia coli width',
            'Distal tenia coli width']:
            if options[key] < 0.0:
                options[key] = 0.0

    @classmethod
    def generateBaseMesh(cls, region, options):
        """
        Generate the base tricubic Hermite mesh.
        :param region: Zinc region to define model in. Must be empty.
        :param options: Dict containing options. See getDefaultOptions().
        :return: list of AnnotationGroup, None
        """
        centralPath = options['Central path']
        segmentProfile = options['Segment profile']
        segmentCount = options['Number of segments']
        startPhase = options['Start phase'] % 360.0
        proximalTCWidth = options['Proximal tenia coli width']
        proximalTransverseTCWidth = options['Proximal-transverse tenia coli width']
        transverseDistalTCWidth = options['Transverse-distal tenia coli width']
        distalTCWidth = options['Distal tenia coli width']
        segmentSettings = segmentProfile.getScaffoldSettings()

        elementsCountAroundTC = segmentSettings['Number of elements around tenia coli']
        elementsCountAroundHaustrum = segmentSettings['Number of elements around haustrum']
        cornerInnerRadiusFactor = segmentSettings['Corner inner radius factor']
        haustrumInnerRadiusFactor = segmentSettings['Haustrum inner radius factor']
        segmentLengthEndDerivativeFactor = segmentSettings['Segment length end derivative factor']
        segmentLengthMidDerivativeFactor = segmentSettings['Segment length mid derivative factor']
        tcCount = segmentSettings['Number of tenia coli']
        tcThickness = segmentSettings['Tenia coli thickness']
        elementsCountAround = (elementsCountAroundTC + elementsCountAroundHaustrum) * tcCount

        elementsCountAlongSegment = segmentSettings['Number of elements along segment']
        elementsCountThroughWall = segmentSettings['Number of elements through wall']
        wallThickness = segmentSettings['Wall thickness']
        mucosaRelThickness = segmentSettings['Mucosa relative thickness']
        submucosaRelThickness = segmentSettings['Submucosa relative thickness']
        circularRelThickness = segmentSettings['Circular muscle layer relative thickness']
        longitudinalRelThickness = segmentSettings['Longitudinal muscle layer relative thickness']
        useCrossDerivatives = segmentSettings['Use cross derivatives']
        useCubicHermiteThroughWall = not (segmentSettings['Use linear through wall'])
        elementsCountAlong = int(elementsCountAlongSegment * segmentCount)

        # Colon coordinates
        lengthToDiameterRatio = 24
        wallThicknessToDiameterRatio = 0.1
        teniaColiThicknessToDiameterRatio = 0.25 * wallThicknessToDiameterRatio
        relativeThicknessListColonCoordinates = [1.0 / elementsCountThroughWall for n3 in range(elementsCountThroughWall)]

        firstNodeIdentifier = 1
        firstElementIdentifier = 1

        # Central path
        if tcCount == 1:
            colonTermsAlong = [None, 'right colon', 'transverse colon', 'left colon']
        elif tcCount == 2:
            colonTermsAlong = [None, 'spiral colon', 'transverse colon', 'descending colon']
        elif tcCount == 3:
            colonTermsAlong = [None, 'ascending colon', 'transverse colon', 'descending colon']

        tmpRegion = region.createRegion()
        centralPath.generate(tmpRegion)
        tmpFieldmodule = tmpRegion.getFieldmodule()
        tmpNodes = tmpFieldmodule.findNodesetByFieldDomainType(Field.DOMAIN_TYPE_NODES)
        tmpCoordinates = tmpFieldmodule.findFieldByName('coordinates')
        arcLengthOfGroupsAlong = []

        for termName in colonTermsAlong:
            tmpGroup = tmpFieldmodule.findFieldByName(termName).castGroup() if termName else None
            tmpNodeset = tmpGroup.getNodesetGroup(tmpNodes) if tmpGroup else tmpNodes

            cxGroup, cd1Group, cd2Group, cd3Group, cd12Group, cd13Group = get_nodeset_path_field_parameters(
                tmpNodeset, tmpCoordinates,
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1,
                 Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D_DS3,
                 Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D2_DS1DS3])
            arcLength = 0.0
            for e in range(len(cxGroup) - 1):
                arcLength += interp.getCubicHermiteArcLength(cxGroup[e], cd1Group[e],
                                                             cxGroup[e + 1], cd1Group[e + 1])
            arcLengthOfGroupsAlong.append(arcLength)

            if not termName:
                cx = cxGroup
                cd1 = cd1Group
                cd2 = cd2Group
                cd3 = cd3Group
                cd12 = cd12Group
                cd13 = cd13Group

            del tmpNodeset
            del tmpGroup

        del tmpCoordinates
        del tmpNodes
        del tmpFieldmodule
        del tmpRegion

        # find arclength of colon
        length = 0.0
        elementsCountIn = len(cx) - 1
        sd1 = interp.smoothCubicHermiteDerivativesLine(cx, cd1, fixAllDirections=True,
                                                       magnitudeScalingMode=interp.DerivativeScalingMode.HARMONIC_MEAN)
        for e in range(elementsCountIn):
            arcLength = interp.getCubicHermiteArcLength(cx[e], sd1[e], cx[e + 1], sd1[e + 1])
            # print(e+1, arcLength)
            length += arcLength
        segmentLength = length / segmentCount

        # Sample central path
        sx, sd1, se, sxi, ssf = interp.sampleCubicHermiteCurves(cx, cd1, elementsCountAlong)
        sd2, sd12 = interp.interpolateSampleCubicHermite(cd2, cd12, se, sxi, ssf)

        centralPathLength = arcLengthOfGroupsAlong[0]
        elementAlongLength = centralPathLength / elementsCountAlong

        elementsCountAlongGroups = []
        groupLength = 0.0
        e = 0
        elementsCount = 1
        length = elementAlongLength
        for i in range(1, len(colonTermsAlong)):
            groupLength += arcLengthOfGroupsAlong[i]
            if e == elementsCountAlong - 2:
                elementsCount += 1
                elementsCountAlongGroups.append(elementsCount)
            else:
                while length < groupLength:
                    elementsCount += 1
                    e += 1
                    length += elementAlongLength

                # check which end is grouplength closer to
                distToUpperEnd = abs(length - groupLength)
                distToLowerEnd = abs(groupLength - (length - elementsCountAlong))
                if distToLowerEnd < distToUpperEnd:
                    elementsCount -= 1
                    elementsCountAlongGroups.append(elementsCount)
                    e -= 1
                    length -= elementAlongLength
                else:
                    elementsCountAlongGroups.append(elementsCount)
            elementsCount = 0

        # Generate variation of radius & tc width along length
        lengthList = [0.0, arcLengthOfGroupsAlong[1], arcLengthOfGroupsAlong[1] + arcLengthOfGroupsAlong[2],
                      arcLengthOfGroupsAlong[0]]

        innerRadiusListCP = [vector.magnitude(c) for c in cd2]
        dInnerRadiusListCP = []
        for n in range(len(innerRadiusListCP) - 1):
            dInnerRadiusListCP.append(innerRadiusListCP[n + 1] - innerRadiusListCP[n])
        dInnerRadiusListCP.append(innerRadiusListCP[-1] - innerRadiusListCP[-2])
        innerRadiusAlongElementList, dInnerRadiusAlongElementList = interp.interpolateSampleCubicHermite(
            innerRadiusListCP, dInnerRadiusListCP, se, sxi, ssf)

        tcWidthList = [proximalTCWidth, proximalTransverseTCWidth, transverseDistalTCWidth, distalTCWidth]
        tcWidthAlongElementList, dTCWidthAlongElementList = interp.sampleParameterAlongLine(lengthList,
                                                                                            tcWidthList,
                                                                                            elementsCountAlong)

        # Account for reduced haustrum appearance in transverse and distal pig colon
        if tcCount == 2:
            haustrumInnerRadiusFactorList = [haustrumInnerRadiusFactor, haustrumInnerRadiusFactor * 0.75,
                                             haustrumInnerRadiusFactor * 0.5, haustrumInnerRadiusFactor * 0.2]
            haustrumInnerRadiusFactorAlongElementList = \
                interp.sampleParameterAlongLine(lengthList, haustrumInnerRadiusFactorList, elementsCountAlong)[0]
        else:
            haustrumInnerRadiusFactorAlongElementList = [haustrumInnerRadiusFactor] * (elementsCountAlong + 1)

        # Create annotation groups for colon sections
        colonGroup = AnnotationGroup(region, get_colon_term("colon"))

        if tcCount == 1:
            proximalGroup = AnnotationGroup(region, get_colon_term("right colon"))
            transverseGroup = AnnotationGroup(region, get_colon_term("transverse colon"))
            distalGroup = AnnotationGroup(region, get_colon_term("left colon"))
            annotationGroupAlong = [[colonGroup, proximalGroup],
                                    [colonGroup, transverseGroup],
                                    [colonGroup, distalGroup]]

        elif tcCount == 2:
            spiralGroup = AnnotationGroup(region, get_colon_term("spiral colon"))
            transverseGroup = AnnotationGroup(region, get_colon_term("transverse colon"))
            distalGroup = AnnotationGroup(region, get_colon_term("descending colon"))
            annotationGroupAlong = [[colonGroup, spiralGroup],
                                    [colonGroup, transverseGroup],
                                    [colonGroup, distalGroup]]

        elif tcCount == 3:
            ascendingGroup = AnnotationGroup(region, get_colon_term("ascending colon"))
            transverseGroup = AnnotationGroup(region, get_colon_term("transverse colon"))
            descendingGroup = AnnotationGroup(region, get_colon_term("descending colon"))
            annotationGroupAlong = [[colonGroup, ascendingGroup],
                                    [colonGroup, transverseGroup],
                                    [colonGroup, descendingGroup]]

        annotationGroupsAlong = []
        for i in range(len(elementsCountAlongGroups)):
            elementsCount = elementsCountAlongGroups[i]
            for n in range(elementsCount):
                annotationGroupsAlong.append(annotationGroupAlong[i])

        xExtrude = []
        d1Extrude = []
        d2Extrude = []
        d3UnitExtrude = []
        sxRefExtrudeList = []

        if elementsCountThroughWall == 1:
            relativeThicknessList = [1.0]
            annotationGroupsThroughWall = [[]]
        else:
            relativeThicknessList = [mucosaRelThickness, submucosaRelThickness,
                                     circularRelThickness, longitudinalRelThickness]
            mucosaGroup = AnnotationGroup(region, get_colon_term("colonic mucosa"))
            submucosaGroup = AnnotationGroup(region, get_colon_term("submucosa of colon"))
            circularMuscleGroup = AnnotationGroup(region, get_colon_term("circular muscle layer of colon"))
            longitudinalMuscleGroup = AnnotationGroup(region, get_colon_term("longitudinal muscle layer of colon"))
            annotationGroupsThroughWall = [[mucosaGroup], [submucosaGroup],
                                           [circularMuscleGroup], [longitudinalMuscleGroup]]

        # Create object
        colonSegmentTubeMeshInnerPoints = ColonSegmentTubeMeshInnerPoints(
            region, elementsCountAroundTC, elementsCountAroundHaustrum, elementsCountAlongSegment,
            tcCount, segmentLengthEndDerivativeFactor, segmentLengthMidDerivativeFactor,
            segmentLength, wallThickness, cornerInnerRadiusFactor, haustrumInnerRadiusFactorAlongElementList,
            innerRadiusAlongElementList, dInnerRadiusAlongElementList, tcWidthAlongElementList,
            startPhase)

        for nSegment in range(segmentCount):
            # Create inner points
            xInner, d1Inner, d2Inner, transitElementList, segmentAxis, annotationGroupsAround \
                = colonSegmentTubeMeshInnerPoints.getColonSegmentTubeMeshInnerPoints(nSegment)

            # Project reference point for warping onto central path
            start = nSegment * elementsCountAlongSegment
            end = (nSegment + 1) * elementsCountAlongSegment + 1
            sxRefList, sd1RefList, sd2ProjectedListRef, zRefList = \
                tubemesh.getPlaneProjectionOnCentralPath(xInner, elementsCountAround, elementsCountAlongSegment,
                                                         segmentLength, sx[start:end], sd1[start:end], sd2[start:end],
                                                         sd12[start:end])

            # Warp segment points
            xWarpedList, d1WarpedList, d2WarpedList, d3WarpedUnitList = tubemesh.warpSegmentPoints(
                xInner, d1Inner, d2Inner, segmentAxis, sxRefList, sd1RefList, sd2ProjectedListRef,
                elementsCountAround, elementsCountAlongSegment, zRefList)

            # Store points along length
            xExtrude += xWarpedList if nSegment == 0 else xWarpedList[elementsCountAround:]
            d1Extrude += d1WarpedList if nSegment == 0 else d1WarpedList[elementsCountAround:]
            d2Extrude += d2WarpedList if nSegment == 0 else d2WarpedList[elementsCountAround:]
            d3UnitExtrude += d3WarpedUnitList if nSegment == 0 else d3WarpedUnitList[elementsCountAround:]
            sxRefExtrudeList += sxRefList if nSegment == 0 else sxRefList[elementsCountAround:]

        contractedWallThicknessList = colonSegmentTubeMeshInnerPoints.getContractedWallThicknessList()

        # Create coordinates and derivatives
        xList, d1List, d2List, d3List, curvatureList = tubemesh.extrudeSurfaceCoordinates(xExtrude, d1Extrude,
                                                                                        d2Extrude, d3UnitExtrude, contractedWallThicknessList, relativeThicknessList,
                                                                                        elementsCountAround, elementsCountAlong, elementsCountThroughWall, transitElementList,
                                                                                          outward=True)

        relaxedLengthList, xiList = colonSegmentTubeMeshInnerPoints.getRelaxedLengthAndXiList()

        closedProximalEnd = False

        if tcThickness > 0:
            tubeTCWidthList = colonSegmentTubeMeshInnerPoints.getTubeTCWidthList()
            xList, d1List, d2List, d3List, annotationArrayAround = getTeniaColi(
                region, xList, d1List, d2List, d3List, curvatureList, tcCount, elementsCountAroundTC,
                elementsCountAroundHaustrum, elementsCountAlong, elementsCountThroughWall,
                tubeTCWidthList, tcThickness, sxRefExtrudeList, annotationGroupsAround,
                closedProximalEnd)

            # Create flat coordinates
            xFlat, d1Flat, d2Flat = createFlatCoordinatesTeniaColi(
                xiList, relaxedLengthList, length, wallThickness, relativeThicknessList, tcCount, tcThickness,
                elementsCountAroundTC, elementsCountAroundHaustrum, elementsCountAlong,
                elementsCountThroughWall, transitElementList, closedProximalEnd)

            # Create colon coordinates
            xColon, d1Colon, d2Colon = createColonCoordinatesTeniaColi(xiList, relativeThicknessListColonCoordinates,
                                                                       lengthToDiameterRatio,
                                                                       wallThicknessToDiameterRatio,
                                                                       teniaColiThicknessToDiameterRatio, tcCount,
                                                                       elementsCountAroundTC,
                                                                       elementsCountAroundHaustrum,
                                                                       elementsCountAlong, elementsCountThroughWall,
                                                                       transitElementList, closedProximalEnd)

            # Create nodes and elements
            nextNodeIdentifier, nextElementIdentifier, annotationGroups = createNodesAndElementsTeniaColi(
                region, xList, d1List, d2List, d3List, xFlat, d1Flat, d2Flat, xColon, d1Colon, d2Colon,
                "colon coordinates", elementsCountAroundTC, elementsCountAroundHaustrum, elementsCountAlong,
                elementsCountThroughWall, tcCount, annotationGroupsAround, annotationGroupsAlong,
                annotationGroupsThroughWall, firstNodeIdentifier, firstElementIdentifier, useCubicHermiteThroughWall,
                useCrossDerivatives, closedProximalEnd)

        else:
            # Create flat coordinates
            xFlat, d1Flat, d2Flat = tubemesh.createFlatCoordinates(
                xiList, relaxedLengthList, length, wallThickness, relativeThicknessList, elementsCountAround,
                elementsCountAlong, elementsCountThroughWall, transitElementList)

            # Create colon coordinates
            xColon, d1Colon, d2Colon = tubemesh.createOrganCoordinates(xiList, relativeThicknessListColonCoordinates,
                                                                       lengthToDiameterRatio,
                                                                       wallThicknessToDiameterRatio,
                                                                       elementsCountAround,
                                                                       elementsCountAlong, elementsCountThroughWall,
                                                                       transitElementList)

            # Create nodes and elements
            nextNodeIdentifier, nextElementIdentifier, annotationGroups = tubemesh.createNodesAndElements(
                region, xList, d1List, d2List, d3List, xFlat, d1Flat, d2Flat, xColon, d1Colon, d2Colon,
                "colon coordinates", elementsCountAround, elementsCountAlong, elementsCountThroughWall,
                annotationGroupsAround, annotationGroupsAlong, annotationGroupsThroughWall,
                firstNodeIdentifier, firstElementIdentifier, useCubicHermiteThroughWall, useCrossDerivatives,
                closedProximalEnd)

        return annotationGroups, None

    @classmethod
    def refineMesh(cls, meshrefinement, options):
        """
        Refine source mesh into separate region, with change of basis.
        :param meshrefinement: MeshRefinement, which knows source and target region.
        :param options: Dict containing options. See getDefaultOptions().
        """
        refineElementsCountAround = options['Refine number of elements around']
        refineElementsCountAlong = options['Refine number of elements along']
        refineElementsCountThroughWall = options['Refine number of elements through wall']

        meshrefinement.refineAllElementsCubeStandard3d(refineElementsCountAround, refineElementsCountAlong,
                                                       refineElementsCountThroughWall)
        return

    @classmethod
    def defineFaceAnnotations(cls, region, options, annotationGroups):
        '''
        Add face annotation groups from the highest dimension mesh.
        Must have defined faces and added subelements for highest dimension groups.
        :param region: Zinc region containing model.
        :param options: Dict containing options. See getDefaultOptions().
        :param annotationGroups: List of annotation groups for top-level elements.
        New face annotation groups are appended to this list.
        '''
        # Create 2d surface mesh groups
        fm = region.getFieldmodule()
        mesh2d = fm.findMeshByDimension(2)

        colonGroup = getAnnotationGroupForTerm(annotationGroups, get_colon_term("colon"))
        is_exterior = fm.createFieldIsExterior()
        is_exterior_face_xi3_1 = fm.createFieldAnd(is_exterior, fm.createFieldIsOnFace(Element.FACE_TYPE_XI3_1))
        is_exterior_face_xi3_0 = fm.createFieldAnd(is_exterior, fm.createFieldIsOnFace(Element.FACE_TYPE_XI3_0))
        is_colon = colonGroup.getGroup()
        is_serosa = fm.createFieldAnd(is_colon, is_exterior_face_xi3_1)
        serosa = findOrCreateAnnotationGroupForTerm(annotationGroups, region, get_colon_term("serosa of colon"))
        serosa.getMeshGroup(mesh2d).addElementsConditional(is_serosa)
        is_mucosaInnerSurface = fm.createFieldAnd(is_colon, is_exterior_face_xi3_0)
        mucosaInnerSurface = findOrCreateAnnotationGroupForTerm(annotationGroups, region,
                                                                get_colon_term("luminal surface of the colonic mucosa"))
        mucosaInnerSurface.getMeshGroup(mesh2d).addElementsConditional(is_mucosaInnerSurface)
