"""
Generates a 3-D small intestine mesh along the central line,
with variable numbers of elements around, along and through
wall, with variable radius and thickness along.
"""

import copy

from cmlibs.zinc.element import Element
from cmlibs.zinc.field import Field
from cmlibs.zinc.node import Node
from scaffoldmaker.annotation.annotationgroup import AnnotationGroup, findOrCreateAnnotationGroupForTerm, \
    getAnnotationGroupForTerm
from scaffoldmaker.annotation.smallintestine_terms import get_smallintestine_term
from scaffoldmaker.meshtypes.meshtype_1d_path1 import MeshType_1d_path1
from scaffoldmaker.meshtypes.scaffold_base import Scaffold_base
from scaffoldmaker.scaffoldpackage import ScaffoldPackage
from scaffoldmaker.utils import interpolation as interp
from scaffoldmaker.utils import tubemesh
from scaffoldmaker.utils import vector
from scaffoldmaker.utils.tubemesh import CylindricalSegmentTubeMeshInnerPoints
from scaffoldmaker.utils.zinc_utils import exnode_string_from_nodeset_field_parameters, \
    get_nodeset_path_field_parameters


class MeshType_3d_smallintestine1(Scaffold_base):
    '''
    Generates a 3-D small intestine mesh with variable numbers
    of elements around, along the central line, and through wall.
    The small intestine is created by a function that generates
    a small intestine segment and uses tubemesh to map the segment
    along a central line profile.
    '''

    centralPathDefaultScaffoldPackages = {
        'Cattle 1' : ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings' : {
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Coordinate dimensions' : 3,
                'Length' : 1.0,
                'Number of elements' : 536
                },
            'meshEdits' : exnode_string_from_nodeset_field_parameters(
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3], [
                (1,   [ [ -502.30,  732.10,   92.00 ], [  -49.80,  -90.70,    0.00 ], [   0.00,   0.00,  10.00 ], [   0.00,   0.00,   0.00 ], [  -8.77,   4.81,   0.00 ], [   0.00,   0.00,   0.00 ] ] ),
                (2,   [ [ -543.10,  654.80,   92.00 ], [  -31.70,  -64.00,    0.00 ], [   0.00,   0.00,  10.00 ], [   0.00,   0.00,   0.00 ], [  -8.96,   4.44,   0.00 ], [   0.00,   0.00,   0.00 ] ] ),
                (3,   [ [ -566.20,  605.00,   92.00 ], [  -43.40,  -40.20,    0.00 ], [   0.00,   0.00,  10.00 ], [   0.00,   0.00,   0.00 ], [  -6.80,   7.34,   0.00 ], [   0.00,   0.00,   0.00 ] ] ),
                (4,   [ [ -623.90,  583.60,   92.00 ], [  -58.90,    0.00,    0.00 ], [   0.00,   0.00,  10.00 ], [   0.00,   0.00,   0.00 ], [   0.00,  10.00,   0.00 ], [   0.00,   0.00,   0.00 ] ] ),
                (5,   [ [ -674.40,  602.30,   92.00 ], [  -25.80,   24.90,    0.00 ], [   0.00,   0.00,  10.00 ], [   1.57,   1.24,  -0.27 ], [   6.94,   7.20,   0.00 ], [   1.57,   1.24,  -0.27 ] ] ),
                (6,   [ [ -676.30,  613.40,   92.00 ], [   21.70,   17.10,   -7.50 ], [   2.05,   1.62,   9.65 ], [   1.52,   0.69,  -0.22 ], [   6.19,  -7.86,   0.01 ], [   1.52,   0.69,  -0.22 ] ] ),
                (7,   [ [ -638.20,  613.20,   80.50 ], [   47.90,    1.90,  -11.30 ], [   2.29,   0.09,   9.73 ], [  -0.55,   2.15,  -1.26 ], [   0.39,  -9.99,  -0.00 ], [  -0.55,   2.15,  -1.25 ] ] ),
                (8,   [ [ -586.80,  617.60,   72.10 ], [    3.30,   41.00,  -47.10 ], [   0.61,   7.52,   6.56 ], [  -1.76,   0.56,  -0.15 ], [   9.97,  -0.79,  -0.02 ], [  -1.77,   0.56,  -0.14 ] ] ),
                (9,   [ [ -644.60,  633.70,   57.50 ], [  -52.20,    4.60,   -6.70 ], [  -1.25,   0.12,   9.92 ], [   0.12,  -4.47,   1.05 ], [   0.89,   9.96,  -0.01 ], [   0.12,  -4.47,   1.05 ] ] ),
                (10,  [ [ -677.20,  630.20,   57.50 ], [    0.10,  -43.70,  -12.10 ], [  -0.00,  -2.66,   9.64 ], [   1.58,  -0.74,  -0.11 ], [ -10.00,  -0.01,  -0.00 ], [   1.58,  -0.74,  -0.10 ] ] ),
                (11,  [ [ -632.90,  609.90,   50.60 ], [   51.90,  -16.20,  -11.70 ], [   2.02,  -0.63,   9.77 ], [   0.39,  -1.35,  -0.70 ], [  -2.97,  -9.55,  -0.00 ], [   0.39,  -1.35,  -0.70 ] ] ),
                (12,  [ [ -581.30,  601.00,   35.40 ], [    4.90,  -48.00,  -34.50 ], [   0.57,  -5.79,   8.13 ], [  -1.42,   0.41,   0.08 ], [  -9.95,  -1.01,  -0.02 ], [  -1.42,   0.40,   0.08 ] ] ),
                (13,  [ [ -631.70,  589.30,   35.90 ], [  -57.50,   19.90,   -5.30 ], [  -0.82,   0.28,   9.96 ], [  -2.13,   7.48,  -3.39 ], [   3.27,   9.45,   0.00 ], [  -2.13,   7.47,  -3.39 ] ] ),
                (14,  [ [ -667.40,  632.00,   27.30 ], [  -60.00,  -18.80,  -39.10 ], [  -3.71,   9.21,   1.15 ], [  -1.20,  -0.91,  -9.00 ], [   4.52,   2.88,  -8.44 ], [  -1.20,  -0.91,  -9.00 ] ] ),
                (15,  [ [ -666.40,  558.00,    7.70 ], [  -33.70,  -54.60,   46.30 ], [  -2.57,  -5.27,  -8.10 ], [  -1.72,  -4.99,  -3.68 ], [   8.68,  -4.95,   0.47 ], [  -1.72,  -4.99,  -3.68 ] ] ),
                (16,  [ [ -698.00,  570.70,   66.50 ], [  -52.90,    5.40,   48.50 ], [  -6.71,  -2.24,  -7.06 ], [  -2.86,  -0.08,   4.78 ], [   0.98,  -9.72,   2.16 ], [  -2.87,  -0.08,   4.78 ] ] ),
                (17,  [ [ -760.20,  567.80,   94.30 ], [  -13.60,   33.40,   46.80 ], [  -8.32,  -5.37,   1.39 ], [   4.50,   1.96,   8.55 ], [   5.04,  -6.28,   5.92 ], [   4.51,   1.96,   8.55 ] ] ),
                (18,  [ [ -724.50,  582.10,   93.50 ], [   45.90,    1.10,   -0.90 ], [   0.19,  -0.07,  10.00 ], [   4.20,   2.65,   4.23 ], [   0.23, -10.00,  -0.07 ], [   4.20,   2.65,   4.24 ] ] ),
                (19,  [ [ -681.80,  570.00,   92.80 ], [   28.80,  -21.00,   -0.70 ], [   0.24,   0.02,  10.02 ], [  -0.11,  -0.21,   0.02 ], [  -5.91,  -8.09,   0.16 ], [  -0.11,  -0.21,   0.02 ] ] ),
                (20,  [ [ -667.90,  548.90,   92.30 ], [    3.90,  -28.50,   -1.00 ], [   0.03,  -0.37,  10.03 ], [  -0.05,  -0.00,   0.02 ], [  -9.94,  -1.38,  -0.02 ], [  -0.05,  -0.00,   0.02 ] ] ),
                (21,  [ [ -677.10,  520.90,   91.00 ], [  -27.30,  -11.70,    0.60 ], [   0.17,   0.09,  10.06 ], [  -0.32,   0.01,   0.01 ], [  -3.97,   9.24,  -0.02 ], [  -0.32,   0.01,   0.01 ] ] ),
                (22,  [ [ -700.50,  532.40,   92.90 ], [  -14.40,   20.00,   -0.20 ], [  -0.56,  -0.30,  10.06 ], [  -0.35,   0.31,   0.00 ], [   8.16,   5.88,   0.63 ], [  -0.36,   0.32,   0.00 ] ] ),
                (23,  [ [ -703.50,  552.30,   91.10 ], [  -16.60,   15.00,   -1.80 ], [  -0.63,   0.54,  10.06 ], [  -0.38,   0.11,  -0.02 ], [   6.75,   7.50,   0.02 ], [  -0.39,   0.11,  -0.02 ] ] ),
                (24,  [ [ -726.50,  553.80,   90.20 ], [  -21.70,   -2.90,   -3.00 ], [  -1.37,  -0.18,  10.02 ], [  -1.63,  -5.16,  -4.21 ], [  -1.35,  10.02,  -0.00 ], [  -1.63,  -5.16,  -4.21 ] ] ),
                (25,  [ [ -739.60,  548.10,   86.40 ], [   15.30,  -13.20,  -28.20 ], [  -3.76,  -9.13,   2.25 ], [  -1.48,  -2.79,  -8.86 ], [  -8.28,   2.07,  -5.44 ], [  -1.48,  -2.79,  -8.86 ] ] ),
                (26,  [ [ -694.50,  550.10,   70.60 ], [   44.70,   -4.70,  -16.30 ], [  -3.06,   2.92,  -9.23 ], [   1.47,   4.72,  -6.06 ], [   1.90,   9.68,   2.43 ], [   1.47,   4.71,  -6.06 ] ] ),
                (27,  [ [ -652.70,  539.30,   54.70 ], [   26.40,  -22.50,   -3.70 ], [  -0.85,   0.66, -10.13 ], [   0.77,  -2.83,  -0.13 ], [   6.61,   7.75,  -0.05 ], [   0.77,  -2.82,  -0.13 ] ] ),
                (28,  [ [ -644.90,  520.20,   58.60 ], [   -8.40,  -26.00,    7.10 ], [  -0.80,  -2.45,  -9.87 ], [  -0.33,  -0.93,   0.06 ], [   9.70,  -3.15,  -0.00 ], [  -0.33,  -0.92,   0.06 ] ] ),
                (29,  [ [ -670.50,  503.30,   65.20 ], [  -49.20,   -7.40,    8.40 ], [  -1.67,  -0.25, -10.09 ], [  -0.69,  -1.22,   1.36 ], [   1.54, -10.11,  -0.00 ], [  -0.69,  -1.22,   1.37 ] ] ),
                (30,  [ [ -723.20,  522.10,   69.80 ], [   13.50,   42.00,  -69.30 ], [  -1.98,  -8.44,  -5.51 ], [  -0.28,  -0.80,   0.83 ], [  -9.93,   2.58,  -0.38 ], [  -0.28,  -0.80,   0.83 ] ] ),
                (31,  [ [ -641.70,  515.00,   30.80 ], [   50.00,  -30.40,  -15.00 ], [  -2.21,   1.29, -10.01 ], [   1.71,  -0.27,  -1.09 ], [   5.37,   8.83,  -0.05 ], [   1.71,  -0.28,  -1.09 ] ] ),
                (32,  [ [ -631.40,  498.10,   31.40 ], [   -4.50,  -28.10,    4.10 ], [  -0.02,  -1.49, -10.24 ], [   0.71,  -0.99,  -0.15 ], [  10.22,  -1.61,   0.21 ], [   0.71,  -0.99,  -0.15 ] ] ),
                (33,  [ [ -652.60,  477.30,   37.00 ], [  -50.00,   -6.30,    6.60 ], [  -1.34,  -0.02, -10.29 ], [  -1.33,   1.31,   0.06 ], [   1.29, -10.29,  -0.15 ], [  -1.33,   1.31,   0.06 ] ] ),
                (34,  [ [ -711.30,  508.40,   47.40 ], [  -65.20,   35.30,   20.70 ], [  -2.68,   0.96, -10.02 ], [  -4.19,   0.21,   3.24 ], [  -4.87,  -9.20,   0.42 ], [  -4.19,   0.21,   3.24 ] ] ),
                (35,  [ [ -756.90,  547.90,   71.40 ], [  -13.40,  -32.80,   30.60 ], [  -9.76,   0.39,  -3.77 ], [  -0.16,  -0.33,  -0.99 ], [   2.37,  -7.50,  -6.91 ], [  -0.16,  -0.33,  -0.99 ] ] ),
                (36,  [ [ -764.20,  524.80,   82.70 ], [    2.40,  -34.00,   -3.00 ], [  -7.09,   0.16,  -7.72 ], [   1.12,   1.17,   2.11 ], [   7.69,   1.14,  -7.04 ], [   1.12,   1.17,   2.11 ] ] ),
                (37,  [ [ -733.10,  495.70,   71.40 ], [   -9.60,  -50.80,   23.30 ], [  -8.61,   3.72,   4.77 ], [   3.16,  -1.87,   8.64 ], [  -5.81,  -2.81,  -8.30 ], [   3.17,  -1.87,   8.64 ] ] ),
                (38,  [ [ -714.70,  462.80,   83.00 ], [   19.70,   -9.40,    0.00 ], [  -1.38,  -2.86,  10.06 ], [  -2.27,   2.74,  -8.49 ], [  -4.34,  -9.07,  -3.17 ], [  -2.27,   2.74,  -8.49 ] ] ),
                (39,  [ [ -703.00,  463.20,   92.20 ], [   -6.90,   19.30,   27.80 ], [  -8.20,   4.34,  -5.03 ], [  -0.98,   2.71, -11.42 ], [  -6.31,  -7.60,   3.73 ], [  -0.98,   2.71, -11.43 ] ] ),
                (40,  [ [ -722.30,  486.20,   90.20 ], [  -30.40,   26.10,   -0.90 ], [   0.20,  -0.15, -10.58 ], [   4.54,  -2.26,  -3.13 ], [  -6.88,  -8.04,  -0.02 ], [   4.54,  -2.26,  -3.13 ] ] ),
                (41,  [ [ -745.10,  520.90,   90.70 ], [  -10.10,   19.90,    1.10 ], [  -0.24,   0.48, -10.60 ], [   0.25,  -0.06,  -0.02 ], [  -9.47,  -4.79,  -0.00 ], [   0.25,  -0.06,  -0.02 ] ] ),
                (42,  [ [ -724.50,  528.30,   92.00 ], [   61.80,   -0.90,    2.80 ], [   0.46,  -0.02, -10.62 ], [  -0.36,  -0.13,   0.04 ], [   0.13,  10.63,  -0.01 ], [  -0.36,  -0.13,   0.04 ] ] ),
                (43,  [ [ -691.00,  492.20,   92.50 ], [   65.60,  -18.50,  -13.50 ], [  -1.98,   0.57, -10.46 ], [  -1.69,   0.78,   0.19 ], [   2.91,  10.26,   0.01 ], [  -1.70,   0.78,   0.19 ] ] ),
                (44,  [ [ -640.00,  474.10,   57.50 ], [   42.10,  -24.20,  -15.10 ], [  -2.75,   1.58, -10.23 ], [   1.45,  -0.62,  -0.23 ], [   5.34,   9.28,  -0.00 ], [   1.45,  -0.63,  -0.23 ] ] ),
                (45,  [ [ -611.10,  454.00,   57.30 ], [   -5.20,  -35.60,    0.00 ], [   0.00,   0.00, -10.73 ], [   1.27,  -0.75,  -0.25 ], [  10.62,  -1.55,  -0.00 ], [   1.28,  -0.75,  -0.25 ] ] ),
                (46,  [ [ -640.70,  443.80,   57.50 ], [  -30.10,   -5.20,    0.10 ], [  -0.04,  -0.02, -10.76 ], [   0.00,   0.00,  -0.02 ], [   1.82, -10.60,   0.01 ], [   0.00,   0.00,  -0.02 ] ] ),
                (47,  [ [ -667.60,  443.10,   57.50 ], [  -25.60,    5.80,    0.00 ], [   0.00,   0.00, -10.77 ], [   0.86,  -0.09,   0.06 ], [  -2.38, -10.51,  -0.00 ], [   0.85,  -0.09,   0.06 ] ] ),
                (48,  [ [ -702.10,  461.20,   42.90 ], [  -35.10,    4.20,   -7.10 ], [   2.12,  -0.26, -10.59 ], [  -2.58,  -3.51,   4.12 ], [  -1.28, -10.73,   0.01 ], [  -2.58,  -3.51,   4.12 ] ] ),
                (49,  [ [ -751.10,  480.00,   41.40 ], [   -2.70,   -4.20,   52.00 ], [  -6.95,  -8.26,  -1.02 ], [   0.95,  -1.86,  -0.23 ], [   8.30,  -6.97,  -0.12 ], [   0.95,  -1.86,  -0.23 ] ] ),
                (50,  [ [ -740.70,  460.80,   81.10 ], [   28.50,  -35.40,   18.50 ], [   2.29,  -5.05,  -9.35 ], [   0.35,   3.93,  -4.00 ], [   9.49,   5.25,  -0.74 ], [   0.34,   3.93,  -4.00 ] ] ),
                (51,  [ [ -715.80,  438.10,   60.60 ], [   47.50,    3.90,  -28.60 ], [  -5.59,  -0.46,  -9.35 ], [  -2.22,   3.02,  -0.69 ], [  -0.89,  10.87,  -0.00 ], [  -2.22,   3.03,  -0.69 ] ] ),
                (52,  [ [ -690.60,  420.70,   23.60 ], [   72.60,  -48.00,   -9.80 ], [  -1.03,   0.68, -10.87 ], [   3.33,  -0.90,  -0.75 ], [   6.02,   9.13,   0.00 ], [   3.33,  -0.90,  -0.75 ] ] ),
                (53,  [ [ -634.40,  437.80,    2.60 ], [  -11.70,   62.00,  -18.70 ], [   0.57,  -3.08, -10.53 ], [   0.70,  -0.09,  -0.11 ], [ -10.80,  -2.04,   0.01 ], [   0.70,  -0.10,  -0.11 ] ] ),
                (54,  [ [ -685.40,  456.30,   14.80 ], [  -80.30,   28.90,   -4.40 ], [   0.54,  -0.19, -11.01 ], [  -0.87,   2.09,  -0.22 ], [  -3.73, -10.38,  -0.00 ], [  -0.87,   2.09,  -0.22 ] ] ),
                (55,  [ [ -713.80,  509.70,   13.10 ], [  -37.40,   33.30,    7.40 ], [  -1.21,   1.06, -10.95 ], [  -3.03,  -3.15,   1.90 ], [  -7.35,  -8.28,   0.01 ], [  -3.03,  -3.15,   1.91 ] ] ),
                (56,  [ [ -762.90,  511.80,   46.30 ], [  -17.40,  -20.30,   31.30 ], [  -5.49,  -6.40,  -7.24 ], [  -0.62,  -3.81,   1.50 ], [   8.45,  -7.22,  -0.02 ], [  -0.62,  -3.81,   1.50 ] ] ),
                (57,  [ [ -778.20,  487.30,   79.10 ], [   -7.60,  -16.30,   20.80 ], [  -3.55,  -7.65,  -7.29 ], [   6.65,   7.15,   6.52 ], [  10.11,  -4.70,   0.01 ], [   6.65,   7.16,   6.52 ] ] ),
                (58,  [ [ -800.50,  488.70,   56.50 ], [   -2.60,   28.50,  -35.30 ], [   7.23,   6.89,   5.01 ], [   6.13,   4.32,   8.27 ], [   8.49,  -5.32,  -4.94 ], [   6.13,   4.32,   8.26 ] ] ),
                (59,  [ [ -767.60,  546.30,   28.50 ], [   27.50,   29.90,   28.50 ], [   4.20,  -8.95,   5.32 ], [  -7.20,  -5.78,  -5.89 ], [   8.34,  -0.54,  -7.50 ], [  -7.20,  -5.78,  -5.89 ] ] ),
                (60,  [ [ -730.30,  549.50,   39.80 ], [   38.00,  -16.90,  -12.30 ], [  -5.37,  -9.00,  -4.10 ], [   0.16,   0.75,  -3.25 ], [  -0.99,   5.13,  -9.97 ], [   0.16,   0.75,  -3.25 ] ] ),
                (61,  [ [ -695.50,  539.60,   -1.80 ], [  -25.60,  -41.70,  -52.60 ], [   8.66,  -7.11,   1.44 ], [   5.87,   1.82,  -1.72 ], [  -6.04,  -5.82,   7.57 ], [   5.87,   1.82,  -1.71 ] ] ),
                (62,  [ [ -753.60,  538.10,   17.50 ], [  -40.40,  -47.50,    4.70 ], [   5.33,  -5.37,  -8.46 ], [  -2.66,  -0.56,  -4.61 ], [   6.83,  -5.06,   7.52 ], [  -2.66,  -0.56,  -4.61 ] ] ),
                (63,  [ [ -716.70,  466.20,   -1.60 ], [  -40.40,  -47.80,   46.20 ], [   3.56,  -8.94,  -6.13 ], [  -1.22,  -1.90,   1.22 ], [   9.07,  -1.07,   6.83 ], [  -1.22,  -1.90,   1.22 ] ] ),
                (64,  [ [ -774.90,  492.50,   14.30 ], [    4.60,  -28.40,   48.20 ], [   2.79,  -9.47,  -5.82 ], [  -2.63,  -0.59,   0.48 ], [  11.08,   2.86,   0.65 ], [  -2.63,  -0.59,   0.47 ] ] ),
                (65,  [ [ -755.00,  488.60,   92.30 ], [  -28.50,    3.50,    3.80 ], [  -1.91, -10.12,  -5.16 ], [  -0.86,  -0.12,  -0.18 ], [   0.72,  -5.33,  10.19 ], [  -0.86,  -0.12,  -0.18 ] ] ),
                (66,  [ [ -793.80,  484.40,  104.20 ], [  -32.90,    3.60,   -1.30 ], [  -0.86,  -9.98,  -5.76 ], [   6.84,   4.46,   2.66 ], [  -1.01,  -5.69,  10.00 ], [   6.84,   4.46,   2.66 ] ] ),
                (67,  [ [ -790.70,  455.50,   85.30 ], [   -2.30,  -19.40,  -39.70 ], [  11.49,  -1.41,   0.00 ], [   7.31,   8.14,   4.70 ], [  -1.27, -10.32,   5.09 ], [   7.31,   8.13,   4.70 ] ] ),
                (68,  [ [ -801.70,  464.20,   17.70 ], [   26.10,  -25.90,  -35.60 ], [   9.60,   5.94,   2.78 ], [  -7.03,   6.33,  -0.90 ], [   2.70,  -8.07,   7.92 ], [  -7.03,   6.33,  -0.90 ] ] ),
                (69,  [ [ -762.40,  459.90,   21.70 ], [   25.40,    2.00,   17.80 ], [  -0.61,  11.64,  -0.39 ], [  -1.65,   0.95,  -2.28 ], [  -6.67,  -0.03,   9.56 ], [  -1.65,   0.96,  -2.28 ] ] ),
                (70,  [ [ -762.80,  454.40,   83.50 ], [   -1.60,   18.90,   65.10 ], [  10.10,   5.74,  -1.38 ], [  -0.60,  -1.49,   6.17 ], [  -5.90,   9.68,  -2.92 ], [  -0.60,  -1.49,   6.16 ] ] ),
                (71,  [ [ -767.20,  479.00,  107.60 ], [  -24.80,   -1.90,    8.70 ], [   2.70,   6.90,   9.09 ], [  -9.96,  -2.61,   4.05 ], [  -2.90,   9.45,  -6.31 ], [  -9.96,  -2.61,   4.05 ] ] ),
                (72,  [ [ -795.10,  432.40,   92.80 ], [   -9.10,  -19.70,  -41.40 ], [ -11.22,  -1.54,   3.22 ], [  -2.62,  -8.20,  -1.68 ], [  -2.72,  10.57,  -4.41 ], [  -2.62,  -8.20,  -1.68 ] ] ),
                (73,  [ [ -769.20,  411.90,   25.90 ], [   68.40,  -16.90,  -33.00 ], [   1.09,  -9.41,   7.07 ], [   5.91,   1.76,   3.98 ], [  -5.53,  -6.68,  -8.04 ], [   5.92,   1.76,   3.98 ] ] ),
                (74,  [ [ -679.40,  419.00,   45.00 ], [   55.50,   10.70,   -0.20 ], [  -0.68,   3.92,  11.21 ], [   1.23,   4.77,   1.54 ], [   2.17, -10.99,   3.98 ], [   1.23,   4.77,   1.54 ] ] ),
                (75,  [ [ -624.80,  428.10,   30.90 ], [   37.00,    1.10,   -8.30 ], [   2.38,   3.47,  11.16 ], [   0.65,  -3.43,   0.31 ], [   1.06, -11.41,   3.32 ], [   0.65,  -3.43,   0.31 ] ] ),
                (76,  [ [ -596.10,  415.90,   17.90 ], [   13.70,  -50.50,  -10.30 ], [   1.44,  -1.92,  11.71 ], [  -2.26,  -0.16,  -0.25 ], [ -11.48,  -3.22,   0.88 ], [  -2.26,  -0.16,  -0.25 ] ] ),
                (77,  [ [ -633.70,  383.40,   10.50 ], [  -53.90,  -32.70,    2.90 ], [  -2.73,   5.47,  10.32 ], [  -3.48,   1.84,  -0.84 ], [  -5.57,   8.70,  -6.09 ], [  -3.48,   1.84,  -0.84 ] ] ),
                (78,  [ [ -721.00,  400.40,   -2.40 ], [  -84.80,  146.00,    5.60 ], [  -4.89,  -3.24,  10.54 ], [   0.98,  -2.25,   0.71 ], [   9.21,   5.13,   5.85 ], [   0.98,  -2.25,   0.71 ] ] ),
                (79,  [ [ -838.90,  413.20,   12.20 ], [   -5.80,    0.00,    0.00 ], [  -0.00,   2.57,  11.87 ], [  -4.26,  -1.08,  -0.91 ], [   0.00,  11.87,  -2.57 ], [  -4.26,  -1.08,  -0.91 ] ] ),
                (80,  [ [ -841.90,  390.30,    9.10 ], [   23.70,  -40.80,   14.10 ], [  -6.15,   0.07,  10.50 ], [  -3.54,  -1.63,  -0.61 ], [  -8.72,  -6.80,  -5.06 ], [  -3.54,  -1.63,  -0.61 ] ] ),
                (81,  [ [ -785.00,  366.40,   25.00 ], [   68.10,  -25.40,   20.50 ], [  -3.33,   0.56,  11.73 ], [   2.70,   1.81,   0.59 ], [  -4.10, -11.48,  -0.62 ], [   2.70,   1.81,   0.59 ] ] ),
                (82,  [ [ -715.10,  372.10,   29.90 ], [   45.50,    7.30,    0.60 ], [  -0.77,   3.83,  11.62 ], [   1.10,   0.20,   0.24 ], [   1.79, -11.48,   3.90 ], [   1.10,   0.20,   0.25 ] ] ),
                (83,  [ [ -676.70,  381.70,   46.00 ], [   33.50,   -1.60,    1.70 ], [  -0.56,   2.17,  12.08 ], [  -2.24,   0.19,  -0.76 ], [  -0.74, -12.08,   2.14 ], [  -2.25,   0.19,  -0.76 ] ] ),
                (84,  [ [ -624.70,  387.70,   43.50 ], [   20.20,   30.90,   -2.50 ], [  -5.85,   4.65,   9.80 ], [   0.24,  -5.21,  -0.06 ], [   8.51,  -4.93,   7.42 ], [   0.24,  -5.21,  -0.06 ] ] ),
                (85,  [ [ -620.00,  414.70,   50.80 ], [  -23.20,   24.50,    6.10 ], [  -2.40,  -5.02,  11.02 ], [   3.51,  -2.77,   1.11 ], [   8.77,   7.03,   5.11 ], [   3.51,  -2.77,   1.11 ] ] ),
                (86,  [ [ -667.60,  418.50,   51.40 ], [  -25.90,   -0.20,    2.60 ], [   1.20,   2.97,  11.96 ], [   0.14,   4.53,   0.11 ], [  -0.37,  12.02,  -2.95 ], [   0.14,   4.53,   0.11 ] ] ),
                (87,  [ [ -705.40,  416.80,   61.80 ], [  -55.10,  -17.10,    0.00 ], [  -1.49,   4.67,  11.40 ], [  -3.64,   0.65,  -0.93 ], [  -3.42,  10.87,  -4.90 ], [  -3.64,   0.65,  -0.93 ] ] ),
                (88,  [ [ -749.80,  368.90,   72.20 ], [  -21.40,  -38.20,   -0.90 ], [  -6.74,   3.55,   9.85 ], [   2.52,   2.85,  -1.54 ], [  -8.52,   4.96,  -7.61 ], [   2.52,   2.84,  -1.54 ] ] ),
                (89,  [ [ -794.90,  366.10,   65.90 ], [  -29.10,   -4.30,    9.30 ], [   1.33,   9.22,   8.31 ], [   4.72,   2.23,  -0.88 ], [  -3.91,   8.25,  -8.52 ], [   4.71,   2.23,  -0.88 ] ] ),
                (90,  [ [ -808.20,  363.30,   93.30 ], [   18.80,  -10.20,   29.80 ], [   3.64,   8.98,   7.91 ], [   1.12,  -0.05,  -0.24 ], [  -6.53,  -2.71,  10.32 ], [   1.12,  -0.05,  -0.24 ] ] ),
                (91,  [ [ -772.30,  355.10,   95.70 ], [   43.70,   -4.20,   -0.20 ], [   3.42,   9.15,   7.86 ], [  -2.39,  -0.32,   0.63 ], [   1.77,  -8.08,   9.42 ], [  -2.39,  -0.32,   0.63 ] ] ),
                (92,  [ [ -723.70,  356.50,   92.00 ], [   34.80,   14.40,   -6.20 ], [  -1.72,   8.22,   9.35 ], [   2.61,  -4.95,   1.60 ], [   4.88,  -8.24,   8.14 ], [   2.61,  -4.94,   1.59 ] ] ),
                (93,  [ [ -704.10,  380.20,   92.30 ], [   -4.50,   25.20,    1.10 ], [   6.07,   0.59,  11.02 ], [  -0.50,   1.17,  -0.66 ], [  10.81,   2.18,  -6.07 ], [  -0.50,   1.17,  -0.66 ] ] ),
                (94,  [ [ -724.40,  396.00,   92.70 ], [  -30.00,   -4.70,   -0.50 ], [  -1.56,   9.32,   8.35 ], [  -4.52,   4.97,  -1.53 ], [  -1.10,   8.28,  -9.45 ], [  -4.52,   4.97,  -1.53 ] ] ),
                (95,  [ [ -761.40,  390.00,   92.00 ], [  -26.80,   -4.40,   -1.30 ], [  -1.95,   9.30,   8.33 ], [   3.52,  -0.78,   0.05 ], [  -0.92,   8.30,  -9.48 ], [   3.52,  -0.78,   0.05 ] ] ),
                (96,  [ [ -796.40,  394.20,   92.00 ], [  -19.90,   14.60,   -1.10 ], [   5.31,   7.80,   8.45 ], [   2.73,   0.40,  -0.79 ], [   5.37,   6.54,  -9.42 ], [   2.73,   0.40,  -0.79 ] ] ),
                (97,  [ [ -808.00,  419.20,   92.00 ], [   22.20,    4.70,  -19.80 ], [   4.02,   9.89,   6.85 ], [  -2.50,   1.30,  -0.56 ], [   7.57,  -7.69,   6.66 ], [  -2.50,   1.30,  -0.56 ] ] ),
                (98,  [ [ -772.40,  406.20,   98.30 ], [   40.50,   -7.10,    9.50 ], [  -0.00,  10.20,   7.59 ], [  -5.63,   0.13,  -2.26 ], [  -3.58,  -7.28,   9.79 ], [  -5.64,   0.13,  -2.26 ] ] ),
                (99,  [ [ -743.70,  434.90,   88.30 ], [  -18.90,  -10.80,  -19.80 ], [  -7.56,  10.12,   1.74 ], [  -5.23,  -2.29,  -1.22 ], [   6.19,   6.22,  -9.25 ], [  -5.24,  -2.29,  -1.23 ] ] ),
                (100, [ [ -799.10,  394.00,   52.00 ], [  -21.00,  -22.00,  -12.00 ], [  -9.00,   4.23,   8.07 ], [   7.99,  -2.65,   4.28 ], [  -3.86,   8.50,  -8.76 ], [   7.99,  -2.64,   4.28 ] ] ),
                (101, [ [ -804.50,  342.90,   51.10 ], [   18.40,  -46.20,   24.50 ], [   5.55,   3.84,  10.93 ], [   5.00,   2.57,   0.35 ], [  -9.65,  -4.07,   7.43 ], [   5.00,   2.57,   0.35 ] ] ),
                (102, [ [ -779.40,  314.90,   94.50 ], [   43.80,  -11.60,   23.20 ], [   0.11,   9.64,   8.54 ], [  -2.57,   2.84,  -1.17 ], [  -4.80,  -7.71,   9.14 ], [  -2.57,   2.84,  -1.17 ] ] ),
                (103, [ [ -725.70,  321.90,   94.80 ], [   50.50,    2.40,   -4.10 ], [   0.22,   9.72,   8.51 ], [   2.24,  -0.48,  -0.06 ], [   1.18,  -8.49,   9.67 ], [   2.24,  -0.48,  -0.06 ] ] ),
                (104, [ [ -673.60,  342.20,   84.90 ], [   40.10,  -23.10,    1.40 ], [   4.71,   8.65,   8.43 ], [   2.31,  -3.92,   1.73 ], [  -4.45,  -7.17,   9.84 ], [   2.31,  -3.92,   1.73 ] ] ),
                (105, [ [ -654.80,  296.30,   92.00 ], [    7.30,  -43.20,    5.10 ], [   5.10,   2.22,  11.75 ], [  -3.45,  -3.20,   1.92 ], [ -11.76,  -1.38,   5.36 ], [  -3.45,  -3.19,   1.92 ] ] ),
                (106, [ [ -638.40,  213.80,   96.30 ], [  -29.60,  -31.80,    0.00 ], [  -4.73,   4.39,  11.35 ], [   4.65,   2.69,  -1.87 ], [  -8.31,   7.73,  -6.45 ], [   4.65,   2.69,  -1.87 ] ] ),
                (107, [ [ -671.50,  206.00,   94.00 ], [  -29.50,   25.20,    0.10 ], [   6.31,   7.31,   8.82 ], [   7.00,  -1.55,  -0.71 ], [   5.73,   6.71,  -9.66 ], [   7.00,  -1.56,  -0.71 ] ] ),
                (108, [ [ -693.10,  254.60,   92.00 ], [   -5.80,   31.80,    6.00 ], [   7.56,  -0.62,  10.70 ], [   0.22,  -0.48,   0.58 ], [  10.46,   3.29,  -7.20 ], [   0.22,  -0.48,   0.58 ] ] ),
                (109, [ [ -692.70,  282.10,   99.60 ], [  -11.70,   28.50,    0.00 ], [   7.22,   2.91,  10.58 ], [  -4.53,   4.59,  -0.52 ], [   9.81,   3.97,  -7.79 ], [  -4.53,   4.59,  -0.52 ] ] ),
                (110, [ [ -714.50,  303.80,  105.30 ], [  -27.80,  -10.40,    3.10 ], [  -2.20,   8.74,   9.59 ], [  -5.61,  -1.93,   0.96 ], [  -4.25,   8.71,  -8.91 ], [  -5.61,  -1.93,   0.96 ] ] ),
                (111, [ [ -732.90,  276.20,  100.40 ], [  -10.50,  -24.20,   -5.40 ], [  -3.94,  -1.07,  12.54 ], [   1.14,  -6.02,   1.63 ], [ -11.48,   5.69,  -3.12 ], [   1.15,  -6.02,   1.63 ] ] ),
                (112, [ [ -739.60,  242.40,   94.90 ], [    6.20,  -21.60,    0.00 ], [   0.16,  -3.20,  12.82 ], [   1.60,   0.58,   0.34 ], [ -11.46,  -6.54,  -0.73 ], [   1.60,   0.58,   0.33 ] ] ),
                (113, [ [ -729.90,  208.40,   92.90 ], [   31.10,  -37.00,    1.30 ], [  -0.79,   0.16,  13.21 ], [  -0.08,   2.62,   0.24 ], [ -10.29,  -8.31,  -0.51 ], [  -0.08,   2.62,   0.25 ] ] ),
                (114, [ [ -671.40,  183.60,   95.20 ], [   26.20,    4.60,  -18.30 ], [   0.71,   1.43,  13.18 ], [   0.74,   0.62,   0.01 ], [  -2.74, -12.05,   4.85 ], [   0.74,   0.62,   0.01 ] ] ),
                (115, [ [ -676.50,  223.40,   56.90 ], [  -64.00,   41.80,   -9.80 ], [   0.72,   1.43,  13.23 ], [   0.01,   0.00,   0.05 ], [   8.61,  10.08,  -1.39 ], [   0.00,   0.01,   0.05 ] ] ),
                (116, [ [ -680.80,  291.90,   29.60 ], [  125.60,   70.80,   -6.60 ], [   0.72,   1.44,  13.29 ], [  -0.00,   0.00,   0.05 ], [   7.21, -11.24,   0.87 ], [   0.00,   0.01,   0.05 ] ] ),
                (117, [ [ -714.10,  347.30,   15.30 ], [  -19.50,   82.90,  -38.40 ], [   0.72,   1.44,  13.34 ], [   0.00,   0.00,   0.06 ], [  13.34,  -1.38,   0.85 ], [   0.00,   0.01,   0.06 ] ] ),
                (118, [ [ -801.90,  404.60,   -7.10 ], [  -58.90,  -30.30,    2.30 ], [   0.73,   1.45,  13.41 ], [   0.00,   0.00,   0.05 ], [  -5.43,  12.33,  -0.98 ], [   0.00,   0.01,   0.05 ] ] ),
                (119, [ [ -808.70,  365.80,  -13.70 ], [   44.60,  -53.00,  -15.80 ], [   0.73,   1.45,  13.45 ], [  -0.00,   0.00,   0.05 ], [ -11.98,  -5.91,   2.26 ], [   0.00,   0.01,   0.05 ] ] ),
                (120, [ [ -688.00,  338.40,  -11.70 ], [   63.80,  -48.80,    8.70 ], [   0.73,   1.46,  13.53 ], [   0.00,   0.01,   0.07 ], [  -7.42, -11.30,   1.72 ], [   0.00,   0.01,   0.07 ] ] ),
                (121, [ [ -673.00,  259.50,   -3.20 ], [   81.00,   -7.10,    7.50 ], [   0.73,   1.47,  13.59 ], [   0.00,   0.01,   0.06 ], [   0.51, -13.58,   1.69 ], [   0.00,   0.01,   0.06 ] ] ),
                (122, [ [ -674.40,  198.70,   27.60 ], [ -166.70,   -5.50,   78.50 ], [   0.74,   1.48,  13.66 ], [   0.01,   0.01,   0.08 ], [  -5.65,  12.51,   0.86 ], [   0.00,   0.01,   0.08 ] ] ),
                (123, [ [ -751.20,  330.40,    5.60 ], [ -146.40,  -62.90,  102.70 ], [   0.74,   1.49,  13.77 ], [   0.01,   0.01,   0.10 ], [ -10.32,   8.91,   2.56 ], [   0.01,   0.01,   0.10 ] ] ),
                (124, [ [ -771.40,  224.30,   -0.50 ], [ -206.60,    3.00,   -5.40 ], [   0.75,   1.50,  13.86 ], [   0.01,   0.01,   0.09 ], [   1.33,  13.82,  -1.48 ], [   0.00,   0.01,   0.09 ] ] ),
                (125, [ [ -809.20,  353.00,   16.20 ], [ -189.70,   36.40,   -2.10 ], [   0.75,   1.51,  13.96 ], [   0.01,   0.00,   0.06 ], [   3.24,  13.58,  -1.62 ], [   0.00,   0.01,   0.06 ] ] ),
                (126, [ [ -838.20,  311.90,   41.30 ], [   13.30,  -30.30,   21.40 ], [   0.76,   1.51,  14.00 ], [   0.00,   0.01,   0.03 ], [  -9.33,  -9.46,   4.73 ], [   0.00,   0.00,   0.03 ] ] ),
                (127, [ [ -817.10,  288.00,   42.40 ], [   34.40,  -12.20,    0.70 ], [   0.76,   1.52,  14.02 ], [  -0.00,   0.01,   0.03 ], [  -4.27, -13.36,   1.69 ], [   0.00,   0.00,   0.03 ] ] ),
                (128, [ [ -752.40,  294.00,   78.80 ], [   68.30,    9.10,    2.00 ], [   0.76,   1.52,  14.08 ], [   0.00,   0.01,   0.05 ], [   3.16, -13.75,   1.44 ], [   0.00,   0.01,   0.05 ] ] ),
                (129, [ [ -741.40,  241.80,   54.90 ], [  -60.00,  -10.60,  -17.50 ], [   0.76,   1.53,  14.13 ], [   0.01,   0.01,   0.05 ], [   2.67,  13.98,   0.02 ], [   0.00,   0.01,   0.05 ] ] ),
                (130, [ [ -802.80,  226.00,   43.40 ], [  -39.00,   36.40,   14.80 ], [   0.77,   1.53,  14.17 ], [   0.00,   0.01,   0.04 ], [   5.91,  12.99,  -0.45 ], [   0.00,   0.00,   0.04 ] ] ),
                (131, [ [ -833.10,  266.00,   68.30 ], [  -22.10,   29.70,   29.30 ], [   0.77,   1.54,  14.21 ], [   0.00,   0.00,   0.03 ], [   3.57,  13.06,   4.65 ], [   0.00,   0.00,   0.03 ] ] ),
                (132, [ [ -822.70,  279.50,   92.00 ], [   26.60,    9.70,    0.30 ], [   0.77,   1.54,  14.23 ], [  -0.00,  -0.00,   0.02 ], [   6.18, -12.88,   1.20 ], [   0.00,   0.00,   0.02 ] ] ),
                (133, [ [ -799.30,  280.50,   92.00 ], [   19.60,   -7.30,    0.00 ], [   0.77,   1.54,  14.25 ], [  -0.00,  -0.00,   0.02 ], [  -4.80, -13.42,   1.71 ], [   0.00,   0.00,   0.02 ] ] ),
                (134, [ [ -788.60,  264.40,   92.00 ], [   -4.80,  -26.50,    0.00 ], [   0.77,   1.54,  14.27 ], [  -0.00,  -0.00,   0.02 ], [ -13.74,   4.17,   0.48 ], [   0.00,   0.00,   0.02 ] ] ),
                (135, [ [ -814.90,  246.10,   92.00 ], [  -26.10,   -9.00,    1.30 ], [   0.77,   1.54,  14.29 ], [   0.00,   0.01,   0.02 ], [  -4.20,  13.71,  -1.23 ], [   0.00,   0.00,   0.02 ] ] ),
                (136, [ [ -826.20,  223.00,   92.00 ], [   12.30,  -25.30,    0.00 ], [   0.77,   1.55,  14.31 ], [   0.00,   0.01,   0.02 ], [ -13.33,  -5.30,   1.37 ], [   0.00,   0.00,   0.02 ] ] ),
                (137, [ [ -802.80,  203.40,   92.00 ], [   45.00,    8.60,    0.00 ], [   0.77,   1.55,  14.33 ], [   0.00,   0.00,   0.03 ], [   3.72, -13.88,   1.38 ], [   0.00,   0.00,   0.03 ] ] ),
                (138, [ [ -783.70,  237.40,   92.10 ], [   32.00,   -1.80,  -16.20 ], [   0.78,   1.55,  14.36 ], [   0.01,   0.00,   0.03 ], [  -5.24, -12.85,   4.06 ], [   0.00,   0.00,   0.03 ] ] ),
                (139, [ [ -754.40,  192.40,   70.50 ], [   45.00,  -18.70,   -9.80 ], [   0.78,   1.56,  14.40 ], [   5.40,  -1.59,  -6.72 ], [  -7.57, -12.17,   2.24 ], [   5.40,  -1.59,  -6.72 ] ] ),
                (140, [ [ -680.80,  179.60,   31.60 ], [    3.90,  -10.60,  -67.30 ], [  14.18,  -2.40,  -2.32 ], [  -5.09,  -0.39, -13.72 ], [  -1.81, -14.39,  -1.36 ], [  -5.09,  -0.39, -13.72 ] ] ),
                (141, [ [ -737.10,  190.00,   35.60 ], [  -37.20,   15.70,   15.30 ], [  -4.87,  -0.09, -13.78 ], [  -6.37,   1.87,  -5.13 ], [  -4.36, -13.86,   1.59 ], [  -6.37,   1.87,  -5.13 ] ] ),
                (142, [ [ -781.90,  194.00,   48.10 ], [  -20.90,   -6.20,   -4.10 ], [  -2.21,   1.47, -14.40 ], [  -2.28,   2.28,   1.31 ], [   0.21, -14.37,  -2.80 ], [  -2.28,   2.28,   1.31 ] ] ),
                (143, [ [ -802.30,  178.10,   41.90 ], [  -11.30,  -27.20,    3.00 ], [  -7.37,   4.17, -11.97 ], [  -3.69,  -0.20,   1.59 ], [  11.43,  -3.27,  -8.59 ], [  -3.69,  -0.20,   1.59 ] ] ),
                (144, [ [ -799.10,  145.80,   39.00 ], [   22.40,  -22.60,  -18.40 ], [  -9.21,   0.32, -11.44 ], [  -1.30,  -2.70,   0.36 ], [   7.14,  11.63,  -5.43 ], [  -1.30,  -2.69,   0.36 ] ] ),
                (145, [ [ -731.20,  157.30,    2.00 ], [   50.10,  -43.00,   -3.70 ], [  -9.25,   0.32, -11.48 ], [  -0.04,  -0.00,  -0.04 ], [   2.50,  13.49,  -5.40 ], [  -0.04,   0.00,  -0.05 ] ] ),
                (146, [ [ -669.40,  112.80,    4.40 ], [  136.30,  170.20,  -47.50 ], [  -9.29,   0.32, -11.53 ], [  -0.05,   0.00,  -0.06 ], [ -10.54,   6.75,   7.91 ], [  -0.04,   0.00,  -0.06 ] ] ),
                (147, [ [ -750.70,  184.10,   17.70 ], [  -76.40,  -20.30,  -34.70 ], [  -9.34,   0.32, -11.59 ], [  -0.04,   0.00,  -0.06 ], [  -8.52,  -9.53,  -7.65 ], [  -0.05,   0.00,  -0.06 ] ] ),
                (148, [ [ -840.70,  197.30,   -8.00 ], [  -65.70,  -54.00,  -27.90 ], [  -9.38,   0.32, -11.65 ], [  -0.04,   0.00,  -0.05 ], [  -0.45, -11.86,  -9.11 ], [  -0.04,   0.00,  -0.05 ] ] ),
                (149, [ [ -770.00,  153.90,  -18.30 ], [  146.60,   96.20,   29.90 ], [  -9.43,   0.33, -11.70 ], [   9.19,   1.65,  -0.79 ], [ -14.26,   2.89,   3.77 ], [   9.19,   1.65,  -0.79 ] ] ),
                (150, [ [ -687.80,  179.20,  -17.80 ], [   93.00,    6.90,   16.70 ], [   6.85,   3.24, -13.05 ], [   9.87,   3.45,   8.75 ], [   3.05,  14.36,   3.50 ], [   9.86,   3.45,   8.75 ] ] ),
                (151, [ [ -660.00,  168.80,  104.60 ], [  -61.10,  -12.30,   52.70 ], [   6.81,   7.53,  11.29 ], [ -10.02,  -0.74,   9.08 ], [  -7.36,  12.69,  -3.93 ], [ -10.02,  -0.74,   9.08 ] ] ),
                (152, [ [ -732.50,  164.50,   92.40 ], [  -75.70,  -13.60,   -3.40 ], [  -8.83,   3.96,  11.77 ], [  -7.28,  -3.78,   0.66 ], [  -9.24,  10.65,  -5.78 ], [  -7.28,  -3.78,   0.66 ] ] ),
                (153, [ [ -785.80,  124.60,   91.40 ], [  -10.20,  -35.80,   -2.20 ], [  -8.66,  -0.00,  12.60 ], [   6.65,  -4.80,   1.10 ], [ -12.54,   2.40,  -8.41 ], [   6.65,  -4.79,   1.10 ] ] ),
                (154, [ [ -759.00,   91.20,   83.60 ], [  119.50,   -1.50,    4.30 ], [   3.31,  -5.48,  13.92 ], [   5.90,  -3.29,   1.00 ], [   3.90, -13.84,  -5.30 ], [   5.90,  -3.29,   1.00 ] ] ),
                (155, [ [ -640.80,  136.60,   98.60 ], [   46.60,   50.10,    7.60 ], [  -4.70,  -3.76,  14.19 ], [  -2.83,   0.60,   0.12 ], [   7.80, -13.29,   0.39 ], [  -2.82,   0.60,   0.12 ] ] ),
                (156, [ [ -579.70,  151.60,  100.30 ], [   19.20,  -49.70,    4.50 ], [  -4.72,  -3.77,  14.23 ], [  -0.02,  -0.02,   0.05 ], [ -11.83,  -8.30,  -5.49 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (157, [ [ -626.30,   99.70,  102.90 ], [  -45.20,   -3.50,    0.20 ], [  -4.73,  -3.79,  14.28 ], [   0.63,   2.69,   0.49 ], [  -6.14,  13.83,   3.44 ], [   0.63,   2.69,   0.49 ] ] ),
                (158, [ [ -677.50,   87.90,  102.10 ], [   -1.10,  -44.00,   15.60 ], [  -3.61,   0.95,  15.10 ], [  -0.22,  -0.88,  -0.12 ], [ -14.64,  -4.84,  -2.01 ], [  -0.22,  -0.88,  -0.12 ] ] ),
                (159, [ [ -658.60,   70.80,  123.20 ], [   64.50,  -18.30,    7.60 ], [  -4.75,  -3.80,  14.34 ], [  -0.57,  -2.37,  -0.37 ], [  -5.27, -13.73,  -5.13 ], [  -0.57,  -2.36,  -0.37 ] ] ),
                (160, [ [ -663.80,   52.80,  121.60 ], [  -54.40,  -14.30,  -27.10 ], [  -4.76,  -3.81,  14.36 ], [  -0.01,  -0.01,   0.03 ], [  -3.89,  14.85,   2.76 ], [  -0.01,  -0.01,   0.03 ] ] ),
                (161, [ [ -714.90,   78.70,   84.40 ], [  -26.00,  -24.40,  -65.40 ], [  -4.77,  -3.82,  14.40 ], [  -0.02,  -0.01,   0.05 ], [  -4.68,  12.40,   8.31 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (162, [ [ -658.40,   32.50,   83.90 ], [    2.10,  116.00,   -7.20 ], [  -4.79,  -3.83,  14.47 ], [  -0.02,  -0.01,   0.06 ], [  14.11,  -4.76,   5.01 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (163, [ [ -604.60,  113.30,   63.50 ], [  -28.80,   78.00,  -53.10 ], [  -4.81,  -3.85,  14.53 ], [  -0.02,  -0.02,   0.06 ], [  12.18,  -0.65,  10.02 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (164, [ [ -692.90,  147.10,   50.50 ], [  -30.70,   48.10,  -22.50 ], [  -4.84,  -3.87,  14.59 ], [  -0.02,  -0.01,   0.06 ], [  13.01,   4.40,   7.92 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (165, [ [ -760.40,  156.30,   47.10 ], [   12.40,  -68.30,   10.10 ], [  -4.85,  -3.88,  14.65 ], [  -0.01,  -0.01,   0.05 ], [ -12.81,  -8.18,  -4.68 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (166, [ [ -714.60,  117.30,   46.30 ], [   59.10,  -31.90,   -3.20 ], [  -4.87,  -3.89,  14.69 ], [   3.35,  -1.81,  -0.44 ], [  -9.91, -11.19,  -5.58 ], [   3.35,  -1.81,  -0.44 ] ] ),
                (167, [ [ -642.10,   94.70,   41.50 ], [   86.40,    3.30,  -36.50 ], [   2.64,  -7.93,  13.65 ], [   8.28,   8.12,  -9.98 ], [  -5.51, -13.70,  -6.17 ], [   8.28,   8.12,  -9.98 ] ] ),
                (168, [ [ -636.10,  134.80,   16.40 ], [  117.50,   19.20,  -27.50 ], [  11.57,  10.42,  -3.86 ], [   5.09,  10.44,  -9.95 ], [  14.85,   3.25,   5.13 ], [   5.09,  10.44,  -9.94 ] ] ),
                (169, [ [ -578.20,   88.00,   25.20 ], [    2.50,  -59.30,   17.40 ], [  11.61,  10.46,  -3.87 ], [   0.54, -10.37,   3.35 ], [   0.34,  13.63,   8.56 ], [   0.54, -10.38,   3.35 ] ] ),
                (170, [ [ -631.30,   39.10,   34.60 ], [ -128.70,  -21.40,  -13.20 ], [  12.62,  -9.73,   2.64 ], [  -7.55, -13.18,   3.33 ], [   9.45,   3.13,  12.72 ], [  -7.54, -13.18,   3.33 ] ] ),
                (171, [ [ -704.70,   57.70,    8.40 ], [ -125.70,  -34.50,   10.50 ], [  -3.96, -15.50,   2.62 ], [ -13.98,   6.42,  -1.88 ], [  -7.24,   0.06,  14.50 ], [ -13.98,   6.42,  -1.88 ] ] ),
                (172, [ [ -784.50,  117.60,   16.90 ], [   39.80,   42.00,  -36.30 ], [ -14.44,   7.33,  -1.76 ], [   1.81,  11.47,  -8.58 ], [   1.08,   6.87,  14.73 ], [   1.81,  11.48,  -8.57 ] ] ),
                (173, [ [ -699.80,  123.10,  -24.30 ], [   55.40, -153.80,  -40.30 ], [  -0.55,   7.65, -14.46 ], [   7.44,   0.19,  -6.84 ], [  13.82,   8.28,   2.90 ], [   7.45,   0.19,  -6.84 ] ] ),
                (174, [ [ -800.50,  108.10,  -31.70 ], [  -87.40,   47.70,   13.70 ], [  -0.55,   7.69, -14.53 ], [  -0.00,   0.03,  -0.05 ], [  -9.81, -11.69,  -6.13 ], [   0.00,   0.03,  -0.05 ] ] ),
                (175, [ [ -802.60,   73.80,  -45.80 ], [   60.70,  -32.10,   -7.20 ], [  -0.55,   7.71, -14.56 ], [  -0.00,   0.02,  -0.05 ], [   5.33,  14.05,   6.79 ], [   0.00,   0.02,  -0.05 ] ] ),
                (176, [ [ -700.50,   58.60,  -44.70 ], [   56.40,   29.00,   19.90 ], [  -0.55,   7.74, -14.63 ], [   0.00,   0.03,  -0.05 ], [  -9.95,  11.61,   6.37 ], [   0.00,   0.03,  -0.05 ] ] ),
                (177, [ [ -650.80,  100.60,  -19.70 ], [   52.40,   82.20,   48.00 ], [  -0.55,   7.77, -14.67 ], [  -2.79,  -2.28,  -0.32 ], [ -14.95,   6.17,   3.78 ], [  -2.78,  -2.28,  -0.33 ] ] ),
                (178, [ [ -659.10,  149.90,  -29.60 ], [   34.00,   79.40,   15.20 ], [  -5.71,   3.52, -15.24 ], [  -0.18,  -0.13,  -0.05 ], [ -15.05,   3.41,   6.24 ], [  -0.18,  -0.13,  -0.05 ] ] ),
                (179, [ [ -609.50,  156.90,  -23.70 ], [   15.60,  -53.60,  -14.70 ], [  -0.56,   7.80, -14.74 ], [   2.21,   1.85,   0.20 ], [  14.69,   7.51,   2.51 ], [   2.21,   1.85,   0.20 ] ] ),
                (180, [ [ -609.70,  110.70,  -23.20 ], [   40.80,  -48.10,   -0.90 ], [  -0.56,   7.82, -14.77 ], [  -2.04,  -1.11,   0.02 ], [   7.42,  14.22,   4.72 ], [  -2.04,  -1.11,   0.02 ] ] ),
                (181, [ [ -646.50,   26.10,  -11.20 ], [ -162.60,  -77.40,   15.80 ], [  -7.08,   4.22, -14.63 ], [  -2.07,   0.47,  11.78 ], [   2.94, -15.18,  -6.54 ], [  -2.07,   0.47,  11.79 ] ] ),
                (182, [ [ -611.30,   41.40,   96.10 ], [   57.30,   45.80,   -2.40 ], [  -3.08,  10.25,  13.06 ], [  -0.50,  -9.06,   9.07 ], [  11.26,  -7.87,   9.80 ], [  -0.50,  -9.06,   9.07 ] ] ),
                (183, [ [ -577.60,   74.90,   90.60 ], [   55.80,  -10.00,   15.40 ], [  -5.16,  -4.13,  15.57 ], [  -1.04,  -7.16,   1.27 ], [  -1.68, -16.12,  -4.83 ], [  -1.04,  -7.15,   1.27 ] ] ),
                (184, [ [ -584.90,   34.80,  107.90 ], [  -13.40,  -36.10,   14.60 ], [  -5.17,  -4.14,  15.60 ], [  -0.01,  -0.00,   0.03 ], [ -15.74,  -6.26,   0.66 ], [  -0.01,  -0.01,   0.03 ] ] ),
                (185, [ [ -599.80,   -4.20,  106.40 ], [  -10.50,  -58.00,  -59.30 ], [  -5.18,  -4.14,  15.63 ], [  -0.01,  -0.00,   0.03 ], [ -12.83,  10.88,   2.29 ], [  -0.01,  -0.01,   0.03 ] ] ),
                (186, [ [ -560.70,   -5.80,  103.50 ], [   22.80,   77.00,   53.00 ], [  -5.19,  -4.15,  15.66 ], [  -0.01,  -0.01,   0.04 ], [  15.79,  -3.31,   5.42 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (187, [ [ -539.50,   62.70,  103.50 ], [   25.80,   91.30,    0.00 ], [  -5.20,  -4.16,  15.71 ], [  -0.01,  -0.01,   0.04 ], [  13.65,  -9.49,   3.87 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (188, [ [ -509.90,  125.70,  103.50 ], [   40.90,   31.10,    0.00 ], [  -5.22,  -4.18,  15.75 ], [  -0.01,  -0.01,   0.03 ], [   4.21, -16.58,  -0.17 ], [  -0.01,  -0.01,   0.03 ] ] ),
                (189, [ [ -480.30,  136.30,  103.50 ], [   27.90,   -1.50,    0.00 ], [  -5.23,  -4.18,  15.77 ], [  -0.00,   0.00,   0.02 ], [  -5.84, -15.48,  -4.46 ], [  -0.01,  -0.00,   0.02 ] ] ),
                (190, [ [ -463.80,  126.60,  103.50 ], [    2.20,  -41.60,    0.00 ], [  -5.23,  -4.18,  15.79 ], [  -0.00,  -0.00,   0.02 ], [ -15.56,  -4.73,  -5.44 ], [  -0.01,  -0.01,   0.02 ] ] ),
                (191, [ [ -502.40,   82.30,  103.50 ], [  -23.60,  -27.20,    0.00 ], [  -5.24,  -4.19,  15.83 ], [  -0.01,  -0.01,   0.02 ], [ -16.28,   5.39,  -1.21 ], [  -0.01,  -0.01,   0.02 ] ] ),
                (192, [ [ -509.30,   56.80,  103.50 ], [   23.00,  -22.90,    0.00 ], [  -5.25,  -4.20,  15.84 ], [  -0.01,  -0.01,   0.02 ], [ -11.71, -10.69,  -6.68 ], [  -0.01,  -0.01,   0.02 ] ] ),
                (193, [ [ -475.50,   64.70,  103.50 ], [   44.10,   -1.90,    0.00 ], [  -5.26,  -4.21,  15.86 ], [  -0.01,  -0.01,   0.03 ], [  -5.75, -15.63,  -4.43 ], [  -0.01,  -0.01,   0.03 ] ] ),
                (194, [ [ -430.20,   49.30,  103.50 ], [   14.50,  -56.10,    1.00 ], [  -5.27,  -4.21,  15.90 ], [  -0.01,  -0.01,   0.04 ], [ -14.56,  -7.00,  -6.10 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (195, [ [ -469.90,    5.30,  104.60 ], [  -58.90,  -25.60,  -18.20 ], [  -5.28,  -4.22,  15.94 ], [  -0.01,  -0.01,   0.04 ], [  -8.97,  14.76,   1.17 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (196, [ [ -530.20,    8.80,   71.20 ], [  -89.40,    9.00,  -28.70 ], [  -5.30,  -4.24,  15.98 ], [  -0.01,  -0.01,   0.04 ], [   0.47,  16.73,   4.60 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (197, [ [ -549.80,   51.30,   74.50 ], [   56.60,   60.50,   38.30 ], [  -5.31,  -4.25,  16.02 ], [  -0.01,  -0.01,   0.04 ], [  12.78, -11.75,   1.14 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (198, [ [ -531.30,  109.40,   71.30 ], [   86.60,   49.60,    3.50 ], [  -5.32,  -4.26,  16.06 ], [  -0.02,  -0.01,   0.05 ], [   2.78, -17.18,  -1.27 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (199, [ [ -456.50,  140.50,   51.60 ], [   45.40,   11.90,   -8.30 ], [  -5.34,  -4.27,  16.11 ], [  -0.02,  -0.01,   0.05 ], [  -5.25, -16.65,  -1.17 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (200, [ [ -423.00,   81.00,   50.80 ], [  -44.50, -103.30,   -0.20 ], [  -5.36,  -4.28,  16.16 ], [  -0.02,  -0.02,   0.07 ], [ -17.24,   0.87,  -3.24 ], [  -0.02,  -0.02,   0.07 ] ] ),
                (201, [ [ -549.80,   34.30,   52.00 ], [  -12.30, -134.80,  -12.90 ], [  -5.39,  -4.31,  16.25 ], [  -0.02,  -0.02,   0.08 ], [ -16.81,  -1.21,  -5.26 ], [  -0.03,  -0.02,   0.08 ] ] ),
                (202, [ [ -465.30,  -17.50,   49.30 ], [  -46.30, -108.10,  -19.60 ], [  -5.41,  -4.33,  16.33 ], [   2.96,   6.02,  -0.20 ], [ -16.81,   4.20,  -3.77 ], [   2.95,   6.02,  -0.20 ] ] ),
                (203, [ [ -578.60,  -16.50,   42.30 ], [  -82.70,   25.50,  -15.40 ], [   0.72,   8.13,  15.84 ], [  -3.24,   8.71,  -6.08 ], [   7.05,  14.46,  -7.67 ], [  -3.24,   8.71,  -6.08 ] ] ),
                (204, [ [ -536.60,   25.20,   15.60 ], [   37.70,   57.90,  -21.80 ], [  -8.45,  14.47,   6.22 ], [  -2.87,  -6.71,   0.69 ], [  12.08,   3.54,  12.69 ], [  -2.87,  -6.71,   0.69 ] ] ),
                (205, [ [ -529.00,   93.90,   11.80 ], [   61.50,   67.50,   -4.90 ], [  -5.47,  -4.37,  16.50 ], [   1.52,  -9.66,   5.29 ], [   6.71, -16.55,   1.52 ], [   1.52,  -9.66,   5.29 ] ] ),
                (206, [ [ -468.20,  137.50,   16.60 ], [   50.60,   13.70,   -2.50 ], [  -5.48,  -4.39,  16.55 ], [  -0.02,  -0.01,   0.05 ], [  -2.85, -17.58,  -2.46 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (207, [ [ -398.80,  135.90,   16.10 ], [   54.50,  -33.90,  -12.10 ], [  -5.50,  -4.40,  16.59 ], [  -0.02,  -0.01,   0.04 ], [ -13.91, -10.03,  -5.53 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (208, [ [ -378.80,   78.80,   14.50 ], [   -8.10,  -48.90,   36.10 ], [  -5.51,  -4.41,  16.63 ], [  -0.01,  -0.01,   0.04 ], [ -12.52, -12.24,   4.45 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (209, [ [ -398.00,   36.20,   45.80 ], [  -34.00,  -45.80,   24.00 ], [  -5.52,  -4.42,  16.67 ], [  -0.01,  -0.01,   0.04 ], [ -17.64,  -2.43,   3.29 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (210, [ [ -442.20,  -17.60,   57.50 ], [  -86.50,  -40.80,   33.20 ], [  -5.54,  -4.43,  16.71 ], [  -0.02,  -0.01,   0.05 ], [ -15.53,   7.63,   5.49 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (211, [ [ -533.50,  -18.10,   72.80 ], [  -62.70,  -13.50,   25.80 ], [  -5.56,  -4.45,  16.77 ], [  -0.01,  -0.01,   0.05 ], [ -12.65,  10.76,   7.49 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (212, [ [ -559.10,  -59.00,   75.10 ], [   21.70,  -55.50,    8.30 ], [  -5.57,  -4.45,  16.81 ], [  -0.01,  -0.01,   0.04 ], [ -13.30, -10.90,  -6.13 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (213, [ [ -500.70,  -84.90,   81.00 ], [   63.70,   28.70,    0.20 ], [  -5.58,  -4.47,  16.85 ], [  -0.02,  -0.02,   0.05 ], [   0.67, -18.20,  -1.80 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (214, [ [ -421.10,  -26.60,   49.00 ], [   38.30,   45.90,  -36.30 ], [  -5.61,  -4.48,  16.92 ], [  -0.02,  -0.02,   0.05 ], [   0.68, -16.06,   8.91 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (215, [ [ -382.00,    4.70,    3.40 ], [    8.90,   68.80,  -40.90 ], [  -5.62,  -4.50,  16.96 ], [  -0.01,  -0.01,   0.04 ], [  10.76, -10.19,  10.95 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (216, [ [ -389.10,   52.20,  -22.10 ], [  -29.80,   81.20,  -22.40 ], [  -5.63,  -4.50,  17.00 ], [  -0.01,  -0.01,   0.04 ], [  16.48,   1.20,   8.24 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (217, [ [ -412.00,  118.00,  -36.80 ], [  -99.40,  115.40,   42.30 ], [  -5.65,  -4.52,  17.04 ], [  -0.02,  -0.02,   0.05 ], [  10.62,  12.72,   8.25 ], [  -0.02,  -0.01,   0.05 ] ] ),
                (218, [ [ -500.70,  120.50,  -22.30 ], [  -52.50, -122.90,   19.40 ], [  -5.67,  -4.53,  17.11 ], [  -0.02,  -0.02,   0.08 ], [ -18.33,  -2.16,  -2.14 ], [  -0.03,  -0.02,   0.08 ] ] ),
                (219, [ [ -430.50,    6.20,  -21.60 ], [ -110.00, -102.90,   20.20 ], [  -5.70,  -4.56,  17.20 ], [  -0.04,  -0.03,   0.10 ], [ -17.92,   5.26,   0.70 ], [  -0.03,  -0.03,   0.10 ] ] ),
                (220, [ [ -561.60,   93.90,  -37.50 ], [ -136.90,  -93.80,   13.70 ], [  -5.74,  -4.59,  17.31 ], [  -0.03,  -0.02,   0.10 ], [ -16.55,   8.84,   1.26 ], [  -0.03,  -0.03,   0.10 ] ] ),
                (221, [ [ -478.90,   -3.90,  -57.80 ], [   64.00, -137.50,   21.40 ], [  -5.77,  -4.61,  17.41 ], [   6.63,   6.39,  -0.25 ], [ -13.24, -11.81,  -6.52 ], [   6.62,   6.39,  -0.25 ] ] ),
                (222, [ [ -431.20,  -73.10,   -4.70 ], [  -33.20,   -2.10,   83.80 ], [   5.72,   6.46,  16.91 ], [  -1.05,  -1.02,   0.11 ], [ -11.35,  11.23,  10.26 ], [  -1.05,  -1.01,   0.11 ] ] ),
                (223, [ [ -498.10,  -26.50,    1.10 ], [ -106.30,   89.80,  -35.30 ], [  -5.81,  -4.65,  17.53 ], [  -6.98,  -6.72,   0.41 ], [  11.98,  12.58,   7.79 ], [  -6.98,  -6.72,   0.41 ] ] ),
                (224, [ [ -612.20,   36.10,   28.20 ], [  -42.00,  -52.50,   15.50 ], [  -5.84,  -4.67,  17.62 ], [  -0.02,  -0.01,   0.06 ], [ -19.10,   0.99,   0.89 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (225, [ [ -599.60,  -31.30,   28.40 ], [   13.40,  -54.20,   -7.20 ], [  -5.85,  -4.68,  17.67 ], [  -0.02,  -0.01,   0.04 ], [ -17.41,  -4.28,  -6.86 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (226, [ [ -551.10,  -68.30,   48.80 ], [   70.00,  -22.30,    2.00 ], [  -5.87,  -4.69,  17.71 ], [  -0.02,  -0.01,   0.04 ], [  -8.76, -15.91,  -6.35 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (227, [ [ -489.30,  -77.50,   38.10 ], [   74.30,   17.00,  -23.00 ], [  -5.88,  -4.70,  17.75 ], [  -0.01,  -0.01,   0.04 ], [  -8.41, -17.35,   0.23 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (228, [ [ -463.60,  -26.20,   12.80 ], [  -33.30,   65.70,   20.70 ], [  -5.90,  -4.72,  17.80 ], [  -0.02,  -0.02,   0.06 ], [  15.11,   9.01,   8.02 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (229, [ [ -582.60,   39.80,  -15.30 ], [  -97.00,  -93.30,  -56.80 ], [  -5.93,  -4.74,  17.89 ], [  -0.03,  -0.02,   0.09 ], [ -13.28,  14.18,  -0.64 ], [  -0.03,  -0.02,   0.09 ] ] ),
                (230, [ [ -535.10,  -48.00,   -7.60 ], [  163.70,  -58.30, -145.00 ], [  -5.95,  -4.76,  17.97 ], [  -0.02,  -0.02,   0.08 ], [ -18.23,  -5.43,   4.37 ], [  -0.03,  -0.02,   0.08 ] ] ),
                (231, [ [ -451.00, -130.20,   37.30 ], [  -10.40,   16.50,   63.10 ], [  -5.98,  -4.79,  18.06 ], [  -0.02,  -0.02,   0.06 ], [   6.41,   1.38,  18.49 ], [  -0.02,  -0.02,   0.06 ] ] ),
                (232, [ [ -404.70,  -90.30,   45.30 ], [   28.50,   23.80,   -5.20 ], [  -6.00,  -4.80,  18.11 ], [  -0.02,  -0.01,   0.04 ], [   3.14, -19.36,   1.56 ], [  -0.01,  -0.01,   0.04 ] ] ),
                (233, [ [ -373.40,  -43.10,   39.40 ], [    5.90,   71.90,   -6.80 ], [  -6.01,  -4.81,  18.14 ], [   0.41,   8.07,  -8.50 ], [  16.98,  -7.83,   6.23 ], [   0.41,   8.07,  -8.50 ] ] ),
                (234, [ [ -350.70,   43.10,   -5.50 ], [    0.30,  -23.00,  -85.30 ], [  -4.83,  18.19,  -6.12 ], [   3.76,   0.99, -18.90 ], [  19.16,   4.38,  -2.31 ], [   3.76,   0.98, -18.90 ] ] ),
                (235, [ [ -326.20,  -52.10,    3.90 ], [  120.10, -136.40,   45.10 ], [   1.61,  -3.65, -19.46 ], [   8.00,  -8.11,  -5.57 ], [  14.41,  13.59,  -1.41 ], [   8.00,  -8.11,  -5.57 ] ] ),
                (236, [ [ -415.40, -151.90,   68.80 ], [  -13.60,  -50.50,  -41.70 ], [  11.95,   8.73, -13.42 ], [  -5.91,   8.77,  17.85 ], [  15.71,  -9.68,   7.65 ], [  -5.91,   8.77,  17.85 ] ] ),
                (237, [ [ -353.60, -118.40,   -8.30 ], [   36.00,   49.60,  -67.80 ], [  -4.64,  15.12,  12.19 ], [  -6.76,  -0.33,  14.51 ], [  16.78,  -2.81,  10.45 ], [  -6.76,  -0.33,  14.52 ] ] ),
                (238, [ [ -394.00,  -54.90,  -43.90 ], [  -58.50,   62.10,   18.60 ], [  -3.17,   9.16,  17.42 ], [  -2.87,  -4.30,   1.63 ], [   2.17,  19.78,  -1.25 ], [  -2.88,  -4.30,   1.64 ] ] ),
                (239, [ [ -374.20,   70.30,  -64.10 ], [  -51.80,  111.20,  -42.80 ], [ -12.32,   7.26,  13.82 ], [  -4.29,  -2.14,  -1.16 ], [  11.60,  15.19,   5.49 ], [  -4.31,  -2.13,  -1.15 ] ] ),
                (240, [ [ -480.40,   88.90,  -81.20 ], [  -77.00,  -72.50,  -24.50 ], [ -12.13,   4.91,  14.92 ], [   4.11,  -0.54,   2.38 ], [ -10.25,  12.01, -12.03 ], [   4.12,  -0.54,   2.40 ] ] ),
                (241, [ [ -447.70,   23.20,  -62.70 ], [   98.20,  -53.90,  -35.40 ], [  -5.03,   5.74,  18.28 ], [   6.87,   3.25,   0.83 ], [ -16.98,  -8.10,   6.21 ], [   6.89,   3.27,   0.84 ] ] ),
                (242, [ [ -382.90, -144.70,  -29.60 ], [ -142.90,  -41.30,  110.60 ], [   1.32,  14.36,  13.47 ], [   2.82,   3.81,  -2.16 ], [ -14.51,  10.33,  -8.50 ], [   2.83,   3.84,  -2.15 ] ] ),
                (243, [ [ -481.80,  -39.40,  -48.90 ], [ -146.60,  -39.20,   12.20 ], [   1.32,  14.32,  13.43 ], [  -0.00,  -0.05,  -0.04 ], [  -0.87,  14.04, -13.76 ], [   0.00,  -0.03,  -0.03 ] ] ),
                (244, [ [ -453.60, -166.50,  -38.80 ], [  150.00,   -6.00,  -58.50 ], [   1.32,  14.27,  13.39 ], [  -0.01,  -0.05,  -0.05 ], [   0.80, -12.79,  14.84 ], [   0.00,  -0.03,  -0.03 ] ] ),
                (245, [ [ -534.60,    2.00,  -60.70 ], [ -321.90,   11.50,   28.90 ], [   1.31,  14.21,  13.34 ], [  -0.01,  -0.04,  -0.04 ], [  -1.19,  13.41, -14.16 ], [   0.00,  -0.03,  -0.03 ] ] ),
                (246, [ [ -620.80,   88.60,  -65.80 ], [ -173.30, -112.90,  -24.50 ], [   1.31,  14.17,  13.30 ], [  -0.00,  -0.04,  -0.04 ], [   3.06,  16.52,  -9.87 ], [   0.00,  -0.03,  -0.02 ] ] ),
                (247, [ [ -601.00,  -61.20,  -12.30 ], [   87.00,  -77.00,   21.40 ], [   1.31,  14.13,  13.26 ], [   9.17, -14.24,  -6.83 ], [ -15.51,  -5.72,  10.20 ], [   9.27, -14.39,  -6.89 ] ] ),
                (248, [ [ -524.30,  -72.50,  -58.30 ], [  -55.40, -200.40,  -26.20 ], [  16.67,  -9.71,   1.85 ], [   5.34,  -1.84,  -3.12 ], [  -4.21,  -6.04,  17.93 ], [   5.41,  -1.85,  -3.14 ] ] ),
                (249, [ [ -516.70, -125.70,   44.60 ], [   99.10,  -71.00,   10.10 ], [  10.37,  14.01,   8.37 ], [   1.88,   3.64,  -8.37 ], [  -5.23,  -6.48,  17.45 ], [   1.91,   3.68,  -8.47 ] ] ),
                (250, [ [ -511.50, -154.20,   13.70 ], [  -22.70,   29.20,  -59.80 ], [  15.95,   8.57,  -6.73 ], [  -3.65,  -0.47,   0.86 ], [   3.16, -14.02, -12.90 ], [  -3.69,  -0.47,   0.87 ] ] ),
                (251, [ [ -507.90, -150.60,  -52.20 ], [   77.80,   65.80,  -26.70 ], [   1.29,  14.04,  13.17 ], [  -8.54,   3.17,  11.58 ], [  16.50,  -6.06,   7.94 ], [  -8.65,   3.22,  11.73 ] ] ),
                (252, [ [ -501.90,  -69.20,  -84.50 ], [  124.80,  -47.20,  -52.70 ], [   1.29,  14.00,  13.15 ], [  -0.00,  -0.03,  -0.03 ], [  -6.41,  -9.18,  15.66 ], [  -0.00,  -0.02,  -0.02 ] ] ),
                (253, [ [ -435.40, -182.40,  -85.30 ], [  133.20,  -88.90,   11.70 ], [   1.29,  13.97,  13.11 ], [  -6.83,  -0.95,  -2.07 ], [ -13.01,  -7.62,  11.89 ], [  -6.94,  -0.95,  -2.09 ] ] ),
                (254, [ [ -364.90, -190.70,  -12.50 ], [  -54.90,  -42.00,   69.00 ], [ -11.49,  12.22,   9.26 ], [  -5.25,   1.16,  -5.56 ], [ -16.97,  -6.24,  -6.34 ], [  -5.33,   1.18,  -5.63 ] ] ),
                (255, [ [ -340.30, -163.80,   48.10 ], [  128.90,   72.90,  -11.00 ], [ -10.62,  15.74,   2.32 ], [   1.45,   0.79,   0.19 ], [   0.86,  -2.05,  19.00 ], [   1.47,   0.81,   0.20 ] ] ),
                (256, [ [ -317.00, -143.90,  -26.80 ], [  -88.40,  -30.90,  -19.80 ], [  -8.52,  13.49,  10.49 ], [   6.33,  -0.81,   5.07 ], [  -1.83,  11.02, -15.49 ], [   6.43,  -0.81,   5.16 ] ] ),
                (257, [ [ -358.50, -142.30,  -98.10 ], [  -66.10,   25.40,   -1.30 ], [   1.28,  13.87,  13.01 ], [   6.65,   0.24,   1.70 ], [   1.62,  13.39, -13.47 ], [   6.77,   0.26,   1.74 ] ] ),
                (258, [ [ -360.30,   29.80, -104.30 ], [  133.20,   -2.90,   37.90 ], [   1.28,  13.82,  12.96 ], [  -6.70,  -8.71,   0.72 ], [   0.26, -12.21,  14.54 ], [  -6.84,  -8.87,   0.75 ] ] ),
                (259, [ [ -315.60,  -96.60,  -32.70 ], [   59.80,  120.90,  165.90 ], [ -11.84,  -3.18,  14.42 ], [   4.77,  -4.90,   1.16 ], [  12.32,  -9.78,  10.52 ], [   4.87,  -4.98,   1.20 ] ] ),
                (260, [ [ -314.20,   51.50,  -89.10 ], [   14.30,  127.70,   -5.60 ], [  10.40,   3.75,  15.28 ], [  -1.05,  -1.37,  -0.07 ], [  15.80,   2.03, -10.10 ], [  -1.08,  -1.40,  -0.05 ] ] ),
                (261, [ [ -333.70,   97.10,   47.00 ], [   52.00,   38.50,   28.60 ], [ -11.07,  -4.90,  14.39 ], [  -3.38,   4.27, -19.59 ], [   6.15, -17.76,   0.41 ], [  -3.45,   4.37, -20.04 ] ] ),
                (262, [ [ -304.70,  120.30,  -32.00 ], [   62.80,  -40.80,  -60.10 ], [  -3.09,   7.47, -16.93 ], [   3.93,   2.96, -20.26 ], [  15.41,  10.71,   0.20 ], [   4.01,   3.02, -20.74 ] ] ),
                (263, [ [ -235.30,  -46.60,   17.40 ], [  156.60,  -68.60,   59.20 ], [  -7.31,  -8.51, -14.95 ], [  11.87,  -4.32,   3.21 ], [   1.54,  13.58, -12.75 ], [  12.18,  -4.43,   3.27 ] ] ),
                (264, [ [ -160.60,    8.90,   41.90 ], [   96.20,   60.70,  -14.40 ], [  13.67,  -6.24, -11.05 ], [   7.32,   9.43,  12.02 ], [   1.39,  12.62, -13.67 ], [   7.50,   9.68,  12.33 ] ] ),
                (265, [ [ -233.60,  108.70,   14.10 ], [ -196.10,  -54.60,  -35.60 ], [   1.25,  13.53,  12.70 ], [  -6.34,  10.06,  12.10 ], [   5.57,  13.67, -11.32 ], [  -6.50,  10.34,  12.42 ] ] ),
                (266, [ [ -327.10,   -7.90,   -1.40 ], [   30.20,  -91.10,   80.60 ], [   1.24,  13.49,  12.66 ], [  -1.77,  -3.98,   2.52 ], [ -18.20,  -1.26,   3.28 ], [  -1.82,  -4.08,   2.60 ] ] ),
                (267, [ [ -223.10,    8.60,   56.80 ], [   64.40,  127.00,   90.60 ], [  -1.99,   6.25,  17.29 ], [   5.46, -14.42,  -3.87 ], [  14.70,   2.28,  10.98 ], [   5.63, -14.85,  -3.98 ] ] ),
                (268, [ [ -253.90,   87.40,  105.10 ], [   -5.80,   39.60,  111.40 ], [  10.24, -13.76,   6.80 ], [  -1.00,   9.40,  -0.35 ], [  15.17,  10.40,  -1.53 ], [  -1.03,   9.68,  -0.35 ] ] ),
                (269, [ [ -243.10,  128.50,  112.80 ], [   95.00,   13.80,    1.40 ], [   1.24,  13.41,  12.59 ], [  -4.00,  12.07,   2.56 ], [   4.92, -11.96,  13.14 ], [  -4.12,  12.44,   2.65 ] ] ),
                (270, [ [ -198.40,  111.30,  103.50 ], [   29.70,  -22.50,    0.00 ], [   1.24,  13.40,  12.58 ], [  -0.01,  -0.01,  -0.01 ], [ -13.25,  -5.74,  11.43 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (271, [ [ -179.80,   86.80,  103.50 ], [   15.50,  -40.20,   -2.20 ], [   1.23,  13.39,  12.57 ], [  -0.01,  -0.01,  -0.01 ], [ -15.58,   7.24,   6.61 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (272, [ [ -181.00,   44.30,  103.40 ], [  -23.30,  -31.20,    0.00 ], [   1.23,  13.38,  12.56 ], [  -0.00,  -0.01,  -0.01 ], [  -3.20,  16.69,  -7.02 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (273, [ [ -212.60,   28.50,  103.50 ], [  -45.50,   -9.50,    0.00 ], [   1.23,  13.37,  12.55 ], [  -0.00,  -0.01,  -0.01 ], [   1.29,  13.09, -12.83 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (274, [ [ -268.30,   32.40,  103.80 ], [  -47.70,   23.20,   -5.70 ], [   1.23,  13.35,  12.54 ], [  -0.00,  -0.01,  -0.02 ], [   3.87,  12.54, -12.83 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (275, [ [ -303.20,   68.70,   92.20 ], [  -16.00,   43.30,   -4.00 ], [   1.23,  13.34,  12.52 ], [  -0.00,  -0.01,  -0.02 ], [   9.07,  14.46,  -6.70 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (276, [ [ -264.40,  113.10,   92.20 ], [   96.40,   13.40,  -16.30 ], [   1.23,  13.32,  12.51 ], [  -0.00,  -0.02,  -0.02 ], [   4.82, -12.29,  12.69 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (277, [ [ -194.80,   79.90,   66.60 ], [   66.40,  101.30,   -2.50 ], [   1.23,  13.30,  12.48 ], [  -0.00,  -0.02,  -0.02 ], [  17.04,   2.78,   6.03 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (278, [ [ -259.90,  145.40,   69.10 ], [  -50.20,   11.70,    0.70 ], [   1.23,  13.27,  12.46 ], [  -0.00,  -0.02,  -0.02 ], [   0.71,  12.60, -13.18 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (279, [ [ -332.40,  128.50,   76.00 ], [  -49.00,  -52.40,   -1.30 ], [   1.23,  13.25,  12.44 ], [  -0.00,  -0.02,  -0.02 ], [  -1.51,  16.32,  -7.96 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (280, [ [ -350.80,   61.90,   80.80 ], [   -2.80,  -49.40,    0.50 ], [   1.23,  13.24,  12.42 ], [  -0.00,  -0.02,  -0.02 ], [ -11.79,  13.85,   0.34 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (281, [ [ -310.50,    4.10,   73.40 ], [  145.80,  -25.20,   -6.40 ], [   1.22,  13.22,  12.40 ], [  -0.01,  -0.02,  -0.02 ], [  -3.10, -11.99,  13.29 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (282, [ [ -271.10,  -43.80,   94.40 ], [   16.10,  -48.30,    3.80 ], [   1.22,  13.20,  12.39 ], [  -0.00,  -0.01,  -0.01 ], [ -16.23,   6.76,   4.48 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (283, [ [ -270.90,  -77.40,  103.40 ], [   39.70,  -14.80,    1.00 ], [   1.21,  13.19,  12.38 ], [  -0.00,  -0.01,  -0.01 ], [  -7.61, -10.46,  12.70 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (284, [ [ -220.70,  -87.80,   97.40 ], [   44.90,   15.00,    0.00 ], [   1.21,  13.17,  12.36 ], [  -0.00,  -0.01,  -0.01 ], [   8.97, -10.04,  12.11 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (285, [ [ -173.40,  -67.30,   82.30 ], [  -11.80,   54.90,  -11.40 ], [   1.21,  13.16,  12.35 ], [  -0.00,  -0.02,  -0.02 ], [  12.41,  11.79,  -5.84 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (286, [ [ -229.90,  -30.60,   62.70 ], [  -49.80,    8.30,   -5.90 ], [   1.21,  13.14,  12.33 ], [  -0.00,  -0.02,  -0.01 ], [   4.00,  11.86, -13.02 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (287, [ [ -291.70,  -26.70,   39.10 ], [  -43.90,    3.80,    0.00 ], [   1.21,  13.12,  12.32 ], [  -0.00,  -0.01,  -0.01 ], [   1.14,  12.27, -13.17 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (288, [ [ -323.60,  -28.80,   45.90 ], [  -29.10,   -6.00,   15.70 ], [   1.21,  13.11,  12.31 ], [  -0.00,  -0.01,  -0.01 ], [ -10.37,  10.80, -10.04 ], [   0.00,  -0.01,  -0.00 ] ] ),
                (289, [ [ -340.40,  -46.90,   56.00 ], [   -7.70,  -28.60,   -0.30 ], [   1.21,  13.10,  12.30 ], [  -0.00,  -0.01,  -0.01 ], [  -8.34,  15.83,  -2.10 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (290, [ [ -341.70,  -83.90,   64.30 ], [    5.20,  -25.30,    4.60 ], [   1.21,  13.10,  12.29 ], [  -0.00,  -0.01,  -0.02 ], [ -16.18,   7.65,   1.97 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (291, [ [ -327.60, -149.50,   76.40 ], [   26.30,  -22.60,    3.30 ], [   1.21,  13.08,  12.27 ], [  -0.01,  -0.02,  -0.02 ], [ -14.04,  -4.99,  10.05 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (292, [ [ -277.60, -101.40,   42.90 ], [   45.10,   -4.20,    0.00 ], [   1.20,  13.05,  12.26 ], [  -0.00,  -0.02,  -0.01 ], [  -1.15, -12.20,  13.11 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (293, [ [ -231.00, -108.20,   27.80 ], [   42.40,   14.10,    0.00 ], [   1.20,  13.04,  12.24 ], [  -0.00,  -0.01,  -0.02 ], [   8.85,  -9.96,  12.00 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (294, [ [ -180.10, -109.90,   40.70 ], [   59.10,  -19.40,   39.60 ], [   1.20,  13.03,  12.23 ], [  -0.00,  -0.01,  -0.02 ], [  -6.93, -10.23,  12.96 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (295, [ [ -174.70, -116.60,   80.30 ], [  -37.40,   27.50,   45.00 ], [   1.20,  13.02,  12.21 ], [  -0.00,  -0.02,  -0.02 ], [ -11.59,  13.58,   1.25 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (296, [ [ -300.70,  -86.10,   54.30 ], [  -63.80,  -23.10,   55.90 ], [   1.20,  12.98,  12.18 ], [  -0.00,  -0.03,  -0.03 ], [ -13.97,   8.69,  -6.90 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (297, [ [ -289.00, -163.30,  110.50 ], [  -99.30,  -43.30,   98.20 ], [   1.20,  12.96,  12.15 ], [  -0.00,  -0.03,  -0.04 ], [ -14.68,   8.02,  -6.09 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (298, [ [ -323.00,  -11.30,   56.10 ], [   39.10,   49.20,   -8.90 ], [   1.20,  12.91,  12.11 ], [  -0.01,  -0.02,  -0.02 ], [  16.77,  -0.61,   5.76 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (299, [ [ -277.50,    6.40,   49.70 ], [   95.70,  -13.00,    0.00 ], [   1.18,  12.89,  12.10 ], [  -0.01,  -0.02,  -0.01 ], [  -2.18, -11.91,  12.93 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (300, [ [ -191.40,  -10.90,   80.50 ], [   62.80,   -7.70,    9.50 ], [   1.18,  12.87,  12.07 ], [  -0.00,  -0.02,  -0.02 ], [  -1.98, -11.85,  12.98 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (301, [ [ -133.30,   24.30,   81.00 ], [   42.20,   67.00,    0.00 ], [   1.18,  12.85,  12.06 ], [  -0.00,  -0.02,  -0.01 ], [  16.33,   3.31,   5.85 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (302, [ [ -125.40,   67.10,   93.50 ], [  -20.50,   21.20,   -1.40 ], [   1.18,  12.84,  12.04 ], [  -0.00,  -0.01,  -0.01 ], [   3.82,  13.92, -10.13 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (303, [ [ -181.30,   58.00,   64.50 ], [  -62.10,  -31.80,  -29.80 ], [   1.18,  12.82,  12.03 ], [  -0.00,  -0.02,  -0.02 ], [   9.05,  14.02,  -5.65 ], [   0.00,  -0.01,  -0.01 ] ] ),
                (304, [ [ -295.90,   96.80,   39.60 ], [ -106.30, -114.70,  -51.10 ], [   1.18,  12.78,  11.99 ], [  -0.00,  -0.04,  -0.03 ], [   4.27,  16.73,  -3.27 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (305, [ [ -170.50,   62.70,  -13.20 ], [  -26.90, -123.60,   24.00 ], [   1.18,  12.74,  11.96 ], [  -0.01,  -0.04,  -0.04 ], [ -11.71,  12.55,  -3.44 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (306, [ [ -274.30,   42.90,   28.40 ], [  -12.60, -108.50,  -11.40 ], [   1.17,  12.71,  11.92 ], [  -0.01,  -0.04,  -0.04 ], [  -8.86,  15.01,   1.14 ], [   0.00,  -0.02,  -0.02 ] ] ),
                (307, [ [ -142.90,  -22.50,   18.30 ], [   11.70,    9.50,   57.30 ], [   1.17,  12.66,  11.88 ], [   1.17,  -1.58,   1.19 ], [  -7.62,   0.99,  15.61 ], [   1.23,  -1.65,   1.27 ] ] ),
                (308, [ [ -150.90,  -36.10,   84.20 ], [   -5.30,  -48.90,   32.20 ], [   2.86,  10.39,  13.63 ], [   0.04,  -0.08,   0.03 ], [ -16.90,   3.99,   0.66 ], [   0.05,  -0.07,   0.04 ] ] ),
                (309, [ [ -135.30, -105.30,   94.80 ], [   -3.20,  -32.60,   55.90 ], [   1.17,  12.62,  11.85 ], [   5.63, -11.05,   0.16 ], [ -17.04,  -0.33,   3.25 ], [   5.94, -11.65,   0.17 ] ] ),
                (310, [ [ -138.40, -135.90,  107.80 ], [    4.80,  -33.70,   -0.70 ], [  10.40,  -4.95,  12.96 ], [   5.06, -12.37,  -7.66 ], [ -12.07,  -8.08,   9.47 ], [   5.34, -13.05,  -8.09 ] ] ),
                (311, [ [ -150.80, -112.60,   40.60 ], [    1.40,  115.30,    8.10 ], [   5.48,  -4.87, -15.68 ], [  -8.73,  -1.10, -16.85 ], [ -15.37,  -5.30,  -5.93 ], [  -9.22,  -1.17, -17.79 ] ] ),
                (312, [ [ -176.80,   19.30,  -24.60 ], [   67.20,  241.30,  -20.60 ], [ -10.06,  -8.08, -11.46 ], [  -5.04,  -2.23,   0.97 ], [ -14.20,  -5.24,   8.27 ], [  -5.32,  -2.36,   1.00 ] ] ),
                (313, [ [ -276.70,   70.90,  -69.10 ], [  -51.90, -162.20,   10.10 ], [  -3.19,  -9.19, -14.18 ], [   2.84,  -0.45,  -1.11 ], [  11.33, -12.94,   0.29 ], [   3.00,  -0.48,  -1.18 ] ] ),
                (314, [ [ -221.20,    2.30,  -47.30 ], [ -165.70, -117.10,   -0.90 ], [  -3.19,  -9.17, -14.15 ], [   0.01,   0.02,   0.03 ], [   1.62, -16.13,   5.61 ], [   0.00,   0.01,   0.02 ] ] ),
                (315, [ [ -283.70,  -70.80,  -59.10 ], [   66.20,  -75.00,   17.60 ], [  -3.18,  -9.15, -14.11 ], [   0.01,   0.02,   0.03 ], [  13.47,   6.99,  -7.92 ], [   0.00,   0.01,   0.02 ] ] ),
                (316, [ [ -200.90,  -32.00,  -51.70 ], [   94.90,   36.10,   50.70 ], [  -3.18,  -9.13, -14.09 ], [   7.54,   1.13,   2.34 ], [ -10.31,   6.59, -11.92 ], [   8.01,   1.20,   2.48 ] ] ),
                (317, [ [ -114.90,   21.00,  -43.50 ], [   88.80,  141.00,  123.30 ], [  12.65,  -6.78,  -9.20 ], [  10.12,   6.89,   4.80 ], [  -4.22,   8.30, -14.28 ], [  10.75,   7.32,   5.08 ] ] ),
                (318, [ [ -154.40,  135.80,  -62.60 ], [  -78.30,   -1.00,   -2.30 ], [  15.19,   6.14,  -4.52 ], [  -4.73,  11.66,   5.50 ], [  15.36,  -4.77,  -5.50 ], [  -5.02,  12.40,   5.84 ] ] ),
                (319, [ [ -274.80,  136.50,    7.90 ], [   22.10,   45.90,   70.80 ], [   3.20,  16.54,   1.79 ], [  -5.93,   5.11,   3.12 ], [  -9.70,   7.92,  11.41 ], [  -6.30,   5.46,   3.32 ] ] ),
                (320, [ [ -141.60,  114.90,   31.30 ], [   60.30,   66.50,   26.50 ], [   3.20,  16.48,   1.79 ], [  -0.01,  -0.03,  -0.00 ], [   5.81,   9.90,  12.39 ], [   0.00,  -0.02,   0.00 ] ] ),
                (321, [ [ -108.40,  101.20,   30.10 ], [   11.30,  -64.60,  -17.80 ], [   3.19,  16.47,   1.79 ], [   5.73,  -9.33,   0.72 ], [   0.02,  13.69,   9.85 ], [   6.12,  -9.95,   0.76 ] ] ),
                (322, [ [ -117.50,   36.30,   28.10 ], [  -11.80,  -88.50,    4.40 ], [  15.95,  -4.24,   3.38 ], [   7.09,  -9.12,  -1.91 ], [  -3.44,  -1.01,  16.46 ], [   7.57,  -9.72,  -2.04 ] ] ),
                (323, [ [ -123.00,  -16.90,   76.90 ], [   -5.40,  -40.50,   20.40 ], [  16.63,  -0.25,  -2.49 ], [  -0.21,   4.61,  -0.38 ], [   2.66,   9.67,  13.50 ], [  -0.21,   4.92,  -0.40 ] ] ),
                (324, [ [  -83.50,  -94.80,  124.40 ], [   -5.30,  -37.10,  -69.10 ], [  15.15,   5.23,   4.96 ], [  -0.96,   3.46,   4.71 ], [   2.79,  -9.28,  13.70 ], [  -1.00,   3.70,   5.04 ] ] ),
                (325, [ [ -114.60,   -3.80,  -22.00 ], [  -48.40,  -89.30,  -74.40 ], [  15.09,   5.21,   4.94 ], [   0.13,  -7.07,  -4.33 ], [   4.35,   1.79,  16.03 ], [   0.15,  -7.57,  -4.63 ] ] ),
                (326, [ [ -106.10,  -89.80,   24.00 ], [  -40.70,  -98.90,  -72.10 ], [  15.33,  -6.21,  -2.04 ], [   0.23,  -2.61,  -5.35 ], [  -2.24, -10.03,  13.12 ], [   0.26,  -2.79,  -5.73 ] ] ),
                (327, [ [ -134.60,  -44.80,  -44.70 ], [  -45.50,  -59.20, -129.40 ], [  15.55,  -0.63,  -5.87 ], [  -4.55,   4.11,  -6.56 ], [   1.59, -15.49,   5.83 ], [  -4.87,   4.41,  -7.04 ] ] ),
                (328, [ [ -152.20, -122.40,   -0.90 ], [ -138.70,  -24.00,  -10.70 ], [   5.62,   1.83, -15.50 ], [  -2.19,  -5.54,   0.63 ], [   7.35, -14.87,  -0.48 ], [  -2.34,  -5.95,   0.67 ] ] ),
                (329, [ [ -215.00,  -66.50,  -51.90 ], [  -95.40,  -77.80,    7.50 ], [  10.89, -11.43,  -4.96 ], [   4.44,  -4.77,   7.49 ], [   5.00,  -2.22,  15.63 ], [   4.78,  -5.13,   8.04 ] ] ),
                (330, [ [ -281.50, -141.60,  -97.10 ], [  -33.10,  -74.40,   -0.80 ], [  14.52,  -7.85,  -0.46 ], [   2.14,   2.15,   2.68 ], [  -0.17,  -1.50,  16.44 ], [   2.32,   2.30,   2.88 ] ] ),
                (331, [ [ -142.90, -169.80,  -24.70 ], [   57.10,   22.50,   62.40 ], [  14.46,  -7.83,  -0.46 ], [  -0.05,   0.02,   0.00 ], [  10.09,  12.44,  -3.77 ], [  -0.03,   0.02,   0.00 ] ] ),
                (332, [ [ -114.90,  -94.60,   56.00 ], [  -22.70,  -27.90,  103.30 ], [  14.42,  -7.80,  -0.46 ], [  -0.04,   0.02,   0.00 ], [   7.79,  13.89,   3.94 ], [  -0.03,   0.01,   0.00 ] ] ),
                (333, [ [ -165.80, -210.20,   55.70 ], [  -82.80,    8.90,  -63.60 ], [  14.38,  -7.78,  -0.46 ], [  -0.04,   0.02,   0.00 ], [   4.52, -10.09,  12.05 ], [  -0.03,   0.01,   0.00 ] ] ),
                (334, [ [ -221.80, -171.30,  -25.00 ], [ -110.60,  -77.10,  -27.50 ], [  14.34,  -7.76,  -0.46 ], [  -7.05,   1.07,  -6.62 ], [   4.40,   0.73,  15.69 ], [  -7.61,   1.16,  -7.17 ] ] ),
                (335, [ [ -212.40, -201.20,   82.90 ], [  136.90,    5.40,   26.70 ], [  -1.40,  -5.38, -15.28 ], [   1.17,  -0.16,   1.14 ], [  -3.98,  14.55,  -6.08 ], [   1.28,  -0.18,   1.23 ] ] ),
                (336, [ [ -187.10, -132.60,   25.20 ], [  -87.20,   10.40, -111.30 ], [  14.26,  -7.71,  -0.46 ], [   9.16,   4.06,  10.76 ], [  -0.57, -12.13,  10.74 ], [   9.93,   4.39,  11.65 ] ] ),
                (337, [ [ -324.00, -192.70,  -37.00 ], [  -25.10,  -36.30,   21.20 ], [  14.59,   5.03,   4.77 ], [   0.10,   4.80,   1.97 ], [  -1.03,  15.54,   4.28 ], [   0.13,   5.20,   2.14 ] ] ),
                (338, [ [ -280.20, -208.50,   37.40 ], [  -89.00,  -10.60,  119.10 ], [  14.55,   5.03,   4.77 ], [ -10.49,   2.92,  -7.65 ], [  -1.23,  14.85,  -6.15 ], [ -11.39,   3.18,  -8.31 ] ] ),
                (339, [ [ -271.90, -187.90,  119.40 ], [   42.70,   34.80,    6.60 ], [  -6.06,  10.79, -10.27 ], [  -5.42,   1.49,  -5.96 ], [  -7.05,   7.74,  12.21 ], [  -5.88,   1.63,  -6.48 ] ] ),
                (340, [ [ -252.20, -147.10,  107.10 ], [   22.70,   45.10,   -4.00 ], [  -4.00,  10.18, -11.77 ], [   2.55,   1.22,   0.64 ], [  -6.00,  12.89,   7.48 ], [   2.78,   1.34,   0.69 ] ] ),
                (341, [ [ -216.60, -117.80,   95.70 ], [   40.30,  -19.20,   -0.50 ], [  -0.92,  13.36,  -8.84 ], [   9.63,  -2.95,   8.59 ], [  -1.90,  10.78,  11.74 ], [  10.48,  -3.21,   9.35 ] ] ),
                (342, [ [ -210.30, -154.50,  112.40 ], [  -20.60,  -33.90,   26.20 ], [  14.48,   5.00,   4.74 ], [   9.65,  -5.25,   8.52 ], [  -3.00,  15.18,   4.18 ], [  10.51,  -5.71,   9.27 ] ] ),
                (343, [ [ -241.10, -217.80,  103.30 ], [  -99.30,  -50.40,  -38.80 ], [  14.45,   4.98,   4.74 ], [  -2.89,   1.23,  -6.88 ], [  12.97,   6.01,   7.20 ], [  -3.14,   1.34,  -7.50 ] ] ),
                (344, [ [ -262.00, -235.10,   20.00 ], [   83.30,   -4.40, -126.20 ], [   7.63,   7.92, -11.58 ], [  -0.53,   0.21,  -1.21 ], [  14.45,  -0.38,  -6.79 ], [  -0.57,   0.23,  -1.31 ] ] ),
                (345, [ [ -341.30, -228.90,  -31.50 ], [   21.90,  106.10,  -95.20 ], [  14.39,   4.97,   4.71 ], [   3.58,  -1.58,   8.69 ], [   7.16,  -8.20, -11.63 ], [   3.93,  -1.72,   9.48 ] ] ),
                (346, [ [ -223.00, -253.50,  -46.00 ], [    3.40,   60.90,  -83.80 ], [  14.34,   4.95,   4.69 ], [   0.66,  -1.98,  -3.36 ], [   6.75, -11.99,  -7.93 ], [   0.74,  -2.16,  -3.66 ] ] ),
                (347, [ [ -309.60, -169.80,  -81.10 ], [   21.90,  106.10,  -95.20 ], [  15.68,   1.07,  -1.89 ], [  -7.72,  -0.17,  -9.71 ], [  -0.02,  -6.82, -14.29 ], [  -8.44,  -0.18, -10.64 ] ] ),
                (348, [ [ -209.50, -219.10,  -37.80 ], [   53.10,  -46.50,  -77.10 ], [  -1.67,   4.84, -14.93 ], [  -6.57,   1.42,  -4.92 ], [  14.28,   5.25,  -4.20 ], [  -7.20,   1.56,  -5.40 ] ] ),
                (349, [ [ -173.60, -172.20,  -85.40 ], [  -26.90,   37.00,  -15.30 ], [  -1.67,   4.82, -14.90 ], [   3.86,  -5.85,   4.16 ], [ -15.11,  -0.60,  -4.38 ], [   4.24,  -6.42,   4.55 ] ] ),
                (350, [ [ -291.30,  -84.30, -106.00 ], [   97.00,  135.70,   42.80 ], [   9.54, -12.11,  -2.88 ], [   6.72,  -8.15,   5.15 ], [  -2.01,   0.14, -15.56 ], [   7.39,  -8.96,   5.63 ] ] ),
                (351, [ [ -120.90, -148.90,  -48.40 ], [   39.70,   69.50,  -18.40 ], [  10.75,  -9.49,  -6.16 ], [   0.44,   1.01,  -1.23 ], [  -8.04,  -0.66, -13.36 ], [   0.50,   1.10,  -1.36 ] ] ),
                (352, [ [ -144.10,  -35.10,  -76.70 ], [  -46.20,    4.30,   -3.90 ], [  10.72,  -9.46,  -6.13 ], [  -0.03,   0.02,   0.02 ], [   9.56,  -8.00,   9.32 ], [  -0.02,   0.02,   0.01 ] ] ),
                (353, [ [ -190.90, -106.20, -108.70 ], [  -85.30,   10.90,  -25.60 ], [  10.69,  -9.44,  -6.13 ], [  -0.03,   0.02,   0.01 ], [   5.65, -10.04,  10.41 ], [  -0.02,   0.01,   0.01 ] ] ),
                (354, [ [ -285.70,  -73.50, -145.80 ], [  -88.80,  -66.70,   -0.90 ], [  10.67,  -9.42,  -6.10 ], [  -0.03,   0.03,   0.02 ], [   5.85,  -3.27,  13.96 ], [  -0.02,   0.02,   0.01 ] ] ),
                (355, [ [ -321.80, -178.80, -111.70 ], [  -31.80,  -89.20,   89.10 ], [  10.64,  -9.39,  -6.09 ], [  -0.02,   0.02,   0.01 ], [  10.69,   6.03,   9.37 ], [  -0.02,   0.01,   0.01 ] ] ),
                (356, [ [ -371.60, -176.40, -124.40 ], [   19.20,   94.50,   -4.50 ], [  10.62,  -9.37,  -6.08 ], [  -1.15,   5.43,  -2.83 ], [  -7.74,  -5.92, -11.94 ], [  -1.26,   6.01,  -3.14 ] ] ),
                (357, [ [ -321.80,  -83.50, -118.80 ], [  -75.90,  112.20,  -24.80 ], [   7.94,   3.35, -12.73 ], [  -7.87,   9.63,  -2.48 ], [ -10.12,  -7.92,  -8.43 ], [  -8.72,  10.67,  -2.75 ] ] ),
                (358, [ [ -395.50, -163.20, -123.30 ], [  -53.40,  -89.20,    1.10 ], [  -5.72,   9.53, -10.56 ], [   1.54,   4.94,   6.19 ], [  11.70,  -0.90,  -9.86 ], [   1.72,   5.48,   6.85 ] ] ),
                (359, [ [ -473.40, -148.90, -123.50 ], [  -56.10,   65.20,  -46.90 ], [   6.94,  13.55,  -1.43 ], [   8.17,  -9.27,   2.34 ], [   2.25,  -0.33, -15.12 ], [   9.07, -10.30,   2.58 ] ] ),
                (360, [ [ -429.70, -106.40, -120.50 ], [  163.60,  -61.70,  -10.20 ], [  10.52,  -9.29,  -6.02 ], [  -2.78,  -9.61,  -6.54 ], [  14.14,   0.27,  -5.75 ], [  -3.09, -10.68,  -7.28 ] ] ),
                (361, [ [ -402.50,  -38.30, -145.20 ], [  -67.70,   81.20,    4.50 ], [   0.22,  -3.27, -14.88 ], [  -6.38,   6.43,  -3.19 ], [  -9.17, -12.03,   1.78 ], [  -7.10,   7.14,  -3.56 ] ] ),
                (362, [ [ -461.90,  -56.90, -150.00 ], [  -80.10,  -50.60,    3.70 ], [  -3.60,   3.42, -14.37 ], [  -5.47,   3.23,   2.15 ], [   6.98, -12.63,  -4.79 ], [  -6.09,   3.59,   2.38 ] ] ),
                (363, [ [ -522.70, -134.30, -113.60 ], [  -48.30,    6.50,   89.20 ], [ -11.56,   1.39,  -9.73 ], [  -0.45,   2.83,   1.70 ], [  -0.44, -14.95,  -2.51 ], [  -0.50,   3.16,   1.88 ] ] ),
                (364, [ [ -492.60,  -30.90, -110.50 ], [   -4.70,   62.60,   70.10 ], [  -4.04,   9.38, -11.16 ], [   3.44,   3.64,  -0.64 ], [ -14.31,  -4.82,   0.82 ], [   3.83,   4.07,  -0.73 ] ] ),
                (365, [ [ -562.40,  -79.20,  -83.10 ], [  -57.30,   18.40,   43.90 ], [  -4.03,   9.36, -11.13 ], [   0.01,  -0.02,   0.03 ], [  -7.38, -11.22,  -6.88 ], [   0.01,  -0.02,   0.02 ] ] ),
                (366, [ [ -607.00,   13.50,  -78.80 ], [   37.30,  101.70,   -7.80 ], [  -4.02,   9.33, -11.10 ], [   0.01,  -0.03,   0.03 ], [  -6.92,  11.73,   6.39 ], [   0.01,  -0.02,   0.02 ] ] ),
                (367, [ [ -511.10,   40.20, -112.70 ], [  103.50,  -93.30,  -48.20 ], [  -4.01,   9.30, -11.07 ], [   0.11,  -3.57,  -1.75 ], [   6.48,  12.30,   5.65 ], [   0.12,  -3.99,  -1.97 ] ] ),
                (368, [ [ -523.20,   -9.50, -122.50 ], [ -147.00,    5.30,   22.40 ], [  -3.83,   3.08, -14.15 ], [  -0.01,  -3.23,  -1.48 ], [  -2.71, -14.50,  -2.64 ], [  -0.01,  -3.61,  -1.67 ] ] ),
                (369, [ [ -588.70,  -24.10,  -95.80 ], [   34.00,  -42.60,    1.70 ], [  -4.00,   2.57, -14.17 ], [   0.65,   0.95,   0.15 ], [   7.91,  12.58,  -1.68 ], [   0.72,   1.07,   0.16 ] ] ),
                (370, [ [ -532.30,  -72.30, -120.30 ], [  111.50,  -25.50,  -41.10 ], [  -2.56,   4.94, -13.85 ], [   1.57,   2.29,   0.53 ], [   5.77,  13.30,   3.55 ], [   1.76,   2.57,   0.59 ] ] ),
                (371, [ [ -460.90,  -12.40, -123.40 ], [  176.70,   55.60,   45.00 ], [  -0.83,   7.12, -13.05 ], [   0.80,   1.00,   0.37 ], [  -7.13,  11.38,   6.42 ], [   0.90,   1.13,   0.40 ] ] ),
                (372, [ [ -427.40,   65.30,  -95.70 ], [  -43.60,   32.50,  -16.00 ], [  -0.99,   6.89, -13.12 ], [   0.27,   0.90,   0.56 ], [ -12.05,  -4.99,  -7.11 ], [   0.30,   1.02,   0.62 ] ] ),
                (373, [ [ -471.40,   54.70, -118.00 ], [  -79.10,  -33.90,    8.90 ], [  -0.50,   8.37, -12.23 ], [  -0.62,   1.29,   1.01 ], [   7.66,  -9.63,  -8.27 ], [  -0.70,   1.46,   1.13 ] ] ),
                (374, [ [ -580.40,  127.50,  -99.50 ], [   11.00,   31.00,   79.80 ], [  -3.95,   9.17, -10.90 ], [   1.47,   0.43,   0.17 ], [ -13.32,  -4.88,  -4.15 ], [   1.66,   0.50,   0.18 ] ] ),
                (375, [ [ -556.80,  144.60,  -36.30 ], [   82.20,  -21.40,    6.30 ], [   0.21,   9.40, -11.37 ], [   0.42,   0.00,  -0.03 ], [  -0.73,  11.74,   8.90 ], [   0.47,   0.02,  -0.04 ] ] ),
                (376, [ [ -519.60,  126.70, -112.70 ], [  103.30,  -10.80,  -25.10 ], [  -3.93,   9.13, -10.86 ], [  -2.19,  -0.16,   0.28 ], [   1.14,  11.64,   8.94 ], [  -2.48,  -0.17,   0.31 ] ] ),
                (377, [ [ -425.90,  145.30,  -95.20 ], [    0.10,    5.40,   86.10 ], [  -3.92,   9.10, -10.83 ], [   0.01,  -0.03,   0.03 ], [  -9.77,  -4.54,  -9.97 ], [   0.01,  -0.02,   0.02 ] ] ),
                (378, [ [ -512.40,  124.30,  -46.70 ], [  -26.40,   -6.80,  -45.50 ], [  -3.91,   9.07, -10.80 ], [   0.01,  -0.02,   0.03 ], [   4.17,  -3.30, -13.64 ], [   0.01,  -0.01,   0.02 ] ] ),
                (379, [ [ -476.90,   99.10,  -97.70 ], [   61.40,   -4.20,  -24.70 ], [  -3.90,   9.06, -10.77 ], [   0.01,  -0.02,   0.03 ], [   3.89,  11.44,   8.21 ], [   0.01,  -0.01,   0.02 ] ] ),
                (380, [ [ -371.40,  119.00, -116.30 ], [   47.50,   67.00,   21.40 ], [  -3.89,   9.03, -10.74 ], [   3.72,   1.74,   2.21 ], [  -9.50,   6.80,   8.69 ], [   4.22,   1.99,   2.49 ] ] ),
                (381, [ [ -341.60,  157.90,  -39.40 ], [ -130.40,  -15.20,  -17.30 ], [   3.62,  12.58,  -6.31 ], [   0.70,   2.51,   4.56 ], [   6.49,  -6.21, -11.42 ], [   0.79,   2.87,   5.16 ] ] ),
                (382, [ [ -401.40,  155.50,  -49.20 ], [  -44.00,   17.80,  118.30 ], [  -0.22,  14.41,  -1.67 ], [  -2.23,   0.58,   5.83 ], [ -13.79,  -0.71,  -4.45 ], [  -2.54,   0.67,   6.62 ] ] ),
                (383, [ [ -326.70,  144.80,  -18.40 ], [   68.80,  -24.00,  -65.20 ], [  -0.21,  13.25,   5.81 ], [   0.05,  -0.90,  -4.01 ], [   2.29,  -2.16,  14.12 ], [   0.06,  -1.01,  -4.56 ] ] ),
                (384, [ [ -264.80,  135.60,  -54.60 ], [   15.70,   44.40,   54.40 ], [  -0.12,  12.55,  -7.14 ], [   3.05,  -0.30,  -4.66 ], [ -13.39,   2.90,   4.57 ], [   3.48,  -0.33,  -5.31 ] ] ),
                (385, [ [ -311.00,  161.60,   22.30 ], [   30.30,  -19.60,   79.00 ], [   6.48,  12.73,  -1.86 ], [   2.95,   0.10,   1.93 ], [ -11.96,   7.06,   3.81 ], [   3.37,   0.13,   2.19 ] ] ),
                (386, [ [ -216.70,  146.00,   34.60 ], [   48.10,  -13.40,    2.20 ], [   5.72,  12.74,  -3.34 ], [  -0.36,  -0.13,  -1.13 ], [   2.21,   2.94,  13.89 ], [  -0.41,  -0.13,  -1.29 ] ] ),
                (387, [ [ -127.20,   99.80,   74.20 ], [   48.10,  -23.70,    1.30 ], [   5.78,  12.45,  -4.09 ], [   3.97,  -5.53,   2.44 ], [   1.13,   3.99,  13.71 ], [   4.55,  -6.30,   2.78 ] ] ),
                (388, [ [  -31.50,   35.80,   58.00 ], [    9.60,  -94.20,   34.00 ], [  14.11,   1.08,   1.91 ], [  -4.96,  -6.31,   9.74 ], [  -2.06,   3.67,  13.65 ], [  -5.67,  -7.21,  11.15 ] ] ),
                (389, [ [  -93.20,   47.90,   78.70 ], [  -40.40,    4.30,   -4.70 ], [   0.38,  -1.89,  14.12 ], [ -13.30,   3.09,  -2.34 ], [   3.43,  13.68,   2.08 ], [ -15.22,   3.54,  -2.68 ] ] ),
                (390, [ [  -83.60,   41.80,   21.70 ], [  -25.40,   95.30,  -36.30 ], [ -12.54,   6.62,  -1.22 ], [   1.19,   7.30,  -7.96 ], [  -1.11,  12.61,   6.50 ], [   1.37,   8.37,  -9.11 ] ] ),
                (391, [ [  -94.90,  120.00,  -16.70 ], [   54.70,   18.00,   -1.30 ], [   6.95,  12.37,   0.40 ], [  13.15,  -4.37,   2.41 ], [  10.33,   2.73,   9.34 ], [  15.08,  -5.02,   2.76 ] ] ),
                (392, [ [   -6.00,    8.70,   44.40 ], [  -86.70,  -73.70,   -3.60 ], [   8.62, -10.35,   4.25 ], [   1.39,  -7.11,  -5.37 ], [  -3.08,   2.96,  13.47 ], [   1.61,  -8.16,  -6.18 ] ] ),
                (393, [ [  -69.50,   67.10,   -5.30 ], [  -28.20,   -6.40,  -92.20 ], [   9.82,  -6.91,  -7.36 ], [  -1.75,  10.80,  -4.15 ], [  -7.45, -11.82,  -1.74 ], [  -2.01,  12.43,  -4.77 ] ] ),
                (394, [ [  -80.80,  -39.40,  -67.40 ], [  124.60,  -54.30,   39.80 ], [   4.34,  13.18,  -2.10 ], [  -6.39,   8.82,   4.76 ], [  -4.51,   3.77,  12.74 ], [  -7.36,  10.17,   5.48 ] ] ),
                (395, [ [  -47.50,   28.90,  -23.70 ], [   59.80,  -55.90,   69.30 ], [  -2.73,  13.54,   2.28 ], [   6.16,  -4.72,   4.76 ], [ -13.90,   0.67,   1.53 ], [   7.12,  -5.43,   5.49 ] ] ),
                (396, [ [  -37.50,  -11.90,  -20.60 ], [  -26.30,  -38.40,   68.20 ], [  10.22,   6.22,   7.23 ], [   7.91,  -7.09,  -3.25 ], [  -8.44,  10.83,   2.63 ], [   9.13,  -8.18,  -3.75 ] ] ),
                (397, [ [  -36.80,  -10.90,   27.50 ], [   -1.60,    7.40,   38.00 ], [  13.48,  -0.66,  -3.58 ], [   1.34,  -2.85,  -4.47 ], [   0.14,  12.27,  -6.66 ], [   1.56,  -3.28,  -5.16 ] ] ),
                (398, [ [  -42.70,   -8.00,   62.70 ], [   -6.50,    1.70,    7.10 ], [  13.47,  -0.66,  -3.57 ], [  -0.01,   0.00,   0.01 ], [   7.61,   5.38, -10.38 ], [  -0.01,   0.00,   0.00 ] ] ),
                (399, [ [  -63.30,    1.20,   96.30 ], [  -26.00,   27.20,    0.00 ], [  13.46,  -0.65,  -3.56 ], [  -0.01,   0.00,   0.00 ], [   4.17,  -9.52,  -9.28 ], [  -0.01,   0.00,   0.00 ] ] ),
                (400, [ [  -78.00,   27.60,  100.90 ], [   37.30,   13.80,    0.20 ], [  13.44,  -0.65,  -3.56 ], [  -8.40,  -4.20,   4.36 ], [  10.36,   7.70,  -5.21 ], [  -9.71,  -4.86,   5.05 ] ] ),
                (401, [ [  -31.50,   35.20,  105.20 ], [   15.60,  -22.20,    3.00 ], [  -6.21, -10.50,   6.66 ], [ -11.72,  -1.75,   3.45 ], [  -1.00,  -9.14, -10.43 ], [ -13.57,  -2.02,   3.99 ] ] ),
                (402, [ [  -34.50,   -1.20,  104.80 ], [  -10.90,  -26.50,  -24.40 ], [ -11.67,  -5.84,   4.76 ], [   4.66,  -0.01,  -4.41 ], [  -8.41,   5.84,  -9.38 ], [   5.39,  -0.02,  -5.10 ] ] ),
                (403, [ [   -5.70,    4.10,   69.80 ], [   22.60,   26.50,   -0.50 ], [   6.31, -12.00,  -2.94 ], [  11.49,   4.26,  -3.48 ], [  -5.65,  -1.98, -12.51 ], [  13.31,   4.94,  -4.03 ] ] ),
                (404, [ [   -8.00,   34.90,   79.80 ], [   -9.00,   30.20,   23.80 ], [  13.38,  -0.65,  -3.55 ], [   3.58,   5.76,  -0.31 ], [  -1.03,   2.91, -13.51 ], [   4.16,   6.67,  -0.35 ] ] ),
                (405, [ [  -24.40,   59.80,   98.70 ], [  -18.30,   17.30,   14.10 ], [  13.37,  -0.65,  -3.54 ], [  -1.58,   4.17,  -0.52 ], [   4.91,  -2.07, -12.78 ], [  -1.83,   4.84,  -0.61 ] ] ),
                (406, [ [  -52.80,   83.20,  117.00 ], [  -19.40,   32.60,   10.00 ], [   9.96,   8.38,  -4.68 ], [  -1.64,  -2.45,   7.05 ], [  -6.44,   0.93, -12.20 ], [  -1.90,  -2.84,   8.18 ] ] ),
                (407, [ [  -33.10,   94.80,  102.90 ], [   22.10,   -6.20,  -30.40 ], [   9.77,  -3.52,   9.11 ], [   0.95,  -7.69,   3.13 ], [  -4.89, -12.92,   0.37 ], [   1.10,  -8.91,   3.63 ] ] ),
                (408, [ [  -19.60,   67.80,   71.00 ], [  -12.00,   23.60,  -71.10 ], [  12.39,  -5.00,  -3.48 ], [  -0.41,   5.00,  -8.19 ], [  -5.72, -12.24,  -2.83 ], [  -0.47,   5.81,  -9.51 ] ] ),
                (409, [ [  -69.80,  126.80,   91.50 ], [  -61.80,   12.70,    9.10 ], [   6.67,  11.38,  -3.94 ], [   1.21,   0.41,   0.04 ], [   2.20,  -3.82, -13.04 ], [   1.42,   0.48,   0.04 ] ] ),
                (410, [ [ -121.30,  169.50,   92.00 ], [  -11.10,   37.00,    0.00 ], [  13.27,  -0.64,  -3.52 ], [   1.04, -10.30,   0.16 ], [  -2.09,  -5.25, -12.53 ], [   1.22, -11.98,   0.19 ] ] ),
                (411, [ [ -114.50,  225.50,   92.00 ], [   69.10,   44.90,    0.00 ], [   9.34,  -9.39,  -3.58 ], [  -8.87,  -5.78,  -0.28 ], [   0.32,   4.48, -12.97 ], [ -10.32,  -6.73,  -0.33 ] ] ),
                (412, [ [  -65.20,  213.40,   92.00 ], [   33.20,  -32.20,    0.00 ], [  -4.05, -12.45,  -4.05 ], [  -9.41,   1.68,  -1.87 ], [   6.95,  -1.10, -11.76 ], [ -10.96,   1.95,  -2.18 ] ] ),
                (413, [ [  -41.70,  168.20,   81.00 ], [    8.20,  -39.50,   -1.80 ], [  -9.73,  -6.34,  -7.24 ], [  -1.36,   1.33,  -1.36 ], [   7.72,  -2.55, -11.01 ], [  -1.58,   1.55,  -1.59 ] ] ),
                (414, [ [  -32.20,  130.30,   62.30 ], [   24.30,  -54.00,  -19.60 ], [  -7.50,  -8.97,  -7.08 ], [   1.34,  -1.11,  -0.38 ], [   6.07,  -1.01, -12.21 ], [   1.56,  -1.30,  -0.45 ] ] ),
                (415, [ [   13.30,   68.90,   25.90 ], [   74.40,   51.70,  -56.50 ], [  -8.09,  -6.79,  -8.63 ], [   6.57,  -0.10,   0.78 ], [ -10.81,   8.18,   1.49 ], [   7.67,  -0.13,   0.90 ] ] ),
                (416, [ [  -52.20,  119.90,  -39.10 ], [  -64.10,  -49.60,   36.30 ], [   7.97,  -9.91,  -4.78 ], [   7.36,   2.12,  -1.37 ], [   8.29,   1.02,  10.72 ], [   8.61,   2.47,  -1.62 ] ] ),
                (417, [ [    0.60,   40.90,   -6.30 ], [  -35.10,  -73.30,  -52.20 ], [   7.52,  -3.08, -10.84 ], [  -8.63,   0.31,   2.35 ], [   4.57, -12.15,   3.87 ], [ -10.11,   0.35,   2.75 ] ] ),
                (418, [ [  -93.60,   68.70,  -44.60 ], [  -75.00,   91.80,  -77.60 ], [  -9.58,  -9.51,   0.20 ], [  -8.60,  -3.22,   5.56 ], [  -4.44,   4.58,  11.90 ], [ -10.08,  -3.79,   6.52 ] ] ),
                (419, [ [  -10.30,   30.20,  -41.90 ], [   25.00,  -66.20,  -30.80 ], [  -9.55,  -9.48,   0.20 ], [   8.19,   6.85,   5.89 ], [  -2.35,  -0.51, -13.24 ], [   9.62,   8.03,   6.92 ] ] ),
                (420, [ [  -50.40,  -34.50,  -84.60 ], [  -46.00,  102.90,  -35.50 ], [   6.06,   3.56,  11.44 ], [   6.94,   5.55,   5.15 ], [  12.08,   0.28,  -5.85 ], [   8.16,   6.52,   6.07 ] ] ),
                (421, [ [  -41.40,   32.40,  -58.90 ], [  -91.30,   85.20,  -56.10 ], [   5.92,   2.99,  11.63 ], [  -2.72,   1.28,   0.04 ], [  13.04,   1.06,  -2.87 ], [  -3.20,   1.52,   0.06 ] ] ),
                (422, [ [ -114.30,  -19.20, -124.30 ], [  -81.00,    8.40,    9.10 ], [  -0.41,   6.85,  11.45 ], [   0.59,  -0.71,   0.09 ], [  -1.94,  11.52,  -6.46 ], [   0.71,  -0.83,   0.13 ] ] ),
                (423, [ [ -129.50,   59.30,  -86.40 ], [  -83.60,   13.00,   50.80 ], [   5.70,   2.51,  11.77 ], [   2.54,  -3.98,   0.65 ], [  -1.06,  13.11,  -2.07 ], [   3.00,  -4.70,   0.79 ] ] ),
                (424, [ [ -193.40,  -50.20, -121.30 ], [ -158.00,   10.70,   45.10 ], [   2.87,  -0.94,  12.92 ], [  -2.63,   2.52,  -0.11 ], [   0.40,  13.23,   0.91 ], [  -3.10,   2.98,  -0.11 ] ] ),
                (425, [ [ -260.70,    9.90,  -99.50 ], [  -72.10,  -15.50,  -24.10 ], [   0.38,   5.69,  11.94 ], [  -1.24,   1.76,  -0.01 ], [   4.21,  12.01,  -3.59 ], [  -1.46,   2.09,   0.01 ] ] ),
                (426, [ [ -339.40,  -58.30, -138.90 ], [  -98.20,    0.40,  -23.00 ], [   0.56,   1.96,  13.03 ], [  -1.75,   1.21,  -0.89 ], [   3.91,  12.54,  -1.11 ], [  -2.08,   1.44,  -1.04 ] ] ),
                (427, [ [ -392.70,    8.10, -127.60 ], [   58.00,   12.90,  -44.30 ], [  -2.91,   7.53,  10.38 ], [  -0.09,   0.88,  -0.16 ], [   0.70,  -7.63,  10.68 ], [  -0.11,   1.05,  -0.17 ] ] ),
                (428, [ [ -318.60,  -12.70, -117.30 ], [   82.30,   72.00,   33.30 ], [  -0.08,   4.36,  12.37 ], [   0.82,  -1.67,   0.99 ], [  11.05,  -4.97,   5.02 ], [   0.97,  -1.97,   1.19 ] ] ),
                (429, [ [ -400.20,   75.30, -134.70 ], [  185.90,  159.10,   45.20 ], [  -2.86,   5.38,  11.56 ], [  -0.07,   1.98,  -1.16 ], [   8.95,  -6.98,   6.47 ], [  -0.09,   2.36,  -1.36 ] ] ),
                (430, [ [ -297.10,   32.20, -114.40 ], [  170.70,  159.00,   13.00 ], [  -0.38,   8.26,  10.06 ], [  -1.28,   1.68,  -1.81 ], [  10.70,  -3.35,   6.62 ], [  -1.53,   2.01,  -2.14 ] ] ),
                (431, [ [ -183.70,   12.90,  -89.70 ], [   13.00,  150.40,   -5.30 ], [  -5.33,   8.76,   7.95 ], [  -3.94,  -4.23,  -0.25 ], [   8.91,   7.45,   5.78 ], [  -4.71,  -5.05,  -0.28 ] ] ),
                (432, [ [ -336.80,   87.50, -103.70 ], [  -17.50,   86.10,   36.10 ], [  -8.00,  -1.41,  10.03 ], [   4.82,   0.55,   1.73 ], [   8.89,   2.44,   9.03 ], [   5.77,   0.68,   2.09 ] ] ),
                (433, [ [ -251.90,   87.50,  -90.80 ], [   85.40, -108.30,   36.10 ], [   1.28,   5.52,  11.55 ], [   2.95,  -0.08,   1.37 ], [ -10.47,  -6.21,   4.16 ], [   3.53,  -0.09,   1.66 ] ] ),
                (434, [ [ -181.50,  101.00,  -68.40 ], [   88.50,  -70.20,   49.50 ], [  -0.99,  -0.33,  12.80 ], [  -1.06,  -4.88,   0.18 ], [  -3.84, -12.21,   1.05 ], [  -1.27,  -5.85,   0.23 ] ] ),
                (435, [ [ -170.30,   54.60, -111.80 ], [   73.70,  -24.10,  -48.10 ], [  -0.96,  -4.33,  12.02 ], [  -2.34,   5.28,  -2.64 ], [ -10.28,  -7.63,  -0.60 ], [  -2.81,   6.35,  -3.16 ] ] ),
                (436, [ [ -123.20,   76.70,  -98.70 ], [   73.80,   14.70,   83.70 ], [  -5.26,   8.63,   7.84 ], [  -4.11,   1.60,  -5.47 ], [  -3.10,  -8.59,   8.96 ], [  -4.95,   1.91,  -6.57 ] ] ),
                (437, [ [  -53.70,  112.30,  -68.10 ], [ -112.20,   97.90,  113.50 ], [  -9.05,  -8.99,   0.20 ], [  -1.90,  -8.86,  -3.85 ], [   5.04,  -4.93,  10.63 ], [  -2.30, -10.67,  -4.63 ] ] ),
                (438, [ [  -42.10,  136.80,   15.40 ], [  139.60,   -1.80,  110.30 ], [  -9.03,  -8.96,   0.19 ], [   2.10,  -1.42,  -0.53 ], [   0.16,  -5.68, -11.38 ], [   2.52,  -1.73,  -0.64 ] ] ),
                (439, [ [    8.20,   95.10,   34.30 ], [  -24.80,   -1.90,   92.70 ], [  -5.35, -11.49,  -0.74 ], [   0.32,   0.74,   3.30 ], [  10.88,  -5.37,   3.73 ], [   0.37,   0.88,   3.99 ] ] ),
                (440, [ [   16.90,  122.10,   92.60 ], [   26.30,    5.60,   42.10 ], [  -7.67,  -8.20,   5.88 ], [  -1.66,   0.65,  -1.50 ], [   7.57,  -9.56,  -3.46 ], [  -2.01,   0.77,  -1.81 ] ] ),
                (441, [ [   25.30,  149.50,   92.30 ], [  -11.10,   35.90,    0.00 ], [  -8.98,  -8.92,   0.19 ], [  -0.75,  -0.41,  -3.28 ], [   1.91,  -5.55,  11.22 ], [  -0.91,  -0.50,  -3.96 ] ] ),
                (442, [ [   -9.30,  181.40,   93.40 ], [  -17.10,   39.40,    0.00 ], [  -8.97,  -8.90,   0.19 ], [   0.35,   9.47,  -2.05 ], [   2.00,  -4.14,  11.77 ], [   0.42,  11.45,  -2.48 ] ] ),
                (443, [ [   -3.30,  218.30,   94.20 ], [   39.10,   16.20,    0.00 ], [  -8.33,   8.77,  -3.63 ], [   8.92,   8.72,  -0.98 ], [  -5.40,   1.69,  11.29 ], [  10.79,  10.55,  -1.19 ] ] ),
                (444, [ [   33.30,  192.20,   83.10 ], [   42.50,  -69.90,  -17.10 ], [  10.72,   6.53,  -1.13 ], [  12.06,  -2.03,   2.94 ], [   2.40,  -1.80,  12.25 ], [  14.60,  -2.45,   3.56 ] ] ),
                (445, [ [   29.00,  113.80,   73.30 ], [   13.20,   -7.90,  -43.20 ], [  11.33,   4.84,   2.53 ], [   0.73,  -5.94,  -0.50 ], [   4.14, -11.41,   3.31 ], [   0.89,  -7.19,  -0.61 ] ] ),
                (446, [ [   53.50,  137.40,   54.40 ], [   22.90,   17.80,    6.30 ], [  12.12,  -3.29,  -0.12 ], [   0.39,  -1.17,  -2.28 ], [   6.30,   7.07,  -8.25 ], [   0.48,  -1.41,  -2.77 ] ] ),
                (447, [ [   57.60,  178.10,   62.90 ], [   -9.40,   32.20,   -3.10 ], [  12.11,   2.57,  -2.03 ], [  -0.01,   2.18,  -0.71 ], [  -1.50,  -2.38, -12.23 ], [  -0.01,   2.65,  -0.87 ] ] ),
                (448, [ [   40.10,  195.20,   59.60 ], [   -3.40,   25.80,   -0.90 ], [  12.10,   2.57,  -2.03 ], [  -0.01,  -0.00,   0.00 ], [  -2.05,   0.34, -12.36 ], [  -0.01,   0.00,   0.00 ] ] ),
                (449, [ [   57.30,  231.00,   78.90 ], [  -15.50,   11.50,   19.40 ], [  12.09,   2.57,  -2.03 ], [  -0.01,  -0.00,   0.00 ], [   1.41,   4.39, -11.64 ], [  -0.01,   0.00,   0.00 ] ] ),
                (450, [ [   34.30,  245.80,   93.80 ], [  -46.90,    3.20,    0.00 ], [  12.07,   2.57,  -2.02 ], [  -0.01,  -0.00,   0.00 ], [  11.71,  -2.83,  -3.38 ], [  -0.01,   0.00,   0.00 ] ] ),
                (451, [ [    5.80,  255.40,   92.60 ], [    2.30,    2.40,   34.80 ], [  12.06,   2.57,  -2.02 ], [  -4.14,  -6.59,  -1.04 ], [  -2.76,  12.07,  -1.70 ], [  -5.03,  -8.01,  -1.27 ] ] ),
                (452, [ [   37.80,  263.40,  105.50 ], [   25.30,    1.40,    0.00 ], [   3.38, -11.25,  -4.21 ], [  -9.58,  -6.24,  -0.76 ], [   2.52,   4.36, -11.42 ], [ -11.64,  -7.58,  -0.92 ] ] ),
                (453, [ [   72.90,  262.00,   95.30 ], [   27.10,  -36.00,  -29.30 ], [  -7.12,  -9.62,  -3.48 ], [  -3.75,   0.11,   0.48 ], [  -0.52,   2.46, -12.21 ], [  -4.56,   0.13,   0.57 ] ] ),
                (454, [ [   74.30,  240.00,   41.90 ], [   65.30,  -11.40,  -10.10 ], [   0.06, -11.97,  -3.42 ], [   7.15,  -0.12,   0.45 ], [   1.32,   2.88, -12.04 ], [   8.71,  -0.16,   0.54 ] ] ),
                (455, [ [  104.20,  247.50,   91.20 ], [   23.80,   44.00,    5.50 ], [   7.18,  -9.81,  -2.58 ], [   2.79,  12.36,   2.88 ], [  -3.76,  -2.77, -11.51 ], [   3.41,  15.07,   3.51 ] ] ),
                (456, [ [  105.20,  286.30,   78.80 ], [  -59.10,   14.70,   15.40 ], [   6.69,  10.28,   1.84 ], [  -4.20,  11.14,   0.08 ], [   1.14,   2.57, -12.08 ], [  -5.12,  13.58,   0.10 ] ] ),
                (457, [ [   79.00,  257.00,   53.90 ], [  -35.60,   -5.00,  -14.30 ], [  -1.41,  12.02,  -2.63 ], [  -5.03,   0.93,  -0.98 ], [   4.13,  -1.99, -11.51 ], [  -6.14,   1.14,  -1.19 ] ] ),
                (458, [ [   20.00,  252.70,   43.90 ], [  -73.70,  -16.00,   -8.60 ], [  -2.90,  12.01,   0.43 ], [  -4.01,  -3.06,  -0.71 ], [   1.04,   0.70, -12.30 ], [  -4.90,  -3.72,  -0.87 ] ] ),
                (459, [ [  -18.80,  195.30,   54.90 ], [  -24.50,  -54.20,  -37.40 ], [ -10.04,   5.15,  -4.98 ], [  -3.78,  -3.64,  -2.86 ], [   5.83,   1.93, -10.71 ], [  -4.63,  -4.44,  -3.50 ] ] ),
                (460, [ [  -85.50,  167.30,   15.40 ], [  -18.50,  -23.90,  -52.20 ], [ -10.02,   5.14,  -4.96 ], [   5.22,  -1.93,  -3.52 ], [   4.78,   5.04, -10.16 ], [   6.38,  -2.36,  -4.33 ] ] ),
                (461, [ [   -9.90,  178.40,   10.90 ], [   70.50,  -60.80,   38.30 ], [   0.64,   1.20, -12.20 ], [   3.58,  -3.04,  -3.48 ], [   3.47,  11.76,  -0.64 ], [   4.38,  -3.72,  -4.28 ] ] ),
                (462, [ [   40.80,  131.00,   22.30 ], [   93.00,   82.10,  -75.60 ], [  -3.07,  -0.91, -11.82 ], [  -0.17,  -2.60,   0.37 ], [  -4.81,  11.24,  -0.76 ], [  -0.21,  -3.20,   0.44 ] ] ),
                (463, [ [   15.50,  223.00,   15.30 ], [   66.30,   73.50,  -18.10 ], [   0.99,  -4.10, -11.46 ], [   5.85,  -2.68,   3.02 ], [  -9.29,   7.17,  -3.38 ], [   7.19,  -3.30,   3.69 ] ] ),
                (464, [ [   84.50,  187.60,    6.70 ], [   63.20,  -22.90,   74.10 ], [   8.35,  -6.35,  -6.19 ], [   1.95,   3.39,   0.57 ], [   7.47,   9.60,  -0.52 ], [   2.40,   4.17,   0.68 ] ] ),
                (465, [ [   81.80,  195.00,   99.00 ], [   85.20,  -17.50,   -8.30 ], [   4.26,   3.34, -10.87 ], [  -5.95,   3.42,  -2.24 ], [   6.90,   9.29,   3.68 ], [  -7.33,   4.21,  -2.77 ] ] ),
                (466, [ [  144.30,  215.80,   88.40 ], [   56.60,   35.60,  -14.80 ], [  -2.94,   2.55, -11.47 ], [  -6.00,  -0.83,   0.66 ], [  -4.28,  10.83,   3.34 ], [  -7.40,  -1.02,   0.80 ] ] ),
                (467, [ [  178.40,  287.40,   86.20 ], [    6.50,   74.90,   -2.80 ], [  -7.52,   1.67,  -9.31 ], [  -2.09,  -5.60,   3.23 ], [  -9.09,   2.44,   7.58 ], [  -2.58,  -6.92,   3.98 ] ] ),
                (468, [ [  160.50,  350.90,   83.50 ], [  -55.40,   54.50,    2.60 ], [  -7.51,  -7.89,  -5.18 ], [   4.12,  -6.89,   3.11 ], [  -3.11,  -4.19,  10.88 ], [   5.08,  -8.53,   3.83 ] ] ),
                (469, [ [   79.50,  372.50,   92.00 ], [  -94.70,  -10.00,   12.80 ], [   1.80, -11.41,  -3.37 ], [   8.18,   5.30,  -0.77 ], [   2.89,  -2.97,  11.29 ], [  10.12,   6.55,  -0.96 ] ] ),
                (470, [ [   56.20,  296.40,   44.30 ], [    5.10,  -77.10,  -40.60 ], [   8.64,   4.24,  -7.16 ], [  -0.89,   8.18,  -2.63 ], [   8.30,  -3.68,   7.84 ], [  -1.10,  10.13,  -3.27 ] ] ),
                (471, [ [   86.50,  241.90,   24.50 ], [   53.00,  -52.30,    0.20 ], [   2.60,   7.44,  -9.01 ], [  -3.05,  -2.57,  -2.05 ], [   3.89,   8.81,   7.11 ], [  -3.77,  -3.18,  -2.56 ] ] ),
                (472, [ [  152.10,  205.80,   56.00 ], [   63.60,   -8.30,   10.00 ], [   3.21,  -2.16, -11.30 ], [  -0.12,  -2.70,  -1.03 ], [   3.42,  11.35,  -1.45 ], [  -0.15,  -3.35,  -1.29 ] ] ),
                (473, [ [  198.40,  215.10,   56.00 ], [   38.20,   22.50,   -1.40 ], [   2.66,  -0.84, -11.59 ], [  -0.74,   1.84,  -0.19 ], [  -3.99,  11.03,  -2.14 ], [  -0.92,   2.29,  -0.25 ] ] ),
                (474, [ [  231.60,  243.60,   55.70 ], [   23.70,   41.90,   -1.20 ], [   1.74,   1.49, -11.69 ], [  -0.06,   0.95,   0.02 ], [  -8.93,   7.83,  -0.84 ], [  -0.08,   1.18,   0.02 ] ] ),
                (475, [ [  254.20,  308.60,   53.70 ], [  -36.30,   58.90,    0.00 ], [   3.07,   0.19, -11.48 ], [   1.44,  -1.14,   0.41 ], [  -9.01,  -7.25,  -2.71 ], [   1.80,  -1.42,   0.49 ] ] ),
                (476, [ [  201.60,  339.10,   45.30 ], [  -53.50,   14.40,  -12.40 ], [   4.61,  -0.82, -10.90 ], [   1.75,  -3.67,   9.37 ], [  -0.90, -11.83,   0.08 ], [   2.19,  -4.57,  11.67 ] ] ),
                (477, [ [  225.60,  364.90,   17.70 ], [   84.00,  -17.10,  -36.30 ], [   6.57,  -7.03,   6.90 ], [  -1.74,  -1.93,  11.58 ], [   0.08,  -9.62,  -6.91 ], [  -2.17,  -2.41,  14.45 ] ] ),
                (478, [ [  262.20,  310.80,   24.60 ], [   19.90,  -60.50,    3.10 ], [   0.45,  -3.90,  11.14 ], [  -3.49,   2.51,   2.39 ], [  -9.02,  -7.61,  -0.58 ], [  -4.35,   3.12,   3.00 ] ] ),
                (479, [ [  266.70,  240.60,   17.50 ], [  -54.80,  -92.30,    1.00 ], [  -0.30,  -2.05,  11.61 ], [  -0.74,   1.97,   0.30 ], [ -10.99,   4.18,   0.80 ], [  -0.93,   2.46,   0.39 ] ] ),
                (480, [ [  191.50,  210.20,   17.50 ], [  -54.20,  -18.80,   19.40 ], [  -1.03,   0.06,  11.71 ], [   0.91,   2.69,  -0.23 ], [  -7.82,   8.71,   1.12 ], [   1.14,   3.36,  -0.27 ] ] ),
                (481, [ [  126.50,  239.40,   20.10 ], [  -85.50,   -4.10,  -17.80 ], [   1.33,   3.26,  11.19 ], [   1.41,   1.92,  -0.33 ], [   3.78,  10.85,  -2.36 ], [   1.77,   2.40,  -0.39 ] ] ),
                (482, [ [   27.70,  256.90,  -12.00 ], [   25.60,   34.70,   43.20 ], [   1.31,   3.24,  11.15 ], [  -0.00,  -0.01,  -0.03 ], [   8.36,   2.10,   7.89 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (483, [ [   83.20,  288.30,   -1.00 ], [   33.20,   -4.80,   -3.70 ], [   1.31,   3.23,  11.13 ], [  -0.00,  -0.01,  -0.03 ], [  -1.61, -11.04,   3.41 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (484, [ [  213.40,  266.00,    3.10 ], [   68.30,  112.60,   23.20 ], [   1.31,   3.23,  11.08 ], [  -0.00,  -0.01,  -0.03 ], [  11.48,  -0.96,   1.46 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (485, [ [  188.20,  299.90,   26.40 ], [  -48.40,  -18.80,    4.40 ], [   1.31,   3.21,  11.06 ], [  -0.00,  -0.01,  -0.02 ], [  -2.92,  10.90,  -2.63 ], [   0.00,  -0.00,  -0.01 ] ] ),
                (486, [ [  130.20,  290.80,   30.50 ], [  -44.60,   71.50,   28.90 ], [   1.31,   3.21,  11.04 ], [  -0.01,  -0.01,  -0.02 ], [   5.06,  10.37,  -0.87 ], [   0.00,  -0.00,  -0.01 ] ] ),
                (487, [ [  136.70,  339.40,   47.80 ], [ -100.80,   75.50,   26.60 ], [   1.30,   3.20,  11.02 ], [  -0.01,  -0.01,  -0.03 ], [   3.35,  10.74,  -2.62 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (488, [ [   47.80,  248.50,   21.60 ], [  -98.10,   56.50,    2.30 ], [   1.30,   3.19,  10.97 ], [  -0.00,  -0.01,  -0.04 ], [   4.81,   9.88,  -3.40 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (489, [ [   59.30,  327.60,   28.10 ], [   34.80,  103.20,  -11.30 ], [   1.29,   3.18,  10.94 ], [  -0.00,  -0.01,  -0.03 ], [  11.37,  -1.46,  -0.44 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (490, [ [  158.50,  381.30,   13.00 ], [  145.50,   -5.80,  -42.40 ], [   1.29,   3.17,  10.90 ], [  -0.01,  -0.01,  -0.04 ], [  -1.38, -10.74,   3.63 ], [   0.00,  -0.01,  -0.03 ] ] ),
                (491, [ [  240.70,  304.70,   -6.30 ], [   49.30,  -67.70,  -27.10 ], [   1.28,   3.15,  10.86 ], [  -0.01,  -0.01,  -0.03 ], [ -10.21,  -2.59,   4.31 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (492, [ [  240.40,  234.90,  -17.20 ], [  -74.10,  -88.50,   16.10 ], [   1.28,   3.15,  10.83 ], [  -0.00,  -0.01,  -0.03 ], [  -7.57,   8.36,  -1.27 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (493, [ [  173.00,  289.90,  -12.50 ], [   56.50,  158.10,    0.10 ], [   1.28,   3.14,  10.80 ], [  -0.00,  -0.01,  -0.04 ], [  11.31,  -0.44,  -0.15 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (494, [ [  109.50,  348.10,   -9.70 ], [  -74.10,  -88.50,   16.10 ], [   1.28,   3.13,  10.76 ], [  -0.01,  -0.01,  -0.02 ], [  -7.52,   8.31,  -1.26 ], [   0.00,  -0.00,  -0.01 ] ] ),
                (495, [ [   88.40,  320.90,   -6.10 ], [  -71.10,  -68.40,   21.40 ], [   1.27,   3.12,  10.75 ], [  -0.01,  -0.01,  -0.01 ], [  -7.43,   8.34,  -1.49 ], [   0.00,  -0.00,  -0.01 ] ] ),
                (496, [ [   61.80,  280.20,  -37.60 ], [   69.10, -114.80,  -49.00 ], [   1.27,   3.12,  10.73 ], [  -0.00,  -0.01,  -0.03 ], [ -10.27,  -1.14,   4.45 ], [   0.00,  -0.00,  -0.02 ] ] ),
                (497, [ [  132.90,  264.70,  -28.90 ], [  -44.70, -165.90,  -57.30 ], [   1.26,   3.11,  10.70 ], [  -0.00,  -0.01,  -0.04 ], [  -7.20,   8.23,   2.46 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (498, [ [  147.20,  179.70,   61.30 ], [   91.20,  -36.60,   30.40 ], [   1.26,   3.10,  10.65 ], [  -0.01,  -0.01,  -0.04 ], [  -1.90, -10.20,   4.13 ], [   0.00,  -0.01,  -0.02 ] ] ),
                (499, [ [  238.20,  177.20,   65.00 ], [   18.30,  -82.80,  -18.10 ], [   1.25,   3.09,  10.62 ], [  -6.46,  -1.97,  -8.08 ], [ -10.53,   2.18,   2.87 ], [  -8.32,  -2.53, -10.38 ] ] ),
                (500, [ [  197.80,  161.40,   50.30 ], [   67.60,   35.60, -126.70 ], [ -10.39,  -0.45,  -3.92 ], [   0.94,  -2.52, -10.53 ], [  -2.01,  10.32,   3.58 ], [   1.21,  -3.24, -13.56 ] ] ),
                (501, [ [  254.20,  183.80,   49.30 ], [  -11.80,   23.50,   62.10 ], [   3.91,  -1.89, -10.20 ], [  10.29,  -1.14,   3.54 ], [   0.06,  -1.92, -10.91 ], [  13.26,  -1.47,   4.56 ] ] ),
                (502, [ [  277.70,  254.80,   52.00 ], [   -1.10,   48.50,  -21.40 ], [  10.34,  -2.74,   2.80 ], [   0.61,   0.19,  -0.28 ], [   1.53,  -7.63,  -7.85 ], [   0.80,   0.24,  -0.38 ] ] ),
                (503, [ [  291.40,  209.50,   13.10 ], [  -39.30,  -47.20,   -6.60 ], [   5.80,  -1.63,  -9.23 ], [  -3.43,   0.14,  -6.40 ], [   7.81,  -5.38,   5.63 ], [  -4.42,   0.18,  -8.27 ] ] ),
                (504, [ [  227.20,  207.90,    6.30 ], [  -93.70,  -69.70,   -8.90 ], [   3.44,  -2.43, -10.16 ], [   0.12,   0.05,   0.07 ], [   6.29,  -8.07,   4.03 ], [   0.16,   0.06,   0.08 ] ] ),
                (505, [ [  186.40,  164.40,   16.90 ], [  -71.70,  -61.20,  -10.00 ], [   5.78,  -1.62,  -9.19 ], [   1.27,   0.44,   0.54 ], [   7.54,  -6.04,   5.21 ], [   1.65,   0.57,   0.68 ] ] ),
                (506, [ [  118.40,  159.70,    1.30 ], [  -74.40,   26.40,  -38.50 ], [   5.76,  -1.62,  -9.17 ], [  -0.02,   0.00,   0.03 ], [  -2.33, -10.70,   0.23 ], [  -0.01,   0.00,   0.02 ] ] ),
                (507, [ [   42.70,  245.00,  -45.30 ], [   77.70,  116.50,   19.30 ], [   5.74,  -1.62,  -9.13 ], [  -0.03,   0.01,   0.04 ], [  -6.98,   6.28,  -5.54 ], [  -0.02,   0.00,   0.03 ] ] ),
                (508, [ [  159.90,  180.50,  -31.40 ], [  106.20,   76.60,   20.20 ], [   5.71,  -1.60,  -9.08 ], [  -0.03,   0.01,   0.04 ], [  -3.19,   9.46,  -4.24 ], [  -0.02,   0.00,   0.03 ] ] ),
                (509, [ [  134.10,  292.30,  -47.10 ], [   34.50,   97.60,   13.80 ], [   5.68,  -1.60,  -9.05 ], [  -0.02,   0.01,   0.03 ], [  -8.51,   2.98,  -5.95 ], [  -0.01,   0.00,   0.02 ] ] ),
                (510, [ [  193.30,  275.10,  -46.30 ], [   23.80,  -61.70,   26.10 ], [   5.67,  -1.59,  -9.02 ], [  -3.56,   1.74,  -0.59 ], [   8.40,   5.13,   4.38 ], [  -4.64,   2.27,  -0.79 ] ] ),
                (511, [ [  190.90,  187.00,  -52.20 ], [   66.50,  -39.30,    3.60 ], [  -2.01,   2.15, -10.33 ], [  -3.82,   1.31,  -0.79 ], [   2.31,  10.47,   0.68 ], [  -5.00,   1.72,  -1.05 ] ] ),
                (512, [ [  254.20,  247.30,  -47.60 ], [  -23.50,   92.50,   19.80 ], [  -1.63,   0.82, -10.55 ], [   0.99,  -1.69,  -0.13 ], [  -9.94,  -3.80,   1.15 ], [   1.29,  -2.21,  -0.20 ] ] ),
                (513, [ [  228.40,  321.30,  -41.20 ], [   52.70,  188.20,   -7.70 ], [  -0.11,  -1.17, -10.61 ], [   3.62,  -1.19,   0.81 ], [ -10.46,   2.15,  -0.18 ], [   4.76,  -1.57,   1.05 ] ] ),
                (514, [ [  176.60,  339.50,  -38.20 ], [  -78.70,   16.00,   21.50 ], [   5.61,  -1.57,  -8.92 ], [   2.47,  -0.17,   0.74 ], [   6.18,  -8.52,  -1.64 ], [   3.25,  -0.23,   0.96 ] ] ),
                (515, [ [  119.00,  367.30,  -29.70 ], [ -127.00,   18.70,    4.70 ], [   5.59,  -1.57,  -8.90 ], [  -0.02,   0.00,   0.03 ], [   4.78,  -9.48,   0.52 ], [  -0.01,   0.00,   0.02 ] ] ),
                (516, [ [   91.40,  252.30,  -63.20 ], [   23.10, -112.20,  118.60 ], [   5.57,  -1.57,  -8.86 ], [  -0.03,   0.01,   0.04 ], [   6.52,   8.32,   0.32 ], [  -0.02,   0.00,   0.03 ] ] ),
                (517, [ [  175.10,  162.40,  -53.90 ], [  -54.80,  -66.80,   32.70 ], [   5.54,  -1.56,  -8.82 ], [  -0.03,   0.01,   0.04 ], [  10.06,   0.55,   3.06 ], [  -0.02,   0.00,   0.03 ] ] ),
                (518, [ [   70.80,  203.50,  -65.80 ], [ -102.40,  -41.90,  -13.50 ], [   5.52,  -1.55,  -8.78 ], [  -0.02,   0.01,   0.03 ], [   6.25,  -7.45,   3.92 ], [  -0.01,   0.00,   0.02 ] ] ),
                (519, [ [   52.30,  140.20,  -39.60 ], [ -180.50,  -24.10,   75.00 ], [   5.51,  -1.54,  -8.75 ], [  -0.01,   0.01,   0.02 ], [   9.17,  -4.92,  -1.03 ], [  -0.01,   0.00,   0.01 ] ] ),
                (520, [ [    6.80,  106.50,  -47.40 ], [ -171.30,   31.80,   -4.10 ], [   5.49,  -1.54,  -8.74 ], [  -0.02,   0.00,   0.02 ], [   3.75,  -9.72,   0.64 ], [  -0.01,   0.00,   0.02 ] ] ),
                (521, [ [   27.90,  191.60,  -30.80 ], [  -96.90,   52.70,  -49.30 ], [   5.48,  -1.54,  -8.71 ], [  -0.02,   0.00,   0.03 ], [  -3.21,  -9.88,  -0.54 ], [  -0.01,   0.00,   0.02 ] ] ),
                (522, [ [  -17.90,  115.60,  -73.10 ], [ -104.10,   42.30,   14.10 ], [   5.45,  -1.54,  -8.67 ], [  -0.02,   0.01,   0.04 ], [   3.08,  -9.78,  -1.45 ], [  -0.01,   0.00,   0.02 ] ] ),
                (523, [ [    1.90,  223.30,  -37.40 ], [   44.00,   82.00,   -9.10 ], [   5.43,  -1.53,  -8.63 ], [  -0.02,   0.01,   0.03 ], [  -6.75,   5.34,  -5.68 ], [  -0.01,   0.00,   0.02 ] ] ),
                (524, [ [   33.20,  282.60,  -58.50 ], [  -23.00,   67.10,  -26.70 ], [   5.41,  -1.52,  -8.61 ], [   0.79,   3.52,   4.91 ], [  -8.17,  -4.48,  -4.35 ], [   1.08,   4.71,   6.57 ] ] ),
                (525, [ [   67.80,  358.20,  -24.30 ], [   -2.30,    0.30,   54.60 ], [   7.28,   6.66,   2.78 ], [   1.46,   2.72,   1.91 ], [  -6.74,   7.41,   2.19 ], [   1.97,   3.65,   2.54 ] ] ),
                (526, [ [   40.70,  358.70,   43.40 ], [  -22.60,   47.20,   59.80 ], [   8.42,   5.16,  -2.63 ], [  -2.81,   1.16,   1.68 ], [  -5.07,   4.81,  -7.46 ], [  -3.76,   1.57,   2.26 ] ] ),
                (527, [ [   29.20,  413.00,   41.50 ], [  -60.50,   10.90,   19.00 ], [   2.27,   8.57,   5.05 ], [  -2.85,   2.07,   3.48 ], [  -2.44,   5.56,  -8.20 ], [  -3.83,   2.79,   4.68 ] ] ),
                (528, [ [  -26.50,  361.80,   43.10 ], [  -56.30,   17.20,    1.60 ], [   3.56,   8.95,   3.27 ], [   2.65,  -1.03,  -4.10 ], [   1.38,   3.02,  -9.62 ], [   3.58,  -1.38,  -5.52 ] ] ),
                (529, [ [  -23.40,  393.00,   43.00 ], [   -7.70,   31.40,   -0.90 ], [   6.85,   7.25,  -1.93 ], [  -0.97,   0.74,  -1.25 ], [  -2.97,   4.65,  -8.53 ], [  -1.31,   1.01,  -1.68 ] ] ),
                (530, [ [  -36.60,  415.60,   41.80 ], [  -25.50,   17.70,    1.00 ], [   2.83,   9.74,  -0.36 ], [  -4.02,   1.55,   1.19 ], [  -3.16,   1.63,  -9.51 ], [  -5.42,   2.09,   1.60 ] ] ),
                (531, [ [  -70.00,  421.50,   45.40 ], [  -30.80,   -9.50,    1.40 ], [  -1.19,  10.06,   0.33 ], [  -5.91,  -4.42,   2.80 ], [   1.19,   0.79, -10.03 ], [  -7.98,  -5.96,   3.78 ] ] ),
                (532, [ [  -90.30,  401.60,   44.50 ], [  -15.10,  -33.40,   -0.70 ], [  -8.70,   1.66,   4.91 ], [  -2.95,  -8.68,   2.68 ], [  -5.26,   0.39,  -8.65 ], [  -3.98, -11.72,   3.63 ] ] ),
                (533, [ [  -91.80,  361.20,   44.50 ], [   35.10,  -38.10,  -18.40 ], [  -4.90,  -7.44,   4.78 ], [   6.01,  -6.11,  -1.81 ], [  -5.53,  -1.71,  -8.29 ], [   8.12,  -8.26,  -2.44 ] ] ),
                (534, [ [  -38.80,  367.10,   17.80 ], [   33.30,   56.60,  -27.60 ], [   4.46,  -9.04,   0.44 ], [   6.11,   6.76,  -2.78 ], [  -5.61,  -6.12,  -5.73 ], [   8.27,   9.14,  -3.76 ] ] ),
                (535, [ [  -55.20,  436.50,   11.50 ], [  -50.10,   54.50,   -0.10 ], [   6.83,   7.37,  -0.54 ], [   0.24,   8.59,  -0.86 ], [  -0.93,   0.22, -10.01 ], [   0.34,  11.63,  -1.17 ] ] ),
                (536, [ [ -121.40,  460.30,   17.10 ], [  -86.80,   60.90,   -1.80 ], [   5.06,   8.57,  -1.28 ], [   0.90,  -2.49,  -1.08 ], [  -1.24,  -0.68,  -9.93 ], [   1.24,  -3.38,  -1.46 ] ] ),
                (537, [ [ -164.50,  536.10,  -17.30 ], [   -9.90,   55.20,   -8.50 ], [   9.53,   1.13,  -2.81 ], [   8.03, -12.39,  -1.98 ], [  -2.54,  -2.06,  -9.45 ], [  10.93, -16.81,  -2.68 ] ] ) ] ),
            
            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-17',
                    'name': get_smallintestine_term('duodenum')[0],
                    'ontId': get_smallintestine_term('duodenum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '18-235',
                    'name': get_smallintestine_term('jejunum')[0],
                    'ontId': get_smallintestine_term('jejunum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '236-536',
                    'name': get_smallintestine_term('ileum')[0],
                    'ontId': get_smallintestine_term('ileum')[1]
                }]
        }),
        'Human 1': ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings': {
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Coordinate dimensions': 3,
                'Length': 1.0,
                'Number of elements': 150
            },
            'meshEdits': exnode_string_from_nodeset_field_parameters(
                [ Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3  ], [
                (1,   [ [ -40.03, -124.48, 1057.88 ], [  -9.88,  12.23,  -4.08 ], [ -4.33, -5.88, -7.13 ], [  2.36, -0.32,  0.12 ], [ -3.77, -1.79,  3.76 ], [ -0.25,  2.08, -4.50] ] ),
                (2,   [ [ -49.64, -111.67, 1050.88 ], [  -6.84,  19.44, -21.94 ], [ -1.59, -6.21, -5.00 ], [  0.90,  0.69,  1.57 ], [ -6.92,  0.02,  2.18 ], [ -0.76,  2.95, -2.23] ] ),
                (3,   [ [ -43.45,  -95.70, 1015.59 ], [  40.73,   3.22, -44.80 ], [ -4.94, -3.15, -4.71 ], [ -0.38,  4.38, -0.33 ], [ -2.12,  5.61, -1.53 ], [  4.59,  3.49, -0.79] ] ),
                (4,   [ [  26.76, -119.31,  992.03 ], [  58.86, -18.68,  -8.45 ], [  0.29,  3.71, -6.21 ], [  4.15,  2.33,  1.07 ], [  2.08,  5.13,  3.16 ], [  3.36, -4.08,  2.37] ] ),
                (5,   [ [  68.06, -131.66,  993.83 ], [   9.68, -30.91, -18.90 ], [  3.82,  3.44, -3.68 ], [ -0.33, -6.20,  3.78 ], [  4.96, -1.02,  4.20 ], [ -3.68, -1.57,  1.51] ] ),
                (6,   [ [  46.28, -139.94,  982.68 ], [ -22.68,  -4.85, -10.94 ], [  0.98, -6.62,  0.91 ], [ -0.93, -3.80,  3.22 ], [ -2.99,  0.39,  6.02 ], [ -5.34, -0.64, -2.85] ] ),
                (7,   [ [  28.56, -141.07,  974.61 ], [   3.75, -10.52, -21.30 ], [  1.46, -5.82,  3.13 ], [ -1.00,  0.06, -0.07 ], [ -6.42, -1.75, -0.27 ], [  1.10, -0.53, -6.40] ] ),
                (8,   [ [  52.39, -146.13,  971.40 ], [  18.40,  -3.70,   0.67 ], [ -1.35, -6.67,  0.28 ], [ -0.59, -0.29, -0.54 ], [  0.18, -0.32, -6.80 ], [  2.42,  0.06, -1.89] ] ),
                (9,   [ [  63.13, -148.18,  973.52 ], [  10.49,  -1.41,   1.47 ], [ -1.00, -6.72,  0.72 ], [  2.41,  0.38,  0.12 ], [  0.83, -0.84, -6.72 ], [ -2.03, -1.10,  0.84] ] ),
                (10,  [ [  72.69, -148.91,  974.30 ], [   7.65,   3.57,  -7.08 ], [  3.28, -5.95,  0.54 ], [  0.43,  0.20, -1.42 ], [ -3.64, -2.47, -5.18 ], [ -3.49,  0.08,  4.65] ] ),
                (11,  [ [  71.05, -143.65,  964.65 ], [  -5.96,   4.70, -11.04 ], [ -0.76, -6.41, -2.32 ], [ -2.36, -0.38, -1.14 ], [ -6.00, -0.40,  3.07 ], [  0.27,  0.95,  6.07] ] ),
                (12,  [ [  61.06, -140.78,  954.52 ], [ -17.96,   3.88,  -6.22 ], [ -1.00, -6.68, -1.29 ], [  0.18, -0.25,  1.37 ], [ -2.40, -0.88,  6.38 ], [  3.84,  0.27,  2.13] ] ),
                (13,  [ [  37.99, -136.93,  957.17 ], [ -18.82,   0.88,   5.08 ], [ -0.15, -6.89,  0.63 ], [  1.90,  0.17,  0.59 ], [  1.82,  0.57,  6.64 ], [  2.12,  1.03, -0.12] ] ),
                (14,  [ [  24.26, -138.06,  963.16 ], [ -17.40,  -6.04,   8.51 ], [  2.43, -6.48,  0.36 ], [  2.42,  1.01,  0.69 ], [  2.61,  1.33,  6.28 ], [ -2.18, -0.67, -0.74] ] ),
                (15,  [ [   6.50, -150.31,  973.05 ], [  -8.25, -16.44, -13.42 ], [  4.59, -4.46,  2.65 ], [ -1.08, -0.49, -0.28 ], [ -4.53, -1.74,  4.92 ], [ -4.27, -0.82, -2.18] ] ),
                (16,  [ [  15.97, -149.80,  957.69 ], [  -4.35,  -1.25, -14.05 ], [  0.93, -6.94,  0.33 ], [ -1.04, -0.78, -0.85 ], [ -6.52, -0.78,  2.09 ], [ -0.48,  0.19, -0.50] ] ),
                (17,  [ [   8.29, -150.98,  954.25 ], [  -3.34,  -0.73,  -8.15 ], [  1.11, -6.93,  0.17 ], [  0.18,  0.05,  0.37 ], [ -6.30, -0.94,  2.67 ], [  0.56, -0.43, -3.27] ] ),
                (18,  [ [  10.77, -150.88,  946.76 ], [   5.27,  -0.05,  -7.15 ], [  1.29, -6.84,  1.00 ], [ -0.12, -0.03,  0.05 ], [ -5.45, -1.61, -4.01 ], [  2.05,  0.22, -4.83] ] ),
                (19,  [ [  18.30, -151.09,  941.26 ], [   7.79,   0.88,  -2.39 ], [  0.83, -7.00,  0.15 ], [  0.05, -0.01, -0.90 ], [ -2.02, -0.38, -6.73 ], [  4.52,  1.49, -0.77] ] ),
                (20,  [ [  24.85, -149.58,  941.59 ], [   5.32,   0.62,   3.02 ], [  1.25, -6.91, -0.79 ], [ -0.35,  0.01, -0.65 ], [  3.30,  1.29, -6.07 ], [  3.43,  0.58,  1.19] ] ),
                (21,  [ [  27.85, -149.81,  945.92 ], [   3.49,  -0.55,   3.99 ], [  0.31, -6.95, -1.22 ], [ -0.75, -0.04, -0.15 ], [  5.27,  1.02, -4.47 ], [ -0.58, -0.16, -0.31] ] ),
                (22,  [ [  31.64, -150.64,  949.37 ], [   5.77,  -0.50,   1.81 ], [ -0.26, -6.99, -1.09 ], [  0.32, -0.01,  0.32 ], [  2.17,  0.96, -6.66 ], [ -4.59, -0.60, -0.74] ] ),
                (23,  [ [  38.16, -150.55,  948.48 ], [   5.40,   1.27,  -4.34 ], [  1.20, -6.96, -0.53 ], [  1.28,  0.32, -0.53 ], [ -4.35, -0.33, -5.52 ], [ -4.44, -1.61,  3.12] ] ),
                (24,  [ [  40.30, -148.53,  942.17 ], [  -0.50,   2.45,  -7.55 ], [  2.28, -6.33, -2.21 ], [  1.39,  0.82, -0.98 ], [ -6.59, -2.27, -0.30 ], [ -0.57, -0.95,  1.02] ] ),
                (25,  [ [  36.76, -146.19,  935.09 ], [   3.27,   5.38,  -6.44 ], [  4.06, -5.28, -2.35 ], [  0.26, -0.08,  1.07 ], [ -5.17, -2.04, -4.32 ], [  3.60,  1.24, -3.41] ] ),
                (26,  [ [  44.60, -142.08,  934.77 ], [   8.65,   3.53,   1.10 ], [  2.68, -6.58,  0.02 ], [ -1.02, -0.77,  1.00 ], [  0.78,  0.29, -7.08 ], [  3.41,  1.45, -1.30] ] ),
                (27,  [ [  53.34, -139.44,  937.21 ], [   8.31,   2.34,   2.22 ], [  2.02, -6.82, -0.36 ], [ -0.54, -0.17, -0.42 ], [  1.61,  0.84, -6.90 ], [  0.07,  0.37,  0.03] ] ),
                (28,  [ [  61.18, -137.39,  939.22 ], [   8.23,   1.72,   1.43 ], [  1.59, -6.92, -0.81 ], [ -0.38, -0.10, -0.17 ], [  1.00,  1.05, -7.00 ], [ -0.84, -0.07, -0.12] ] ),
                (29,  [ [  69.72, -136.03,  940.04 ], [   8.42,   1.51,   0.05 ], [  1.27, -7.02, -0.70 ], [  0.30,  0.21, -0.63 ], [ -0.08,  0.70, -7.13 ], [ -3.56, -0.97,  1.73] ] ),
                (30,  [ [  77.28, -134.50,  939.34 ], [   4.49,   4.18,  -8.53 ], [  2.17, -6.51, -2.05 ], [ -0.29,  0.07, -0.48 ], [ -6.03, -0.87, -3.60 ], [ -3.19, -0.95,  5.53] ] ),
                (31,  [ [  71.89, -131.27,  929.11 ], [  -6.78,   1.09,  -6.79 ], [  0.08, -7.09, -1.22 ], [ -0.44, -0.26,  1.25 ], [ -5.09, -0.91,  4.94 ], [  0.24,  0.45,  2.90] ] ),
                (32,  [ [  66.11, -131.60,  925.90 ], [  -4.09,  -0.45,  -4.44 ], [  0.51, -7.19,  0.26 ], [  0.08,  0.05,  1.40 ], [ -5.23, -0.20,  4.83 ], [ -0.15,  0.88, -0.16] ] ),
                (33,  [ [  63.93, -132.09,  921.07 ], [  -3.93,  -1.16,  -4.37 ], [  0.30, -7.03,  1.60 ], [  0.72,  0.30,  0.95 ], [ -5.38,  0.82,  4.63 ], [  0.66,  0.51,  0.56] ] ),
                (34,  [ [  58.47, -133.89,  917.96 ], [  -6.67,  -3.36,  -3.67 ], [  2.16, -6.56,  2.08 ], [  2.28,  0.92, -0.11 ], [ -3.73,  0.71,  6.12 ], [  2.25,  0.03,  1.31] ] ),
                (35,  [ [  51.02, -139.09,  914.02 ], [  -6.63,  -6.74,   0.39 ], [  5.08, -4.94,  1.06 ], [  1.07,  0.42, -1.32 ], [ -0.56,  0.96,  7.15 ], [  3.12,  0.37,  0.20] ] ),
                (36,  [ [  46.77, -145.49,  918.12 ], [  -5.05,  -4.37,   2.87 ], [  4.53, -5.57, -0.52 ], [ -1.31, -0.89, -0.94 ], [  2.52,  1.43,  6.62 ], [  1.26, -0.68, -0.12] ] ),
                (37,  [ [  41.95, -147.92,  919.78 ], [  -5.03,  -2.38,   1.94 ], [  2.73, -6.62, -1.04 ], [ -1.32, -0.68, -0.33 ], [  2.61,  0.01,  6.76 ], [ -0.11, -1.00,  0.12] ] ),
                (38,  [ [  36.76, -150.21,  922.00 ], [  -5.10,  -1.68,   1.58 ], [  1.91, -6.90, -1.17 ], [ -0.51, -0.22,  0.81 ], [  2.30, -0.53,  6.87 ], [ -0.45,  0.50,  0.13] ] ),
                (39,  [ [  31.89, -151.33,  922.99 ], [  -5.71,  -1.25,   1.56 ], [  1.67, -7.06,  0.43 ], [ -0.57, -0.18,  0.73 ], [  1.73,  0.83,  7.02 ], [ -2.05,  0.33,  0.03] ] ),
                (40,  [ [  25.55, -152.65,  925.18 ], [  -6.28,  -0.56,  -2.17 ], [  0.66, -7.26, -0.03 ], [ -0.69, -0.09, -0.62 ], [ -2.35, -0.24,  6.88 ], [ -3.19, -0.78, -0.83] ] ),
                (41,  [ [  22.17, -152.27,  920.06 ], [  -4.85,   0.30,  -4.20 ], [  0.25, -7.25, -0.80 ], [ -0.24, -0.02,  0.24 ], [ -4.73, -0.76,  5.42 ], [  2.41,  0.37, -0.06] ] ),
                (42,  [ [  16.60, -152.11,  917.51 ], [  -6.97,   0.00,   2.68 ], [  0.19, -7.29,  0.48 ], [ -0.30, -0.03,  0.60 ], [  2.61,  0.51,  6.79 ], [  5.78,  0.37, -0.74] ] ),
                (43,  [ [  13.12, -152.32,  924.90 ], [  -3.73,   0.41,   7.15 ], [ -0.42, -7.30,  0.20 ], [ -0.14, -0.02, -0.16 ], [  6.38, -0.28,  3.34 ], [ -0.23, -0.18,  0.20] ] ),
                (44,  [ [   9.44, -151.37,  931.31 ], [  -7.88,   0.19,   2.81 ], [ -0.12, -7.33,  0.16 ], [  0.49, -0.01,  0.12 ], [  2.46,  0.11,  6.89 ], [ -3.70,  0.33,  2.08] ] ),
                (45,  [ [   1.23, -152.17,  928.31 ], [  -8.13,  -0.72,  -1.08 ], [  0.58, -7.31,  0.46 ], [  0.09, -0.01, -0.29 ], [ -1.00,  0.38,  7.26 ], [  1.43, -0.16, -0.65] ] ),
                (46,  [ [  -5.97, -152.74,  928.93 ], [  -5.64,  -0.31,   4.39 ], [  0.16, -7.35, -0.31 ], [ -0.93,  0.04,  0.09 ], [  4.48, -0.15,  5.76 ], [  3.98, -0.84, -4.11] ] ),
                (47,  [ [  -8.05, -152.70,  935.12 ], [   0.84,   0.43,   7.28 ], [ -1.23, -7.23,  0.57 ], [ -1.08,  0.17,  0.74 ], [  7.08, -1.26, -0.74 ], [  0.79, -0.43, -1.11] ] ),
                (48,  [ [  -4.18, -151.98,  940.93 ], [  -5.50,   2.85,   8.02 ], [ -1.95, -7.01,  1.15 ], [ -0.37,  0.07, -0.10 ], [  5.81, -0.90,  4.30 ], [ -3.01,  0.56,  4.29] ] ),
                (49,  [ [ -15.92, -150.23,  939.85 ], [ -11.53,   2.83,   0.08 ], [ -1.76, -7.18, -0.02 ], [  0.12,  0.15,  0.78 ], [  0.04, -0.03,  7.39 ], [  0.81,  0.00, -1.16] ] ),
                (50,  [ [ -25.87, -146.77,  940.97 ], [  -2.69,   4.15,   9.23 ], [ -1.70, -6.74,  2.54 ], [  0.69,  0.20,  1.48 ], [  6.83, -0.83,  2.36 ], [  1.72, -1.41, -6.49] ] ),
                (51,  [ [ -19.43, -146.58,  946.99 ], [   7.70,   1.83,   5.13 ], [ -0.46, -6.73,  3.10 ], [  1.24, -0.13, -0.09 ], [  4.24, -2.77, -5.37 ], [ -0.59, -0.04, -3.45] ] ),
                (52,  [ [ -11.61, -143.43,  950.54 ], [   5.25,   2.52,   5.68 ], [  0.78, -7.00,  2.38 ], [  0.75, -0.24, -0.87 ], [  5.57, -0.98, -4.71 ], [  0.83,  1.41,  0.57] ] ),
                (53,  [ [  -9.22, -141.82,  956.88 ], [   3.60,   1.58,   5.20 ], [  1.15, -7.22,  1.40 ], [ -0.05, -0.13, -0.42 ], [  6.02,  0.14, -4.21 ], [  0.12,  0.28,  0.10] ] ),
                (54,  [ [  -4.88, -140.39,  960.70 ], [   3.78,   1.36,   4.90 ], [  0.74, -7.28,  1.45 ], [  0.07, -0.06, -0.65 ], [  5.87, -0.29, -4.45 ], [  0.53,  0.53,  1.66] ] ),
                (55,  [ [  -1.98, -139.20,  966.47 ], [   0.65,   0.13,   7.29 ], [  1.34, -7.34,  0.01 ], [  0.23, -0.06, -0.88 ], [  7.17,  1.31, -0.66 ], [  0.66,  0.74,  3.14] ] ),
                (56,  [ [  -4.15, -140.36,  973.96 ], [  -2.38,  -0.65,  10.17 ], [  1.12, -7.39, -0.21 ], [ -0.65, -0.07, -0.20 ], [  7.07,  1.03,  1.72 ], [ -0.21, -0.70,  2.00] ] ),
                (57,  [ [  -6.44, -140.06,  986.68 ], [  -4.21,  -0.33,   9.07 ], [ -0.23, -7.49, -0.38 ], [ -0.77,  0.02, -0.44 ], [  6.68, -0.36,  3.09 ], [ -1.82, -1.02,  2.46] ] ),
                (58,  [ [ -10.86, -140.67,  991.98 ], [  -5.80,   0.04,   3.85 ], [ -0.69, -7.41, -0.97 ], [  0.04, -0.00, -0.15 ], [  4.08, -1.18,  6.14 ], [ -3.51, -0.15,  2.18] ] ),
                (59,  [ [ -17.28, -140.01,  993.95 ], [  -7.30,   0.18,  -0.28 ], [ -0.16, -7.49, -0.69 ], [  0.59, -0.04,  0.75 ], [ -0.30, -0.68,  7.48 ], [ -4.07,  0.69,  0.16] ] ),
                (60,  [ [ -24.40, -140.40,  991.18 ], [  -6.25,  -0.73,  -3.97 ], [  0.50, -7.49,  0.59 ], [ -0.08,  0.16,  1.48 ], [ -4.03,  0.23,  6.31 ], [ -3.14,  0.95, -2.22] ] ),
                (61,  [ [ -29.47, -141.40,  986.35 ], [  -2.47,  -1.37,  -4.69 ], [ -0.26, -7.19,  2.25 ], [ -1.43,  0.31,  0.61 ], [ -6.62,  1.22,  3.13 ], [ -0.94,  0.18, -5.15] ] ),
                (62,  [ [ -29.90, -142.61,  982.98 ], [   1.86,  -2.14,  -4.74 ], [ -2.05, -6.88,  2.30 ], [ -1.16,  0.11, -0.29 ], [ -6.69,  0.97, -3.06 ], [  0.49, -0.35, -4.87] ] ),
                (63,  [ [ -24.82, -145.19,  978.58 ], [   4.65,  -2.18,  -4.59 ], [ -2.01, -7.15,  1.36 ], [  0.29, -0.26, -0.91 ], [ -5.15,  0.42, -5.42 ], [ -0.33,  0.23,  1.00] ] ),
                (64,  [ [ -20.75, -146.91,  974.01 ], [   1.23,  -0.70,  -6.94 ], [ -1.49, -7.40,  0.48 ], [  0.08, -0.09, -0.69 ], [ -7.16,  1.35, -1.40 ], [  0.46,  0.35,  5.77] ] ),
                (65,  [ [ -23.39, -146.10,  967.33 ], [  -6.32,   1.65,  -4.01 ], [ -1.92, -7.32,  0.01 ], [ -0.16,  0.07,  0.36 ], [ -3.81,  1.01,  6.42 ], [  4.83, -0.40,  4.28] ] ),
                (66,  [ [ -30.60, -144.33,  967.78 ], [  -6.69,   2.04,   2.19 ], [ -1.82, -7.26,  1.20 ], [  0.05,  0.00,  0.44 ], [  2.50,  0.55,  7.13 ], [  3.03, -0.35,  0.34] ] ),
                (67,  [ [ -36.17, -142.25,  971.34 ], [  -6.59,   1.94,   2.27 ], [ -1.83, -7.30,  0.95 ], [  0.04, -0.01,  0.09 ], [  2.55,  0.29,  7.14 ], [  0.37, -0.02, -0.15] ] ),
                (68,  [ [ -43.40, -140.59,  972.07 ], [  -6.04,   1.97,   2.74 ], [ -1.73, -7.28,  1.40 ], [  0.02,  0.34,  1.18 ], [  3.28,  0.53,  6.83 ], [  2.30, -0.41, -2.27] ] ),
                (69,  [ [ -47.62, -138.63,  976.14 ], [  -1.78,   2.68,   4.62 ], [ -1.77, -6.68,  3.20 ], [  0.01,  0.35,  0.98 ], [  6.89, -0.43,  2.91 ], [  1.72, -1.44, -3.71] ] ),
                (70,  [ [ -47.10, -135.98,  980.16 ], [   0.98,   1.95,   4.05 ], [ -1.73, -6.52,  3.55 ], [  0.21, -0.14, -0.20 ], [  7.12, -2.24, -0.64 ], [  0.04, -0.76, -2.21] ] ),
                (71,  [ [ -45.78, -134.75,  984.02 ], [   1.92,   1.94,   5.58 ], [ -1.38, -6.93,  2.88 ], [ -0.53,  0.37,  0.09 ], [  7.01, -2.09, -1.69 ], [ -0.24,  0.26,  1.07] ] ),
                (72,  [ [ -43.37, -132.13,  990.97 ], [  -1.92,   7.78,   7.36 ], [ -3.57, -5.07,  4.43 ], [ -2.14,  1.26, -0.06 ], [  6.52, -1.62,  3.41 ], [ -3.47,  2.54,  4.15] ] ),
                (73,  [ [ -50.92, -122.17,  992.94 ], [  -8.28,   6.98,  -7.76 ], [ -5.63, -4.80,  1.69 ], [  0.61, -1.11, -1.84 ], [ -1.92,  4.37,  5.99 ], [ -6.54,  1.98, -0.43] ] ),
                (74,  [ [ -54.52, -122.57,  980.90 ], [  -5.10,   0.46, -13.05 ], [ -2.46, -7.24,  0.71 ], [  2.85, -1.09,  0.85 ], [ -6.63,  2.51,  2.68 ], [ -2.86, -1.79, -2.31] ] ),
                (75,  [ [ -60.93, -121.16,  967.78 ], [  -3.40,  -6.95, -12.83 ], [  0.01, -6.78,  3.67 ], [ -1.69,  1.75,  1.25 ], [ -7.36,  0.81,  1.51 ], [  1.21, -0.70, -4.51] ] ),
                (76,  [ [ -59.94, -133.62,  960.46 ], [   4.44, -11.12,  -4.92 ], [ -5.83, -3.75,  3.22 ], [ -1.80,  0.34, -0.38 ], [ -4.21,  1.12, -6.33 ], [  1.72, -0.86, -3.60] ] ),
                (77,  [ [ -53.65, -142.09,  958.04 ], [   8.15,  -8.95,  -3.45 ], [ -4.68, -5.37,  2.89 ], [  2.04, -1.62, -0.11 ], [ -3.54, -0.59, -6.83 ], [ -0.92, -0.86,  0.97] ] ),
                (78,  [ [ -43.78, -150.90,  953.55 ], [   6.86,  -6.21, -10.82 ], [ -1.48, -6.98,  3.06 ], [  0.10, -0.47, -1.48 ], [ -6.55, -0.34, -3.95 ], [ -0.50,  1.88,  6.12] ] ),
                (79,  [ [ -42.79, -151.92,  940.92 ], [ -10.16,   7.35, -12.87 ], [ -4.48, -6.31, -0.07 ], [ -1.78,  0.37, -1.94 ], [ -4.55,  3.17,  5.40 ], [  1.37,  1.93,  5.90] ] ),
                (80,  [ [ -59.56, -139.32,  940.87 ], [ -10.47,   7.53, -10.55 ], [ -4.41, -6.40, -0.20 ], [ -0.78,  0.74,  0.77 ], [ -4.15,  2.67,  6.02 ], [ -0.63,  1.74, -4.22] ] ),
                (81,  [ [ -61.19, -138.43,  930.23 ], [  -0.37,  -2.04, -12.19 ], [ -5.64, -5.22,  1.05 ], [ -0.86,  0.96,  0.96 ], [ -5.33,  5.60, -0.78 ], [  0.65, -0.72, -6.69] ] ),
                (82,  [ [ -60.05, -143.37,  918.84 ], [   6.12,  -9.41,  -3.11 ], [ -6.09, -4.52,  1.69 ], [  0.25, -0.37, -0.76 ], [ -2.60,  0.74, -7.36 ], [  3.39, -3.10, -3.06] ] ),
                (83,  [ [ -54.81, -148.97,  924.54 ], [   7.07,  -6.60,   2.05 ], [ -5.30, -5.72, -0.16 ], [  0.76, -0.86, -0.33 ], [  1.31, -0.99, -7.69 ], [ -1.61,  1.40,  1.58] ] ),
                (84,  [ [ -48.00, -154.67,  922.57 ], [   3.54,  -3.72,  -7.48 ], [ -4.57, -6.25,  0.95 ], [  1.21, -0.76, -0.14 ], [ -5.53,  3.38, -4.30 ], [ -4.06,  1.81,  3.83] ] ),
                (85,  [ [ -49.39, -154.51,  914.46 ], [   0.27,   0.35, -10.67 ], [ -2.94, -7.22, -0.31 ], [  1.86, -0.79, -0.48 ], [ -7.12,  2.90, -0.09 ], [  1.64, -1.57, -0.54] ] ),
                (86,  [ [ -46.73, -154.01,  904.00 ], [  18.99,  -1.89,  -1.48 ], [ -0.74, -7.74,  0.38 ], [  1.31, -0.30,  0.50 ], [ -0.63, -0.32, -7.74 ], [  5.19, -2.31, -4.98] ] ),
                (87,  [ [ -31.90, -157.00,  923.20 ], [  20.27,  -2.84,   5.71 ], [ -1.21, -7.60,  0.51 ], [  1.63,  0.18,  0.17 ], [  1.98, -0.81, -7.42 ], [ -3.21, -0.77,  1.42] ] ),
                (88,  [ [ -17.04, -158.47,  917.33 ], [  10.69,   1.60, -10.94 ], [  1.83, -7.40,  0.70 ], [  0.87,  0.01, -0.11 ], [ -5.13, -1.77, -5.27 ], [ -1.71,  0.19, -0.09] ] ),
                (89,  [ [ -12.92, -154.88,  905.59 ], [  14.28,   1.56,  -5.12 ], [  0.95, -7.55,  0.35 ], [ -0.98, -0.09, -0.58 ], [ -2.50, -0.65, -7.16 ], [  1.89,  1.14, -1.15] ] ),
                (90,  [ [   0.79, -157.22,  913.22 ], [  17.05,  -0.10,  -3.50 ], [ -0.15, -7.56, -0.50 ], [  0.24,  0.09, -0.24 ], [ -1.52,  0.52, -7.39 ], [  0.10,  0.16,  0.01] ] ),
                (91,  [ [  11.37, -154.85,  899.81 ], [  20.47,   4.34,  -7.04 ], [  1.53, -7.36, -0.09 ], [  0.67,  0.24, -0.62 ], [ -2.36, -0.40, -7.11 ], [ -1.60,  0.07,  0.92] ] ),
                (92,  [ [  35.28, -149.96,  905.73 ], [  15.01,   5.71, -13.20 ], [  0.75, -7.07, -2.20 ], [  1.12,  0.33,  0.34 ], [ -5.05,  1.10, -5.27 ], [ -1.05, -1.38,  0.95] ] ),
                (93,  [ [  31.51, -148.90,  893.07 ], [  11.51,   6.02, -14.64 ], [  2.99, -6.72, -0.41 ], [  1.65,  0.51,  0.45 ], [ -5.13, -1.99, -4.85 ], [  2.20, -0.03, -0.53] ] ),
                (94,  [ [  53.21, -142.18,  892.15 ], [  16.13,   8.56,   4.60 ], [  3.73, -5.96, -2.01 ], [  0.42,  0.20,  0.80 ], [  0.55,  2.65, -6.83 ], [  3.44,  1.00, -0.61] ] ),
                (95,  [ [  62.43, -134.45,  898.99 ], [  11.35,   7.61,   6.27 ], [  3.97, -6.09,  0.21 ], [ -1.03, -0.53,  1.15 ], [  2.65,  1.50, -6.63 ], [ -2.20, -1.80,  0.66] ] ),
                (96,  [ [  74.83, -127.92,  903.83 ], [  13.27,   2.58, -11.41 ], [  1.50, -7.08,  0.14 ], [ -0.63, -0.25,  0.20 ], [ -4.49, -1.05, -5.45 ], [ -4.38, -1.93,  4.88] ] ),
                (97,  [ [  72.51, -133.32,  887.20 ], [ -12.49,  -8.25, -22.65 ], [  3.07, -6.44,  0.66 ], [  0.84,  0.36, -0.04 ], [ -5.54, -2.24,  3.87 ], [  0.64, -0.37,  7.04] ] ),
                (98,  [ [  47.82, -142.63,  864.93 ], [ -22.72,  -8.47,  -7.52 ], [  2.59, -6.58, -0.40 ], [  0.03,  0.23, -1.27 ], [ -1.82, -1.13,  6.76 ], [  5.33,  1.35,  0.11] ] ),
                (99,  [ [  33.71, -147.84,  866.62 ], [  -7.31,  -5.02,   5.80 ], [  2.84, -6.19, -1.77 ], [ -0.71,  0.24, -1.45 ], [  4.23,  0.33,  5.62 ], [  2.35,  2.56, -6.78] ] ),
                (100, [ [  33.45, -149.55,  870.08 ], [   6.66,  -2.67,   8.62 ], [  1.80, -6.00, -3.25 ], [ -0.81,  0.03, -0.86 ], [  5.31,  3.27, -3.10 ], [  0.80,  2.08, -6.77] ] ),
                (101, [ [  48.50, -149.86,  877.56 ], [   8.44,  -1.40,  10.88 ], [  1.73, -6.47, -2.17 ], [ -0.55, -0.39,  1.17 ], [  5.25,  2.65, -3.73 ], [ -1.37, -2.16,  5.89] ] ),
                (102, [ [  49.17, -151.34,  885.88 ], [ -14.13,  -2.76,   6.71 ], [  0.88, -6.81, -0.95 ], [ -0.57,  0.33, -0.41 ], [  3.06, -0.48,  6.24 ], [ -3.09, -3.47,  6.17] ] ),
                (103, [ [  30.33, -151.82,  877.70 ], [ -17.16,   1.53,  -4.96 ], [  0.86, -5.16, -4.59 ], [ -1.51, -0.14,  0.40 ], [ -1.82, -4.63,  4.87 ], [  0.64, -0.13,  0.13] ] ),
                (104, [ [  16.36, -149.06,  875.57 ], [ -10.97,   1.71,   5.38 ], [ -1.69, -6.55, -1.37 ], [ -0.54, -0.73,  1.97 ], [  2.66, -1.95,  6.05 ], [  2.86,  1.80, -0.06] ] ),
                (105, [ [  11.69, -148.72,  883.92 ], [ -10.14,   1.00,   8.83 ], [ -0.88, -6.83, -0.24 ], [  0.66, -0.15,  1.09 ], [  4.42, -0.75,  5.16 ], [ -0.40,  1.32,  0.04] ] ),
                (106, [ [  -3.83, -146.99,  890.11 ], [ -14.28,   1.09,   0.55 ], [ -0.49, -6.77,  0.78 ], [  0.36,  0.01, -0.26 ], [  0.32,  0.76,  6.79 ], [ -1.67, -0.06,  0.66] ] ),
                (107, [ [ -14.79, -146.52,  886.71 ], [ -10.85,   0.22,   0.51 ], [ -0.15, -6.79, -0.36 ], [  0.73,  0.03, -0.30 ], [  0.31, -0.37,  6.78 ], [  2.18,  0.02, -0.94] ] ),
                (108, [ [ -23.91, -146.55,  890.53 ], [  -7.40,  -0.95,   6.58 ], [  0.92, -6.71,  0.07 ], [ -0.52,  0.10,  0.63 ], [  4.39,  0.65,  5.03 ], [  0.75,  0.45, -0.15] ] ),
                (109, [ [ -28.03, -148.18,  898.37 ], [ -13.61,   2.90,   3.80 ], [ -1.15, -6.58,  0.90 ], [ -1.43,  0.37, -0.84 ], [  1.91,  0.54,  6.43 ], [ -3.99, -0.65,  0.17] ] ),
                (110, [ [ -42.02, -140.32,  889.27 ], [  -9.93,   7.70,  -9.68 ], [ -1.35, -5.73, -3.18 ], [  0.84, -0.04, -0.89 ], [ -4.99, -1.16,  4.20 ], [ -3.62,  0.11, -4.10] ] ),
                (111, [ [ -47.83, -133.82,  880.59 ], [   1.67,   2.81,  -9.02 ], [  0.18, -6.37, -1.95 ], [  0.02, -0.24,  2.82 ], [ -6.43,  0.16, -1.14 ], [  2.63, -0.65, -5.17] ] ),
                (112, [ [ -43.36, -134.64,  876.97 ], [   7.37,  -1.12,  -1.77 ], [ -0.57, -6.40,  1.69 ], [ -0.49, -0.09,  1.49 ], [ -1.72, -1.49, -6.23 ], [  3.37, -0.42, -3.09] ] ),
                (113, [ [ -34.80, -135.76,  878.37 ], [   7.33,  -0.80,  -0.16 ], [ -0.71, -6.57,  0.26 ], [ -0.51,  0.02, -0.71 ], [ -0.17, -0.24, -6.61 ], [ -2.44,  1.36,  1.66] ] ),
                (114, [ [ -29.32, -136.24,  877.18 ], [   3.39,  -0.80,  -5.31 ], [ -1.47, -6.43,  0.04 ], [ -0.23,  0.05,  0.03 ], [ -5.30,  1.19, -3.57 ], [ -3.22,  0.72,  3.41] ] ),
                (115, [ [ -30.61, -136.87,  870.77 ], [  -0.25,  -0.25,  -5.29 ], [ -1.11, -6.47,  0.36 ], [  0.28, -0.03, -0.03 ], [ -6.35,  1.10,  0.25 ], [  0.35, -0.33, -1.21] ] ),
                (116, [ [ -30.23, -136.86,  867.08 ], [   2.79,  -0.45,  -3.89 ], [ -0.88, -6.50,  0.13 ], [  0.33, -0.03, -0.12 ], [ -5.21,  0.63, -3.81 ], [  3.30, -0.63, -3.48] ] ),
                (117, [ [ -25.37, -137.76,  865.06 ], [   8.76,  -0.50,   1.85 ], [ -0.41, -6.53,  0.17 ], [  0.24,  0.00, -0.17 ], [  1.34, -0.25, -6.40 ], [  3.73, -0.35, -1.77] ] ),
                (118, [ [ -17.01, -136.93,  873.25 ], [   9.54,  -0.88,  -1.35 ], [ -0.66, -6.46, -0.44 ], [ -0.79,  0.17,  0.25 ], [ -0.86,  0.52, -6.43 ], [ -3.99,  0.94,  2.86] ] ),
                (119, [ [ -14.62, -138.10,  867.37 ], [   2.47,  -1.15,  -8.78 ], [ -1.76, -6.23,  0.32 ], [  1.25,  0.41, -0.04 ], [ -5.90,  1.57, -1.86 ], [ -1.18,  0.21,  1.02] ] ),
                (120, [ [ -13.21, -138.67,  858.26 ], [  15.02,   8.84,  -1.86 ], [  3.15, -5.55, -0.95 ], [  3.02,  0.32, -0.46 ], [ -1.07,  0.48, -6.36 ], [  4.30, -0.76, -2.96] ] ),
                (121, [ [  -2.01, -129.09,  875.92 ], [  14.73,   5.57,   5.54 ], [  2.15, -6.01,  0.34 ], [ -3.38,  0.09,  1.42 ], [  2.11,  0.42, -6.03 ], [ -4.26,  0.28,  1.85] ] ),
                (122, [ [   5.71, -128.31,  873.45 ], [   4.22,  -3.80,  -6.92 ], [ -2.18, -5.70,  1.80 ], [  1.54,  1.12,  0.79 ], [ -5.12,  0.83, -3.57 ], [ -2.90, -1.92,  4.89] ] ),
                (123, [ [   3.86, -133.45,  867.51 ], [  -1.78,  -5.98,  -7.59 ], [  4.53, -3.88,  2.00 ], [  1.96,  0.65,  1.01 ], [ -4.22, -3.14,  3.46 ], [  0.30, -2.10,  1.25] ] ),
                (124, [ [   2.44, -139.69,  858.80 ], [   7.02,  -5.19,  -6.23 ], [  0.10, -4.81,  4.11 ], [ -1.69, -0.89,  0.21 ], [ -4.72, -2.72, -3.06 ], [  2.22,  0.25, -4.49] ] ),
                (125, [ [  11.98, -140.15,  859.35 ], [  17.70,   2.25,  -1.14 ], [  0.89, -5.66,  2.60 ], [  1.39, -0.39, -2.08 ], [ -0.04, -2.63, -5.72 ], [  2.19,  0.11, -1.39] ] ),
                (126, [ [  35.30, -132.99,  854.91 ], [   8.74,   9.55, -15.22 ], [  3.78, -4.88, -0.90 ], [  0.82,  0.89, -2.82 ], [ -4.13, -2.47, -3.92 ], [ -1.33, -1.30,  4.05] ] ),
                (127, [ [  29.93, -128.55,  843.92 ], [ -12.95,   5.12, -19.36 ], [  3.43, -3.92, -3.33 ], [ -1.41,  0.00, -0.87 ], [ -3.91, -4.60,  1.39 ], [  1.70, -0.30,  5.11] ] ),
                (128, [ [   9.39, -125.79,  820.06 ], [ -27.75,   1.44,   5.84 ], [ -0.44, -6.13, -0.59 ], [ -1.27, -0.92,  1.78 ], [  1.23, -0.67,  6.02 ], [  4.74,  2.19, -0.36] ] ),
                (129, [ [   4.65, -126.74,  840.92 ], [  -6.02,   1.18,  15.73 ], [  0.12, -6.14,  0.51 ], [ -0.73,  0.09,  0.19 ], [  5.69,  0.29,  2.16 ], [ -2.11,  0.21,  1.57] ] ),
                (130, [ [   0.04, -125.01,  849.26 ], [  -9.24,   1.85,   1.00 ], [ -1.17, -6.01,  0.30 ], [  0.68,  0.07, -0.54 ], [  0.70,  0.17,  6.10 ], [ -5.74, -0.82, -0.60] ] ),
                (131, [ [  -4.44, -124.79,  843.15 ], [  -5.02,   0.46, -16.66 ], [  1.17, -5.99, -0.52 ], [  1.98,  0.08, -0.42 ], [ -5.67, -1.25,  1.67 ], [ -4.27, -1.02, -2.57] ] ),
                (132, [ [  -4.10, -124.24,  817.80 ], [ -14.82,  -5.98, -12.69 ], [  2.08, -5.71,  0.27 ], [ -0.08, -0.02,  0.18 ], [ -3.61, -1.09,  4.72 ], [  6.33,  1.59, -0.02] ] ),
                (133, [ [ -13.82, -128.40,  823.84 ], [  -7.49,  -1.58,  13.51 ], [  1.44, -5.89,  0.11 ], [ -0.55,  0.00,  0.71 ], [  5.11,  1.30,  2.98 ], [  4.92,  1.22, -2.01] ] ),
                (134, [ [ -15.54, -126.40,  839.81 ], [  -2.26,   3.63,  12.78 ], [  0.99, -5.69,  1.79 ], [ -1.86,  0.05, -0.36 ], [  5.77,  1.22,  0.68 ], [ -0.95, -1.43,  1.43] ] ),
                (135, [ [ -17.86, -122.19,  848.91 ], [  -6.97,   2.29,   6.40 ], [ -1.77, -5.74,  0.13 ], [  0.21, -0.11, -0.73 ], [  3.78, -1.07,  4.51 ], [ -5.50, -0.82,  1.29] ] ),
                (136, [ [ -24.46, -122.95,  849.04 ], [ -11.05,  -2.26, -14.69 ], [  0.96, -5.91,  0.19 ], [  1.60, -0.05,  0.47 ], [ -4.70, -0.65,  3.63 ], [ -6.40,  0.95, -0.75] ] ),
                (137, [ [ -23.98, -124.71,  821.70 ], [ -13.90,  -3.45, -17.21 ], [ -1.01, -5.56,  1.93 ], [ -0.11, -0.11, -0.76 ], [ -4.54,  1.96,  3.28 ], [  5.05, -0.43,  0.87] ] ),
                (138, [ [ -38.60, -127.22,  820.18 ], [ -12.58,  -0.43,   8.78 ], [ -0.02, -5.94, -0.31 ], [ -0.17, -0.02, -0.17 ], [  3.41, -0.27,  4.87 ], [  4.86, -1.54, -2.07] ] ),
                (139, [ [ -42.52, -125.65,  832.78 ], [   1.33,   2.91,  14.25 ], [ -1.16, -5.67,  1.27 ], [ -0.03,  0.21,  1.19 ], [  5.67, -1.22, -0.28 ], [  1.08, -0.26, -3.49] ] ),
                (140, [ [ -35.80, -121.92,  845.73 ], [   4.20,   3.78,  10.20 ], [  0.01, -5.54,  2.04 ], [ -0.16,  0.23,  0.57 ], [  5.46, -0.72, -1.98 ], [  0.09,  0.03,  0.69] ] ),
                (141, [ [ -33.62, -118.72,  852.84 ], [   0.50,   3.01,   6.47 ], [ -0.86, -5.25,  2.51 ], [ -0.79, -0.03, -0.57 ], [  5.71, -0.94,  0.00 ], [ -0.33, -0.03,  2.50] ] ),
                (142, [ [ -34.44, -116.11,  858.28 ], [  -2.64,   1.81,   5.04 ], [ -1.59, -5.52,  1.14 ], [  0.58, -0.29, -1.41 ], [  4.94, -0.83,  2.89 ], [ -1.83,  0.37,  2.71] ] ),
                (143, [ [ -38.38, -115.33,  862.17 ], [  -6.05,  -0.34,   2.35 ], [  0.20, -5.83, -0.32 ], [  1.37, -0.09, -0.25 ], [  2.12, -0.22,  5.43 ], [ -4.21,  0.45,  0.94] ] ),
                (144, [ [ -44.50, -116.93,  861.58 ], [ -11.59,  -3.81, -10.16 ], [  1.03, -5.65,  0.95 ], [  0.44,  0.12,  0.61 ], [ -3.83,  0.03,  4.36 ], [ -4.54,  0.05, -0.79] ] ),
                (145, [ [ -51.78, -120.72,  839.03 ], [ -15.88,  -5.45, -10.36 ], [  1.70, -5.54,  0.30 ], [ -1.36,  0.27,  0.87 ], [ -2.99, -0.65,  4.93 ], [  4.29,  0.49,  0.09] ] ),
                (146, [ [ -64.65, -124.35,  841.48 ], [ -11.51,   0.42,   9.85 ], [  0.14, -5.74,  0.41 ], [ -0.55,  0.33,  1.54 ], [  3.68,  0.40,  4.29 ], [  4.38, -0.52, -2.90] ] ),
                (147, [ [ -69.37, -119.95,  854.33 ], [   5.11,   4.21,  16.62 ], [ -1.23, -5.32,  1.73 ], [  0.69,  0.27,  0.73 ], [  5.25, -1.61, -1.21 ], [  0.56, -1.58, -4.18] ] ),
                (148, [ [ -55.63, -118.57,  867.17 ], [   8.19,  15.70,   8.57 ], [  0.64, -2.10,  3.23 ], [  0.89,  1.63,  1.44 ], [  3.88, -1.19, -1.54 ], [  0.13,  0.20,  0.14] ] ),
                (149, [ [ -52.38, -100.90,  869.76 ], [   4.76,  24.01,  12.61 ], [  0.45, -2.26,  4.14 ], [ -0.28, -0.30, -1.61 ], [  4.16, -0.46, -0.70 ], [ -2.29,  1.10,  4.28] ] ),
                (150, [ [ -41.73,  -76.29,  892.90 ], [ -14.02,  14.24,  19.65 ], [ -3.16, -2.70, -0.30 ], [  3.24, -2.89,  0.75 ], [  1.74, -2.36,  2.95 ], [ -5.09,  5.86,  6.61] ] ),
                (151, [ [ -58.69,  -78.02,  896.31 ], [  -2.60,  -3.83,  -1.89 ], [  4.23, -2.00, -1.76 ], [  3.24, -2.89,  0.75 ], [  0.59, -2.51,  4.28 ], [ -5.09,  5.86,  6.61] ] ) ] ),

            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-5',
                    'name': get_smallintestine_term('duodenum')[0],
                    'ontId': get_smallintestine_term('duodenum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '6-82',
                    'name': get_smallintestine_term('jejunum')[0],
                    'ontId': get_smallintestine_term('jejunum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '83-150',
                    'name': get_smallintestine_term('ileum')[0],
                    'ontId': get_smallintestine_term('ileum')[1]
                }]
        }),

        'Mouse 1' : ScaffoldPackage(MeshType_1d_path1, {
            'scaffoldSettings' : {
                'D2 derivatives': True,
                'D3 derivatives': True,
                'Coordinate dimensions' : 3,
                'Length' : 1.0,
                'Number of elements' : 45
                },
            'meshEdits' : exnode_string_from_nodeset_field_parameters(
                [ Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1, Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D_DS3, Node.VALUE_LABEL_D2_DS1DS3  ], [
                (1,  [ [  -2.30, 18.50,  -4.40 ], [ -4.20, -0.80,   3.70 ], [  0.00,  0.60,  0.00 ], [  0.00,  0.11,  0.00 ], [ -0.33,  0.01, -0.50 ], [ 0.00, 0.00, 0.50 ] ] ),
                (2,  [ [  -8.60, 16.30,  -0.40 ], [ -7.10, -2.70,   1.60 ], [  0.00,  0.73,  0.00 ], [  0.00,  0.14,  0.00 ], [  0.08,  0.09, -0.72 ], [ 0.00, 0.00, 0.50 ] ] ),
                (3,  [ [ -18.30, 12.60,  -1.50 ], [ -6.40, -1.70,  -3.80 ], [  0.00,  0.90,  0.00 ], [  0.00,  0.13,  0.00 ], [  0.61,  0.04, -0.65 ], [ 0.00, 0.00, 0.50 ] ] ),
                (4,  [ [ -15.60, 13.70,  -6.10 ], [  7.00,  2.10,  -1.80 ], [  0.00,  1.00,  0.00 ], [  0.00,  0.05,  0.00 ], [  0.50,  0.08,  0.86 ], [ 0.00, 0.00, 0.50 ] ] ),
                (5,  [ [  -9.30, 14.80,  -4.90 ], [  4.70,  0.70,   1.80 ], [  0.00,  1.00,  0.00 ], [  0.00,  0.00,  0.00 ], [ -0.23,  0.02,  0.97 ], [ 0.00, 0.00, 0.50 ] ] ),
                (6,  [ [  -3.90, 15.70,  -3.00 ], [  4.30,  0.70,   2.00 ], [  0.00,  1.00,  0.00 ], [  0.17, -1.19,  0.09 ], [ -0.29,  0.02,  0.96 ], [ 0.00, 0.00, 0.50 ] ] ),
                (7,  [ [  -3.40, 13.40,  -2.80 ], [ -4.10, -0.70,  -1.70 ], [  0.28, -0.95,  0.14 ], [  0.03, -1.11,  0.04 ], [ -0.23,  0.05,  0.97 ], [ 0.00, 0.00, 0.50 ] ] ),
                (8,  [ [  -7.60, 12.40,  -4.60 ], [ -3.70, -0.80,  -0.90 ], [  0.00, -1.00,  0.05 ], [ -0.14, -0.02, -0.04 ], [ -0.43,  0.01,  0.90 ], [ 0.00, 0.00, 0.50 ] ] ),
                (9,  [ [ -11.60, 11.60,  -5.70 ], [ -4.20, -0.70,  -0.20 ], [  0.00, -1.00,  0.05 ], [  0.09,  0.01, -0.08 ], [ -0.21,  0.03,  0.98 ], [ 0.00, 0.00, 0.50 ] ] ),
                (10, [ [ -16.50, 11.70,  -3.90 ], [ -1.00,  0.20,   5.80 ], [  0.21, -0.97, -0.14 ], [ -0.04,  0.05, -0.25 ], [  0.98,  0.17, -0.04 ], [ 0.00, 0.00, 0.50 ] ] ),
                (11, [ [ -12.50, 11.70,  -1.40 ], [  3.60,  0.10,   0.60 ], [ -0.06, -0.89, -0.45 ], [ -0.31,  0.12, -0.21 ], [ -0.02,  0.43, -0.90 ], [ 0.00, 0.00, 0.50 ] ] ),
                (12, [ [  -6.80, 11.80,  -0.60 ], [  2.90,  0.00,   0.70 ], [ -0.42, -0.72, -0.54 ], [ -0.16,  1.09,  0.46 ], [ -0.35,  0.43, -0.83 ], [ 0.00, 0.00, 0.50 ] ] ),
                (13, [ [  -6.40,  9.80,  -1.60 ], [ -2.90, -0.30,  -1.40 ], [ -0.48,  0.85,  0.21 ], [  0.07,  0.85,  0.47 ], [  0.12,  0.37, -0.92 ], [ 0.00, 0.00, 0.50 ] ] ),
                (14, [ [  -9.50,  9.50,  -2.90 ], [ -4.60,  0.00,  -1.80 ], [ -0.26,  0.89,  0.37 ], [  0.21,  0.05,  0.07 ], [  0.22,  0.44, -0.87 ], [ 0.00, 0.00, 0.50 ] ] ),
                (15, [ [ -14.30,  9.40,  -4.60 ], [ -3.40,  0.10,  -1.60 ], [ -0.06,  0.95,  0.30 ], [  0.14,  0.05, -0.18 ], [  0.46,  0.29, -0.84 ], [ 0.00, 0.00, 0.50 ] ] ),
                (16, [ [ -19.00,  9.40,  -2.90 ], [  0.30,  0.20,   6.70 ], [  0.00,  1.00,  0.00 ], [  0.17, -0.03,  0.03 ], [ -1.00,  0.00,  0.07 ], [ 0.00, 0.00, 0.50 ] ] ),
                (17, [ [ -14.50,  9.70,   0.20 ], [  3.60, -1.20,   1.00 ], [  0.28,  0.89,  0.36 ], [  0.37, -0.41,  0.31 ], [ -0.26, -0.29,  0.92 ], [ 0.00, 0.00, 0.50 ] ] ),
                (18, [ [ -12.60,  7.70,   0.70 ], [  0.60, -2.70,   0.20 ], [  0.69,  0.32,  0.65 ], [  0.15, -0.68,  0.19 ], [ -0.68,  0.03,  0.73 ], [ 0.00, 0.00, 0.50 ] ] ),
                (19, [ [ -13.10,  3.80,   0.30 ], [ -4.00, -3.60,  -1.50 ], [  0.47, -0.52,  0.71 ], [ -0.73, -0.40, -0.16 ], [ -0.46,  0.51,  0.73 ], [ 0.00, 0.00, 0.50 ] ] ),
                (20, [ [ -15.20,  5.10,  -0.80 ], [  6.00,  6.90,   1.80 ], [ -0.87, -0.39,  0.29 ], [ -0.53,  0.27,  0.04 ], [ -0.22, -0.94,  0.24 ], [ 0.00, 0.00, 0.50 ] ] ),
                (21, [ [ -17.30,  6.90,  -1.00 ], [ -2.50,  0.00,  -0.40 ], [ -0.69,  0.00,  0.73 ], [ -0.05, -0.05, -0.12 ], [ -0.55,  0.83, -0.09 ], [ 0.00, 0.00, 0.50 ] ] ),
                (22, [ [ -19.90,  6.80,  -2.50 ], [ -1.50, -1.10,  -3.40 ], [ -0.91, -0.35,  0.20 ], [  0.18, -0.33, -0.70 ], [ -0.48,  0.80, -0.36 ], [ 0.00, 0.00, 0.50 ] ] ),
                (23, [ [ -17.20,  6.30,  -5.10 ], [  4.00,  0.80,  -1.30 ], [ -0.20, -0.64, -0.74 ], [  0.52, -0.04, -0.62 ], [ -0.41,  0.74, -0.54 ], [ 0.00, 0.00, 0.50 ] ] ),
                (24, [ [ -12.20,  7.80,  -6.80 ], [  4.80,  1.70,  -0.30 ], [  0.06, -0.35, -0.93 ], [  0.24,  0.14, -0.07 ], [ -0.34,  0.87, -0.35 ], [ 0.00, 0.00, 0.50 ] ] ),
                (25, [ [  -7.90,  9.60,  -6.50 ], [  3.70,  1.70,   0.70 ], [  0.28, -0.34, -0.90 ], [ -0.23,  0.13,  0.01 ], [ -0.35,  0.84, -0.42 ], [ 0.00, 0.00, 0.50 ] ] ),
                (26, [ [  -3.80, 10.30,  -5.50 ], [  3.80, -2.70,  -0.10 ], [ -0.39, -0.09, -0.92 ], [ -0.22,  0.47,  0.06 ], [  0.33,  0.90, -0.29 ], [ 0.00, 0.00, 0.50 ] ] ),
                (27, [ [  -5.30,  7.60,  -6.40 ], [ -3.50, -1.00,  -1.30 ], [ -0.21,  0.58, -0.79 ], [ -0.10,  0.22,  0.09 ], [  0.12, -0.72, -0.68 ], [ 0.00, 0.00, 0.50 ] ] ),
                (28, [ [  -9.00,  6.40,  -7.30 ], [ -3.20, -1.30,   1.90 ], [ -0.60,  0.33, -0.73 ], [ -0.33, -0.33,  0.51 ], [  0.06, -0.89, -0.46 ], [ 0.00, 0.00, 0.50 ] ] ),
                (29, [ [ -11.60,  4.00,  -2.00 ], [  5.60, -0.20,   4.30 ], [ -0.82, -0.14,  0.55 ], [  0.58, -0.22,  0.77 ], [ -0.18, -0.93, -0.32 ], [ 0.00, 0.00, 0.50 ] ] ),
                (30, [ [  -5.90,  5.00,  -3.10 ], [  4.10,  1.20,  -1.60 ], [  0.46, -0.14,  0.88 ], [ -0.22, -0.22, -0.11 ], [  0.25, -0.93, -0.27 ], [ 0.00, 0.00, 0.50 ] ] ),
                (31, [ [  -2.50,  6.00,  -3.80 ], [  3.60,  0.70,   3.20 ], [ -0.71, -0.50,  0.50 ], [ -0.65,  0.12, -0.73 ], [  0.20, -0.87, -0.44 ], [ 0.00, 0.00, 0.50 ] ] ),
                (32, [ [  -4.10,  3.20,  -0.40 ], [ -3.50, -1.70,   2.60 ], [ -0.65,  0.26, -0.71 ], [  0.19,  0.53, -0.71 ], [  0.12, -0.89, -0.43 ], [ 0.00, 0.00, 0.50 ] ] ),
                (33, [ [  -9.70,  1.70,   2.30 ], [ -7.90, -1.00,   1.00 ], [ -0.30,  0.51, -0.81 ], [  0.69,  0.07,  0.25 ], [ -0.09, -0.85, -0.52 ], [ 0.00, 0.00, 0.50 ] ] ),
                (34, [ [ -19.00,  0.60,  -0.40 ], [  0.20,  3.70,  -6.80 ], [  0.96,  0.29,  0.00 ], [  0.12, -0.28,  0.90 ], [  0.26, -0.76, -0.59 ], [ 0.00, 0.00, 0.50 ] ] ),
                (35, [ [ -13.90,  2.30,  -5.80 ], [  4.40,  0.60,  -1.10 ], [  0.20, -0.04,  0.98 ], [ -0.57, -0.13,  0.44 ], [  0.08, -1.00, -0.06 ], [ 0.00, 0.00, 0.50 ] ] ),
                (36, [ [  -7.70,  1.20,  -4.60 ], [  3.90, -3.40,   1.50 ], [ -0.22,  0.00,  0.98 ], [ -0.31,  0.34, -0.17 ], [ -0.54, -0.84, -0.11 ], [ 0.00, 0.00, 0.50 ] ] ),
                (37, [ [  -4.80, -4.00,  -1.30 ], [ -4.20, -3.30,   3.10 ], [ -0.40,  0.67,  0.62 ], [  0.01,  0.39, -0.20 ], [ -0.82,  0.10, -0.56 ], [ 0.00, 0.00, 0.50 ] ] ),
                (38, [ [ -10.90, -6.10,  -0.60 ], [ -5.30, -1.20,  -0.90 ], [ -0.21,  0.81,  0.55 ], [  0.05,  0.11, -0.14 ], [  0.07,  0.58, -0.81 ], [ 0.00, 0.00, 0.50 ] ] ),
                (39, [ [ -19.90, -6.40,  -5.50 ], [ -0.30,  1.70, -10.50 ], [ -0.39,  0.87,  0.30 ], [  0.06,  0.09, -0.23 ], [  0.91,  0.37,  0.19 ], [ 0.00, 0.00, 0.50 ] ] ),
                (40, [ [ -10.70, -3.20,  -8.80 ], [  7.80,  0.40,   0.10 ], [ -0.09,  0.99,  0.09 ], [ -0.32, -0.44, -0.23 ], [ -0.05, -0.10,  0.99 ], [ 0.00, 0.00, 0.50 ] ] ),
                (41, [ [  -1.20, -1.90,  -7.30 ], [  0.80,  8.10,   2.50 ], [ -0.99,  0.03, -0.16 ], [  0.29, -0.98, -0.16 ], [ -0.17, -0.38,  0.91 ], [ 0.00, 0.00, 0.50 ] ] ),
                (42, [ [  -6.30,  0.50,  -8.10 ], [ -9.80, -1.20,   0.50 ], [  0.10, -0.96, -0.25 ], [  0.58, -0.56, -0.12 ], [  0.08, -0.25,  0.97 ], [ 0.00, 0.00, 0.50 ] ] ),
                (43, [ [ -16.00, -0.70,  -7.40 ], [ -7.60,  1.20,   1.50 ], [ -0.05, -0.91, -0.41 ], [ -0.62,  0.61,  0.13 ], [  0.28, -0.43,  0.86 ], [ 0.00, 0.00, 0.50 ] ] ),
                (44, [ [ -20.50,  2.30,  -6.10 ], [  3.50,  7.20,  -2.90 ], [ -0.99,  0.09, -0.09 ], [ -0.21,  0.77, -0.11 ], [ -0.17,  0.12,  0.98 ], [ 0.00, 0.00, 0.50 ] ] ),
                (45, [ [ -11.40,  2.60, -10.10 ], [ 10.40,  1.50,  -0.20 ], [ -0.08,  0.51, -0.86 ], [  0.49,  0.10, -0.38 ], [ -0.10,  0.85,  0.51 ], [ 0.00, 0.00, 0.50 ] ] ),
                (46, [ [  -3.80,  4.20,  -7.30 ], [  3.50,  0.90,   2.70 ], [  0.07,  0.36, -0.93 ], [ -0.19, -0.40,  0.24 ], [ -0.73,  0.68,  0.01 ], [ 0.00, 0.00, 0.50 ] ] ) ] ),
                
            'userAnnotationGroups': [
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '1-3',
                    'name': get_smallintestine_term('duodenum')[0],
                    'ontId': get_smallintestine_term('duodenum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '4-43',
                    'name': get_smallintestine_term('jejunum')[0],
                    'ontId': get_smallintestine_term('jejunum')[1]
                },
                {
                    '_AnnotationGroup': True,
                    'dimension': 1,
                    'identifierRanges': '44-45',
                    'name': get_smallintestine_term('ileum')[0],
                    'ontId': get_smallintestine_term('ileum')[1]
                }]
            } )
        }

    @staticmethod
    def getName():
        return '3D Small Intestine 1'

    @staticmethod
    def getParameterSetNames():
        return [
            'Default',
            'Cattle 1',
            'Human 1',
            'Mouse 1']

    @classmethod
    def getDefaultOptions(cls, parameterSetName='Default'):
        if 'Cattle 1' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Cattle 1']
        elif 'Mouse 1' in parameterSetName:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Mouse 1']
        else:
            centralPathOption = cls.centralPathDefaultScaffoldPackages['Human 1']
        options = {
            'Central path': copy.deepcopy(centralPathOption),
            'Number of segments': 80,
            'Number of elements around': 8,
            'Number of elements along segment': 3,
            'Number of elements through wall': 4,
            'Wall thickness': 0.45,
            'Mucosa relative thickness': 0.55,
            'Submucosa relative thickness': 0.15,
            'Circular muscle layer relative thickness': 0.25,
            'Longitudinal muscle layer relative thickness': 0.05,
            'Use cross derivatives': False,
            'Use linear through wall': True,
            'Refine': False,
            'Refine number of elements around': 1,
            'Refine number of elements along': 1,
            'Refine number of elements through wall': 1
            }
        if 'Cattle 1' in parameterSetName:
            options['Number of segments'] = 300
            options['Wall thickness'] = 2.0
        elif 'Mouse 1' in parameterSetName:
            options['Number of segments'] = 100
            options['Wall thickness'] = 0.1

        return options

    @staticmethod
    def getOrderedOptionNames():
        return [
            'Central path',
            'Number of segments',
            'Number of elements around',
            'Number of elements along segment',
            'Number of elements through wall',
            'Wall thickness',
            'Mucosa relative thickness',
            'Submucosa relative thickness',
            'Circular muscle layer relative thickness',
            'Longitudinal muscle layer relative thickness',
            'Use cross derivatives',
            'Use linear through wall',
            'Refine',
            'Refine number of elements around',
            'Refine number of elements along',
            'Refine number of elements through wall' ]

    @classmethod
    def getOptionValidScaffoldTypes(cls, optionName):
        if optionName == 'Central path':
            return [ MeshType_1d_path1 ]
        return []

    @classmethod
    def getOptionScaffoldTypeParameterSetNames(cls, optionName, scaffoldType):
        if optionName == 'Central path':
            return list(cls.centralPathDefaultScaffoldPackages.keys())
        assert scaffoldType in cls.getOptionValidScaffoldTypes(optionName), \
            cls.__name__ + '.getOptionScaffoldTypeParameterSetNames.  ' + \
            'Invalid option \'' + optionName + '\' scaffold type ' + scaffoldType.getName()
        return scaffoldType.getParameterSetNames()

    @classmethod
    def getOptionScaffoldPackage(cls, optionName, scaffoldType, parameterSetName=None):
        '''
        :param parameterSetName:  Name of valid parameter set for option Scaffold, or None for default.
        :return: ScaffoldPackage.
        '''
        if parameterSetName:
            assert parameterSetName in cls.getOptionScaffoldTypeParameterSetNames(optionName, scaffoldType), \
                'Invalid parameter set ' + str(parameterSetName) + ' for scaffold ' + str(scaffoldType.getName()) + \
                ' in option ' + str(optionName) + ' of scaffold ' + cls.getName()
        if optionName == 'Central path':
            if not parameterSetName:
                parameterSetName = list(cls.centralPathDefaultScaffoldPackages.keys())[0]
            return copy.deepcopy(cls.centralPathDefaultScaffoldPackages[parameterSetName])
        assert False, cls.__name__ + '.getOptionScaffoldPackage:  Option ' + optionName + ' is not a scaffold'

    @classmethod
    def checkOptions(cls, options):
        if not options['Central path'].getScaffoldType() in cls.getOptionValidScaffoldTypes('Central path'):
            options['Central path'] = cls.getOptionScaffoldPackage('Central path', MeshType_1d_path1)
        for key in [
            'Number of segments',
            'Number of elements around',
            'Number of elements along segment',
            'Refine number of elements around',
            'Refine number of elements along',
            'Refine number of elements through wall']:
            if options[key] < 1:
                options[key] = 1
        if options['Number of elements through wall'] != (1 or 4):
            options['Number of elements through wall'] = 4
        if options['Wall thickness'] < 0.0:
            options['Wall thickness'] = 0.0

    @classmethod
    def generateBaseMesh(cls, region, options):
        """
        Generate the base tricubic Hermite mesh.
        :param region: Zinc region to define model in. Must be empty.
        :param options: Dict containing options. See getDefaultOptions().
        :return: list of AnnotationGroup, None
        """
        centralPath = options['Central path']
        segmentCount = options['Number of segments']
        elementsCountAround = options['Number of elements around']
        elementsCountAlongSegment = options['Number of elements along segment']
        elementsCountThroughWall = options['Number of elements through wall']
        wallThickness = options['Wall thickness']
        mucosaRelThickness = options['Mucosa relative thickness']
        submucosaRelThickness = options['Submucosa relative thickness']
        circularRelThickness = options['Circular muscle layer relative thickness']
        longitudinalRelThickness = options['Longitudinal muscle layer relative thickness']
        useCrossDerivatives = options['Use cross derivatives']
        useCubicHermiteThroughWall = not(options['Use linear through wall'])
        elementsCountAlong = int(elementsCountAlongSegment*segmentCount)
        startPhase = 0.0

        # Small intestine coordinates
        lengthToDiameterRatio = 200
        wallThicknessToDiameterRatio = 0.15
        relativeThicknessListSmallIntestineCoordinates = [1.0 / elementsCountThroughWall
                                                          for n3 in range(elementsCountThroughWall)]

        firstNodeIdentifier = 1
        firstElementIdentifier = 1

        # Central path
        tmpRegion = region.createRegion()
        centralPath.generate(tmpRegion)
        tmpFieldmodule = tmpRegion.getFieldmodule()
        tmpNodes = tmpFieldmodule.findNodesetByFieldDomainType(Field.DOMAIN_TYPE_NODES)
        tmpCoordinates = tmpFieldmodule.findFieldByName('coordinates')
        smallIntestineTermsAlong = [None, 'duodenum', 'jejunum', 'ileum']
        arcLengthOfGroupsAlong = []

        for termName in smallIntestineTermsAlong:
            tmpGroup = tmpFieldmodule.findFieldByName(termName).castGroup() if termName else None
            tmpNodeset = tmpGroup.getNodesetGroup(tmpNodes) if tmpGroup else tmpNodes

            cxGroup, cd1Group, cd2Group, cd3Group, cd12Group, cd13Group = get_nodeset_path_field_parameters(
                tmpNodeset, tmpCoordinates,
                [Node.VALUE_LABEL_VALUE, Node.VALUE_LABEL_D_DS1,
                 Node.VALUE_LABEL_D_DS2, Node.VALUE_LABEL_D_DS3,
                 Node.VALUE_LABEL_D2_DS1DS2, Node.VALUE_LABEL_D2_DS1DS3])
            arcLength = 0.0
            for e in range(len(cxGroup) - 1):
                arcLength += interp.getCubicHermiteArcLength(cxGroup[e], cd1Group[e],
                                                             cxGroup[e + 1], cd1Group[e + 1])
            arcLengthOfGroupsAlong.append(arcLength)

            if not termName:
                cx = cxGroup
                cd1 = cd1Group
                cd2 = cd2Group
                cd3 = cd3Group
                cd12 = cd12Group
                cd13 = cd13Group

            del tmpNodeset
            del tmpGroup

        del tmpCoordinates
        del tmpNodes
        del tmpFieldmodule
        del tmpRegion

        # find arclength of colon
        length = 0.0
        elementsCountIn = len(cx) - 1
        sd1 = interp.smoothCubicHermiteDerivativesLine(cx, cd1, fixAllDirections = True,
            magnitudeScalingMode = interp.DerivativeScalingMode.HARMONIC_MEAN)

        for e in range(elementsCountIn):
            arcLength = interp.getCubicHermiteArcLength(cx[e], sd1[e], cx[e + 1], sd1[e + 1])
            # print(e+1, arcLength)
            length += arcLength
        segmentLength = length / segmentCount
        elementAlongLength = length / elementsCountAlong
        # print('Length = ', length)

        # Sample central path
        sx, sd1, se, sxi, ssf = interp.sampleCubicHermiteCurves(cx, cd1, elementsCountAlongSegment*segmentCount)
        sd2, sd12 = interp.interpolateSampleCubicHermite(cd2, cd12, se, sxi, ssf)

        innerRadiusListCP = [vector.magnitude(c) for c in cd2]
        dInnerRadiusListCP = []
        for n in range(len(innerRadiusListCP) - 1):
            dInnerRadiusListCP.append(innerRadiusListCP[n + 1] - innerRadiusListCP[n])
        dInnerRadiusListCP.append(innerRadiusListCP[-1] - innerRadiusListCP[-2])
        innerRadiusList, dInnerRadiusList = interp.interpolateSampleCubicHermite(
            innerRadiusListCP, dInnerRadiusListCP, se, sxi, ssf)

        centralPathLength = arcLengthOfGroupsAlong[0]
        elementAlongLength = centralPathLength / elementsCountAlong

        elementsCountAlongGroups = []
        groupLength = 0.0
        e = 0
        elementsCount = 1
        length = elementAlongLength
        for i in range(1, len(smallIntestineTermsAlong)):
            groupLength += arcLengthOfGroupsAlong[i]
            if e == elementsCountAlong - 2:
                elementsCount += 1
                elementsCountAlongGroups.append(elementsCount)
            else:
                while length < groupLength:
                    elementsCount += 1
                    e += 1
                    length += elementAlongLength

                # check which end is grouplength closer to
                distToUpperEnd = abs(length - groupLength)
                distToLowerEnd = abs(groupLength - (length - elementsCountAlong))
                if distToLowerEnd < distToUpperEnd:
                    elementsCount -= 1
                    elementsCountAlongGroups.append(elementsCount)
                    e -= 1
                    length -= elementAlongLength
                else:
                    elementsCountAlongGroups.append(elementsCount)
            elementsCount = 0

        # Groups along small intestine
        smallintestineGroup = AnnotationGroup(region, get_smallintestine_term("small intestine"))
        duodenumGroup = AnnotationGroup(region, get_smallintestine_term("duodenum"))
        jejunumGroup = AnnotationGroup(region, get_smallintestine_term("jejunum"))
        ileumGroup = AnnotationGroup(region, get_smallintestine_term("ileum"))

        annotationGroupAlong = [[smallintestineGroup, duodenumGroup],
                                [smallintestineGroup, jejunumGroup],
                                [smallintestineGroup, ileumGroup]]

        annotationGroupsAlong = []
        for i in range(len(elementsCountAlongGroups)):
            elementsCount = elementsCountAlongGroups[i]
            for n in range(elementsCount):
                annotationGroupsAlong.append(annotationGroupAlong[i])

        annotationGroupsAround = []
        for i in range(elementsCountAround):
            annotationGroupsAround.append([ ])

        # Groups through wall
        longitudinalMuscleGroup = AnnotationGroup(region,
                                                  get_smallintestine_term(
                                                      "longitudinal muscle layer of small intestine"))
        circularMuscleGroup = AnnotationGroup(region,
                                              get_smallintestine_term("circular muscle layer of small intestine"))
        submucosaGroup = AnnotationGroup(region, get_smallintestine_term("submucosa of small intestine"))
        mucosaGroup = AnnotationGroup(region, get_smallintestine_term("mucosa of small intestine"))

        if elementsCountThroughWall == 1:
            relativeThicknessList = [1.0]
            annotationGroupsThroughWall = [[]]
        else:
            relativeThicknessList = [mucosaRelThickness, submucosaRelThickness,
                                     circularRelThickness, longitudinalRelThickness]
            annotationGroupsThroughWall = [[mucosaGroup], [submucosaGroup],
                                           [circularMuscleGroup], [longitudinalMuscleGroup]]

        xExtrude = []
        d1Extrude = []
        d2Extrude = []
        d3UnitExtrude = []

        # Create object
        smallIntestineSegmentTubeMeshInnerPoints = CylindricalSegmentTubeMeshInnerPoints(
            elementsCountAround, elementsCountAlongSegment, segmentLength,
            wallThickness, innerRadiusList, startPhase)

        for nSegment in range(segmentCount):
            # Create inner points
            xInner, d1Inner, d2Inner, transitElementList, segmentAxis, radiusAlongSegmentList = \
               smallIntestineSegmentTubeMeshInnerPoints.getCylindricalSegmentTubeMeshInnerPoints(nSegment)

            # Project reference point for warping onto central path
            start = nSegment*elementsCountAlongSegment
            end = (nSegment + 1)*elementsCountAlongSegment + 1
            sxRefList, sd1RefList, sd2ProjectedListRef, zRefList = \
                tubemesh.getPlaneProjectionOnCentralPath(xInner, elementsCountAround, elementsCountAlongSegment,
                                                         segmentLength, sx[start:end], sd1[start:end], sd2[start:end],
                                                         sd12[start:end])

            # Warp segment points
            xWarpedList, d1WarpedList, d2WarpedList, d3WarpedUnitList = tubemesh.warpSegmentPoints(
                xInner, d1Inner, d2Inner, segmentAxis, sxRefList, sd1RefList, sd2ProjectedListRef,
                elementsCountAround, elementsCountAlongSegment, zRefList)

            # Store points along length
            xExtrude = xExtrude + (xWarpedList if nSegment == 0 else xWarpedList[elementsCountAround:])
            d1Extrude = d1Extrude + (d1WarpedList if nSegment == 0 else d1WarpedList[elementsCountAround:])

            # Smooth d2 for nodes between segments and recalculate d3
            if nSegment == 0:
                d2Extrude = d2Extrude + (d2WarpedList[:-elementsCountAround])
                d3UnitExtrude = d3UnitExtrude + (d3WarpedUnitList[:-elementsCountAround])
            else:
                xSecondFace = xWarpedList[elementsCountAround:elementsCountAround*2]
                d2SecondFace = d2WarpedList[elementsCountAround:elementsCountAround*2]
                for n1 in range(elementsCountAround):
                    nx = [xLastTwoFaces[n1], xLastTwoFaces[n1 + elementsCountAround], xSecondFace[n1]]
                    nd2 = [d2LastTwoFaces[n1], d2LastTwoFaces[n1 + elementsCountAround], d2SecondFace[n1]]
                    d2 = interp.smoothCubicHermiteDerivativesLine(nx, nd2, fixStartDerivative = True,
                                                                  fixEndDerivative = True)[1]
                    d2Extrude.append(d2)
                    d3Unit = vector.normalise(vector.crossproduct3(vector.normalise(d1LastTwoFaces[n1 + elementsCountAround]),
                                                                   vector.normalise(d2)))
                    d3UnitExtrude.append(d3Unit)
                d2Extrude = d2Extrude + \
                            (d2WarpedList[elementsCountAround:-elementsCountAround] if nSegment < segmentCount - 1 else
                             d2WarpedList[elementsCountAround:])
                d3UnitExtrude = d3UnitExtrude + \
                                (d3WarpedUnitList[elementsCountAround:-elementsCountAround] if nSegment < segmentCount - 1 else
                                 d3WarpedUnitList[elementsCountAround:])
            xLastTwoFaces = xWarpedList[-elementsCountAround*2:]
            d1LastTwoFaces = d1WarpedList[-elementsCountAround*2:]
            d2LastTwoFaces = d2WarpedList[-elementsCountAround*2:]

        # Create coordinates and derivatives
        xList, d1List, d2List, d3List, curvatureList = tubemesh.extrudeSurfaceCoordinates(xExtrude, d1Extrude,
            d2Extrude, d3UnitExtrude, [wallThickness]*(elementsCountAlong+1), relativeThicknessList,
            elementsCountAround, elementsCountAlong, elementsCountThroughWall, transitElementList, outward=True)

        flatWidthList, xiList = smallIntestineSegmentTubeMeshInnerPoints.getFlatWidthAndXiList()

        # Create flat coordinates
        xFlat, d1Flat, d2Flat = tubemesh.createFlatCoordinates(
            xiList, flatWidthList, length, wallThickness, relativeThicknessList, elementsCountAround,
            elementsCountAlong, elementsCountThroughWall, transitElementList)

        # Create small intestine coordinates
        xSmallIntestine, d1SmallIntestine, d2SmallIntestine = \
            tubemesh.createOrganCoordinates(xiList, relativeThicknessListSmallIntestineCoordinates,
                                            lengthToDiameterRatio, wallThicknessToDiameterRatio,
                                            elementsCountAround, elementsCountAlong, elementsCountThroughWall,
                                            transitElementList)

        # Create nodes and elements
        nextNodeIdentifier, nextElementIdentifier, annotationGroups = tubemesh.createNodesAndElements(
            region, xList, d1List, d2List, d3List, xFlat, d1Flat, d2Flat,
            xSmallIntestine, d1SmallIntestine, d2SmallIntestine, "small intestine coordinates",
            elementsCountAround, elementsCountAlong, elementsCountThroughWall,
            annotationGroupsAround, annotationGroupsAlong, annotationGroupsThroughWall,
            firstNodeIdentifier, firstElementIdentifier, useCubicHermiteThroughWall, useCrossDerivatives,
            closedProximalEnd=False)

        return annotationGroups, None

    @classmethod
    def refineMesh(cls, meshrefinement, options):
        """
        Refine source mesh into separate region, with change of basis.
        :param meshrefinement: MeshRefinement, which knows source and target region.
        :param options: Dict containing options. See getDefaultOptions().
        """
        refineElementsCountAround = options['Refine number of elements around']
        refineElementsCountAlong = options['Refine number of elements along']
        refineElementsCountThroughWall = options['Refine number of elements through wall']

        meshrefinement.refineAllElementsCubeStandard3d(refineElementsCountAround, refineElementsCountAlong,
                                                       refineElementsCountThroughWall)
        return

    @classmethod
    def defineFaceAnnotations(cls, region, options, annotationGroups):
        '''
        Add face annotation groups from the highest dimension mesh.
        Must have defined faces and added subelements for highest dimension groups.
        :param region: Zinc region containing model.
        :param options: Dict containing options. See getDefaultOptions().
        :param annotationGroups: List of annotation groups for top-level elements.
        New face annotation groups are appended to this list.
        '''
        # Create 2d surface mesh groups
        fm = region.getFieldmodule()
        mesh2d = fm.findMeshByDimension(2)

        smallIntestineGroup = getAnnotationGroupForTerm(annotationGroups, get_smallintestine_term("small intestine"))
        is_exterior = fm.createFieldIsExterior()
        is_exterior_face_xi3_1 = fm.createFieldAnd(is_exterior, fm.createFieldIsOnFace(Element.FACE_TYPE_XI3_1))
        is_exterior_face_xi3_0 = fm.createFieldAnd(is_exterior, fm.createFieldIsOnFace(Element.FACE_TYPE_XI3_0))
        is_smallIntestine = smallIntestineGroup.getGroup()
        is_serosa = fm.createFieldAnd(is_smallIntestine, is_exterior_face_xi3_1)
        serosa = findOrCreateAnnotationGroupForTerm(annotationGroups, region,
                                                    get_smallintestine_term("serosa of small intestine"))
        serosa.getMeshGroup(mesh2d).addElementsConditional(is_serosa)

        duodenumGroup = getAnnotationGroupForTerm(annotationGroups, get_smallintestine_term("duodenum"))
        duodenumSerosa = findOrCreateAnnotationGroupForTerm(annotationGroups, region,
                                                                 get_smallintestine_term("serosa of duodenum"))
        is_duodenum = duodenumGroup.getGroup()
        is_duodenumSerosa = fm.createFieldAnd(is_duodenum, is_exterior_face_xi3_1)
        duodenumSerosa.getMeshGroup(mesh2d).addElementsConditional(is_duodenumSerosa)

        is_duodenumLuminal = fm.createFieldAnd(is_duodenum, is_exterior_face_xi3_0)
        duodenumLuminal = findOrCreateAnnotationGroupForTerm(annotationGroups, region,
                                                             get_smallintestine_term(
                                                                 "luminal surface of duodenum"))
        duodenumLuminal.getMeshGroup(mesh2d).addElementsConditional(is_duodenumLuminal)

